/*
 * Copyright (c) 2009-2010 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/* This code is associated to IPOL publication:
*
* Image Color Cube Dimensional Filtering an Visualization,
* J.L. Lisani, A. Buades, J-M. Morel, 2011 
*
*/

/**
 * @file rgbcubeIPOL.cpp
 *
 * @brief Auxiliary functions for color filtering and visualization
 *
 * @version 1.0 
 * @author Jose Luis Lisani; <joseluis.lisani@uib.es> 
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "rgbcube.h"


/**
* \brief Normalize R^3 vector
*
* @param[in/out] input/output vector
*/
void normalize_vector3D(float u[3])
{
  float d=sqrt(u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
  u[0]/=d;
  u[1]/=d;
  u[2]/=d;
}

/**
* \brief Scalar product of two R^3 vectors 
*
* @param[in] u, v input vectors
* \return scalar product (float)
*
*/
float scalar_product3D(float u[3], float v[3])
{
  float p=u[0]*v[0]+u[1]*v[1]+u[2]*v[2];
  return p;
}

/**
* \brief Vectorial product of two R^3 vectors 
*
* @param[in] u, v input vectors
* @param[out] w output vector
*
*/
void vectorial_product3D(float u[3], float v[3], float w[3])
{
  w[0]=u[1]*v[2]-u[2]*v[1];
  w[1]=u[2]*v[0]-u[0]*v[2];
  w[2]=u[0]*v[1]-u[1]*v[0];
}

/**
* \brief Length of R^3 vector 
*
* @param[in] u input vector
* \return vector norm (float)
*
*/
float vector_norm3D(float u[3])
{
  float d=sqrt(u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
  return d;
}



/**
* \brief Samples an arc of circunference in 3D
*
* @param[in] A, B end points of the arc
* @param[in] O center of the circunference
* @param[in] npoints number of sampling points (including both end-points)
*
* \return list of sampled points
*
*/
float **sample_arclength(float A[3], float B[3], float O[3], int npoints)
{
  float **P;
  float u[3], v[3], vv[3], w[3];
  float unorm, vvnorm, R, x, y;
  float angmax, stepang;

  if (npoints == 1) return NULL;

  //initialize list
  P=new float*[npoints];
  for (int n=0; n < npoints; n++) P[n]=new float[3];

  //u: vector in the direction OA
  u[0]=A[0]-O[0];
  u[1]=A[1]-O[1];
  u[2]=A[2]-O[2];
  unorm=vector_norm3D(u);
  //vv: vector in the direction OB
  vv[0]=B[0]-O[0];
  vv[1]=B[1]-O[1];
  vv[2]=B[2]-O[2];
  vvnorm=vector_norm3D(vv);
  //Since OA and OB belong to the same circunference centered at O, 
  //unorm=vvnorm=R (radius of circunference)
  R=unorm;

  //angle from OA to OB
  angmax=acos((double) (scalar_product3D(u, vv)/(unorm*vvnorm)));

  //w: normal vector to the plane containing A-O-B
  normalize_vector3D(u);
  normalize_vector3D(vv);
  vectorial_product3D(u, vv, w);
  normalize_vector3D(w);
  //v: vector on the plane, normal to u
  vectorial_product3D(w, u, v);
  normalize_vector3D(v);
  //u-v define an orthogonal coordinates system on the plane A-O-B

  //angular step
  stepang=angmax/(npoints-1);

  //get samples
  //first point
  P[0][0]=A[0];
  P[0][1]=A[1];
  P[0][2]=A[2];
  for (int n=1; n < npoints-1; n++) {
    x=R*cos((double) (n*stepang));
    y=R*sin((double) (n*stepang));
    P[n][0]=O[0]+x*u[0]+y*v[0];
    P[n][1]=O[1]+x*u[1]+y*v[1];
    P[n][2]=O[2]+x*u[2]+y*v[2];
  }
  //last point
  P[npoints-1][0]=B[0];
  P[npoints-1][1]=B[1];
  P[npoints-1][2]=B[2];

  //return list
  return P;
}


/**
* \brief Delete a list of 3D points
*
* @param[in] points list of points
* @param[in] npoints list of points
*
*/
void delete_list3D(float **points, int npoints)
{
  for (int n=0; n < npoints; n++) delete[] points[n];
  delete[] points;
}


/**
* \brief Generate the parameters of the projection planes for a 
* rotating sequence of views of the RGB cube
*
* @param[out] paramsfile output file name (text file)
*
* \return number of views in the sequence
*
*/
int RGBviews_sequence_params(const char *paramsfile)
{
  int nviews=0;
  int idview;

  //Notation:
  //A=(255, 255, 255) starting point of the sequence
  //O=(127.5, 127.5, 127.5) center of color cube, all the views point 
  //at this center
  //S sphere centered at O and containing all the vertexes of the cube
  //R=sqrt(3)*127.5 radius of S
  //B=(127.5, 127.5, 127.5+R) point in S in the vertical of O (above O)
  //C=(0, 255, 255)
  //D=(0, 255, 0)
  //E=(127.5, 127.5, 127.5-R) point in S in the vertical of O (below O)
  //F=(0, 0, 0)
  //G=(0, 0, 255)
  //H=(255, 0, 255)

  //Remarks:
  //alpha=angle OA-OB=angle OB-OC=angle OD-OE=angle OE-OF=54.736 degrees
  //beta=angle OC-OD=70.53 degrees

  float R;
  float A[3], O[3], B[3], C[3], D[3], E[3], F[3], G[3], H[3];
  A[0]=255.0f; A[1]=255.0f; A[2]=255.0f;
  O[0]=127.5f; O[1]=127.5f; O[2]=127.5f;
  R=(float) sqrt((A[0]-O[0])*(A[0]-O[0])+(A[1]-O[1])*(A[1]-O[1])+(A[2]-O[2])*(A[2]-O[2]));
  B[0]=O[0]; B[1]=O[1]; B[2]=O[2]+R;
  C[0]=0.0f; C[1]=255.0f; C[2]=255.0f;
  D[0]=0.0f; D[1]=255.0f; D[2]=0.0f;
  E[0]=O[0]; E[1]=O[1]; E[2]=O[2]-R;
  F[0]=0.0f; F[1]=0.0f; F[2]=0.0f;
  G[0]=0.0f; G[1]=0.0f; G[2]=255.0f;
  H[0]=255.0f; H[1]=0.0f; H[2]=255.0f;


  int nsamples1; //number of samples of arc OA-OB=arc OB-OC=arc OD-OE=
                 //arc OE-OF (including both end-points)
  int nsamples2; //number of samples of arc OC-OD (including both end points)


  nsamples1=6; //that is, 10.95 degrees between samples
  nsamples2=7; //that is, 11.75 degrees between samples
               //nviews=27 


  nviews=(nsamples1-1)*4+(nsamples2-1)+1; //end points are only counted once
  printf("%i views\n", nviews);

  //Save data to a text file
  FILE *params=fopen(paramsfile, "w");
  fprintf(params, "%i\n", nviews);

  //Generate views parameters
  idview=0;
  float **P; //list of sampling points
  float w[3]; //normal vector to the projection plane
  float u[3]; //direction of horizontal projection onto the plane
  float v[3]; //direction of vertical projection onto the plane
  //Remark: u x v = w (vectorial product)

  //Get sampling points

  //sample OA-OB
  P=sample_arclength(A, B, O, nsamples1);
  //direction of horizontal projection: HC
  u[0]=C[0]-H[0];
  u[1]=C[1]-H[1];
  u[2]=C[2]-H[2];
  normalize_vector3D(u);
  for (int n=0; n < nsamples1-1; n++) { //do not use last end-point
    w[0]=O[0]-P[n][0];
    w[1]=O[1]-P[n][1];
    w[2]=O[2]-P[n][2];
    normalize_vector3D(w);
    vectorial_product3D(w, u, v);
    normalize_vector3D(v);
    //save params
    fprintf(params, "%f %f %f %f %f %f %f %f %f\n", 
            P[n][0], P[n][1], P[n][2], u[0], u[1], u[2], v[0], v[1], v[2]);
    idview++;
  }
  //delete samples
  delete_list3D(P, nsamples1);

  //sample OB-OC
  P=sample_arclength(B, C, O, nsamples1);
  //direction of vertical projection: GA
  v[0]=A[0]-G[0];
  v[1]=A[1]-G[1];
  v[2]=A[2]-G[2];
  normalize_vector3D(v);
  for (int n=0; n < nsamples1-1; n++) { //do not use last end-point
    w[0]=O[0]-P[n][0];
    w[1]=O[1]-P[n][1];
    w[2]=O[2]-P[n][2];
    normalize_vector3D(w);
    vectorial_product3D(v, w, u);
    normalize_vector3D(u);
    //save params
    fprintf(params, "%f %f %f %f %f %f %f %f %f\n", 
            P[n][0], P[n][1], P[n][2], u[0], u[1], u[2], v[0], v[1], v[2]);
    idview++;
  }
  //delete samples
  delete_list3D(P, nsamples1);

  //sample OC-OD
  P=sample_arclength(C, D, O, nsamples2);
  //direction of vertical projection: AG
  //(same as before)
  for (int n=0; n < nsamples2-1; n++) { //do not use last end-point
    w[0]=O[0]-P[n][0];
    w[1]=O[1]-P[n][1];
    w[2]=O[2]-P[n][2];
    normalize_vector3D(w);
    vectorial_product3D(v, w, u);
    normalize_vector3D(u);
    //save params
    fprintf(params, "%f %f %f %f %f %f %f %f %f\n", 
            P[n][0], P[n][1], P[n][2], u[0], u[1], u[2], v[0], v[1], v[2]);
    idview++;
  }
  //delete samples
  delete_list3D(P, nsamples2);

  //sample OD-OE
  P=sample_arclength(D, E, O, nsamples1);
  //direction of vertical projection: AG
  //(same as before)
  for (int n=0; n < nsamples1-1; n++) { //do not use last end-point
    w[0]=O[0]-P[n][0];
    w[1]=O[1]-P[n][1];
    w[2]=O[2]-P[n][2];
    normalize_vector3D(w);
    vectorial_product3D(v, w, u);
    normalize_vector3D(u);
    //save params
    fprintf(params, "%f %f %f %f %f %f %f %f %f\n", 
            P[n][0], P[n][1], P[n][2], u[0], u[1], u[2], v[0], v[1], v[2]);
    idview++;
  }
  //delete samples
  delete_list3D(P, nsamples1);

  //sample OE-OF
  P=sample_arclength(E, F, O, nsamples1);
  //direction of horizontal projection: CH
  u[0]=H[0]-C[0];
  u[1]=H[1]-C[1];
  u[2]=H[2]-C[2];
  normalize_vector3D(u);
  for (int n=0; n < nsamples1; n++) { // use last end-point
    w[0]=O[0]-P[n][0];
    w[1]=O[1]-P[n][1];
    w[2]=O[2]-P[n][2];
    normalize_vector3D(w);
    vectorial_product3D(w, u, v);
    normalize_vector3D(v);
    //save params
    fprintf(params, "%f %f %f %f %f %f %f %f %f\n", 
            P[n][0], P[n][1], P[n][2], u[0], u[1], u[2], v[0], v[1], v[2]);
    idview++;
  }
  //delete samples
  delete_list3D(P, nsamples1);

  fclose(params);
  return nviews;
}

/**
* \brief Create a set of color points along the edges of the RGB color cube 
*
* @param[out] Rout, Gout, Bout arrays containing the RGB values of the created 
*                              set of points
* @param[out] npoints number of points in the set
*
*/
void cube_edges(float **Rout, float **Gout, float **Bout, int &npoints)
{
  float s, disc=0.2f;
  int ndisc=(int) (1/disc);
  int npointsmax=12*256*(ndisc+1); //12 edges, 256 integer colors per edge, 
                                   //ndisc points between integer colors
  //create output arrays
  float *R=new float[npointsmax];
  float *G=new float[npointsmax];
  float *B=new float[npointsmax];
  //fill arrays of RGB values
  npoints=0;
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=0; G[npoints]=0; B[npoints]=s; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=255; G[npoints]=0; B[npoints]=s; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=0; G[npoints]=255; B[npoints]=s; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=255; G[npoints]=255; B[npoints]=s; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=0; G[npoints]=s; B[npoints]=0; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=0; G[npoints]=s; B[npoints]=255; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=255; G[npoints]=s; B[npoints]=0; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=255; G[npoints]=s; B[npoints]=255; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=s; G[npoints]=0; B[npoints]=0; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=s; G[npoints]=255; B[npoints]=0; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=s; G[npoints]=0; B[npoints]=255; npoints++; }
  for (s=0; s <= 255; s+=disc) { 
    R[npoints]=s; G[npoints]=255; B[npoints]=255; npoints++; }

  *Rout=R; *Gout=G; *Bout=B;
}


//option=1 -> project to line
//option=2 -> project to plane
//u: normalized vector
//d: distance to line/plane (signed for planes)
/**
* \brief Find the projection of a R^3 point onto a line or a plane
*
* @param[in] P input point
* @param[out] PP projected point
* @param[in] M point belonging to the line or plane
* @param[in] u unitary direction vector of the line or orthonormal vector to
*              the plane
* @param[out] d distance from original point to its projection
* @param[in] option option=1 projection onto line, 
*                   otherwise projection onto plane
*
*/
void project_point(float P[3], float PP[3], float M[3], float u[3], float &d, 
                   unsigned char option)
{
  if (option == 1) {
    float dp[3], lambda;
    //project point to line
    dp[0]=P[0]-M[0];
    dp[1]=P[1]-M[1];
    dp[2]=P[2]-M[2];
    lambda=scalar_product3D(dp, u);
    PP[0]=M[0]+lambda*u[0];
    PP[1]=M[1]+lambda*u[1];
    PP[2]=M[2]+lambda*u[2];
    //distance point to line
    dp[0]=PP[0]-P[0];
    dp[1]=PP[1]-P[1];
    dp[2]=PP[2]-P[2];
    d=vector_norm3D(dp);
  } else {
    //project point to plane
    float lambda;
    lambda=scalar_product3D(u, M)-scalar_product3D(u, P);
    PP[0]=P[0]+lambda*u[0];
    PP[1]=P[1]+lambda*u[1];
    PP[2]=P[2]+lambda*u[2];
    d=lambda;
  }
}



//u, v: normalized vectors
/**
* \brief Create image displaying the projection of the color cube of a color 
* image onto a plane (in this case the center of projection is located at the 
* center of the output image, which was not necessarily the case in
* 'projectRGBcube')
*
* @param[in] R, G, B color channels of the input image
* @param[in] npoints image size
* @param[in] IR, IG, IB if not NULL, draw pixel n with color (IR[n] IG[n], IB[n]), else use color 
*            (R[n], G[n], B[n])
* @param[out] Rout, Gout, Bout color channels of the output image
* @param[in] wout, hout dimensions of the output image
* @param[in] M origin of the plane
* @param[in] u, v unitary vectors defining the coordinates system of the 
*                 projection plane
*
*/
void projectRGBcubeB(unsigned char *R, unsigned char *G, unsigned char *B, 
                     int npoints, unsigned char *IR, unsigned char *IG, unsigned char *IB, 
                     unsigned char *Rout, unsigned char *Gout, 
                     unsigned char *Bout, 
                     int wout, int hout, 
                     float M[3], float u[3], float v[3])
{
  float **cplane; //coordinates of RGB points w.r.t. the plane u-v
  float w[3], P[3], PP[3], PPP[3], d;
  float zmin, zmax, *dplane;
  int x, y, smax, npointsB;
  float *RB, *GB, *BB;

  //generate cube edges
  RB=GB=BB=NULL;
  npointsB=0;
  cube_edges(&RB, &GB, &BB, npointsB); 

  //create an array to store the set of R^3 points to be projected
  cplane=new float*[npoints+npointsB];
  for (int n=0; n < npoints+npointsB; n++) cplane[n]=new float[3];

  //normal vector to the plane
  vectorial_product3D(u, v, w);
  normalize_vector3D(w);

  //fill the set of points to be projected: image points + cube edges 
  for (int n=0; n < npoints+npointsB; n++) {
    if (n < npoints) { //image points
      P[0]=(float) R[n]; P[1]=(float) G[n]; P[2]=(float) B[n]; 
    } else { //cube edges
      P[0]=(float) RB[n-npoints]; 
      P[1]=(float) GB[n-npoints]; 
      P[2]=(float) BB[n-npoints]; 
    }
    //project point onto the plane and store the distance to the projection
    project_point(P, PP, M, w, d, 2);
    PPP[0]=PP[0]-M[0]; PPP[1]=PP[1]-M[1]; PPP[2]=PP[2]-M[2];
    cplane[n][0]=scalar_product3D(PPP, u); 
    cplane[n][1]=scalar_product3D(PPP, v); 
    cplane[n][2]=d;
  }

  //find maximum and minimum of coordinates of the projected points
  //find maximum and minimum distance from color cube to the projection plane
  zmin=zmax=cplane[0][2];
  for (int n=1; n < npoints+npointsB; n++) {
    if (cplane[n][2] < zmin) zmin=cplane[n][2];
    if (cplane[n][2] > zmax) zmax=cplane[n][2];
  }

  //maximum distance between projected points for any view: 
  smax=255.0*(float)sqrt((float) 3);

  //output image
  //array storing distances to projection plane: 
  //d(x, y)=distance to plane from point projected at normalized position (x,y)
  dplane=new float[wout*hout];
  for (int n=0; n < wout*hout; n++) dplane[n]=-1;
  memset(Rout, 255, wout*hout);
  memset(Gout, 255, wout*hout);
  memset(Bout, 255, wout*hout);
  for (int n=0; n < npoints+npointsB; n++) {
    //normalized coordinates of projection: 0 <= x < wout, 0 <= y < hout
    x=(int) ((cplane[n][0]*((float) wout-1)/smax)+wout/2);
    y=(int) ((cplane[n][1]*((float) hout-1)/smax)+hout/2);
    if ((x < wout) && (y < hout) && (x >= 0) && (y >= 0)) {
      //if two or more points are to be projected at the same position (x, y), 
      //display the one closest to the observer, ie the one whose distance 
      //to the plane is bigger
      if (cplane[n][2]-zmin > dplane[x+y*wout]) {
        if (n < npoints) { //color cloud
          if (!IR) {
            Rout[x+y*wout]=R[n];
            Gout[x+y*wout]=G[n];
            Bout[x+y*wout]=B[n];
          } else {
            Rout[x+y*wout]=IR[n];
            Gout[x+y*wout]=IG[n];
            Bout[x+y*wout]=IB[n];
          }
          // dilate projected point
          if (x+1 < wout) {
            if (!IR) {
              Rout[x+1+y*wout]=R[n];
              Gout[x+1+y*wout]=G[n];
              Bout[x+1+y*wout]=B[n];
            } else {
              Rout[x+1+y*wout]=IR[n];
              Gout[x+1+y*wout]=IG[n];
              Bout[x+1+y*wout]=IB[n];
            }
          }
          if (y+1 < hout) {
            if (!IR) {
              Rout[x+(y+1)*wout]=R[n];
              Gout[x+(y+1)*wout]=G[n];
              Bout[x+(y+1)*wout]=B[n];
            } else {
              Rout[x+(y+1)*wout]=IR[n];
              Gout[x+(y+1)*wout]=IG[n];
              Bout[x+(y+1)*wout]=IB[n];
            }
          }
          if ((x+1 < wout) && (y+1 < hout)) {
            if (!IR) {
              Rout[x+1+(y+1)*wout]=R[n];
              Gout[x+1+(y+1)*wout]=G[n];
              Bout[x+1+(y+1)*wout]=B[n];
            } else {
              Rout[x+1+(y+1)*wout]=IR[n];
              Gout[x+1+(y+1)*wout]=IG[n];
              Bout[x+1+(y+1)*wout]=IB[n];
            }
          }  
          //
        } else { //cube edges
          Rout[x+y*wout]=(int) (RB[n-npoints]+0.5f);
          Gout[x+y*wout]=(int) (GB[n-npoints]+0.5f);
          Bout[x+y*wout]=(int) (BB[n-npoints]+0.5f);
        }
        //update distance to plane for point projected at normalized 
        //position (x, y)
        dplane[x+y*wout]=cplane[n][2]-zmin;
      }
    }
  }

  for (int n=0; n < npoints+npointsB; n++) delete[] cplane[n];
  delete[] cplane;
  delete[] dplane;

  if (RB) delete[] RB;
  if (GB) delete[] GB;
  if (BB) delete[] BB;
}


