/*
 * Copyright (c) 2011 Toni Buades <toni.buades@uib.es>
 *
 * Based on code from Megawave libraries created by Jacques Froment
 *
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "splinesMW.h"


////////////////////////////////////////////////////////////////////////////////
//! Begin Spline Interpolation MEGAWAVE
////////////////////////////////////////////////////////////////////////////////



//! extract image value (even outside image domain) 
float vMW(float *in,int x,int y,float bg, int width, int height)
{
  if (x<0 || x>=width || y<0 || y>=height)
    return bg; 
  else return(in[y*width+x]);
}





//! c[] = values of interpolation function at ...,t-2,t-1,t,t+1,... 
//! coefficients for cubic interpolant (Keys' function)
void keysMW(float *c,float t,float a)
{
  float t2,at;
  
  t2 = t*t;
  at = a*t;
  c[0] = a*t2*(1.0-t);
  c[1] = (2.0*a+3.0 - (a+2.0)*t)*t2 - at;
  c[2] = ((a+2.0)*t - a-3.0)*t2 + 1.0;
  c[3] = a*(t-2.0)*t2 + at;
}



//! Coefficients for cubic spline
void spline3MW(float *c,float t)
{
  float tmp;
  
  tmp = 1.-t;
  c[0] = 0.1666666666*t*t*t;
  c[1] = 0.6666666666-0.5*tmp*tmp*(1.+t);
  c[2] = 0.6666666666-0.5*t*t*(2.-t);
  c[3] = 0.1666666666*tmp*tmp*tmp;
}






//! Init spline n
void init_splinenMW(float *a,int n)
{
  int k;
  
  a[0] = 1.;
  for (k=2;k<=n;k++) a[0]/=(float)k;
  for (k=1;k<=n+1;k++)
    a[k] = - a[k-1] *(float)(n+2-k)/(float)k;
}


//! Fast integral power function
float ipowMW(float x,int n)
{
  float res;
  
  for (res=1.;n;n>>=1) {
    if (n&1) res*=x;
    x*=x;
  }
  return(res);
}


//! coefficients for spline of order >3 
void splinenMW(float *c,float t,float *a,int n)
{
  int i,k;
  float xn;
  
  memset((void *)c,0,(n+1)*sizeof(float));
  for (k=0;k<=n+1;k++) { 
    xn = ipowMW(t+(float)k,n);
    for (i=k;i<=n;i++) 
      c[i] += a[i-k]*xn;
  }
}




//! called by invspline1d
//! takes c as input and output some kind of sum combining c and z
double initcausalMW(double *c,int n,double z)
{
  double zk,z2k,iz,sum;
  int k;
  
  zk = z; iz = 1./z;
  z2k = pow(z,(double)n-1.);
  sum = c[0] + z2k * c[n-1];
  z2k = z2k*z2k*iz;
  for (k=1;k<=n-2;k++) {
    sum += (zk+z2k)*c[k];
    zk *= z;
    z2k *= iz;
  }
  return (sum/(1.-zk*zk));
}



//! called by invspline1d
//! takes c as input and output some kind of sum combining c and z
double initanticausalMW(double *c,int n,double z)
{
  return((z/(z*z-1.))*(z*c[n-2]+c[n-1]));
}



//! called by finvspline
void invspline1DMW(double *c,int size,double *z,int npoles)
{
  double lambda;
  int n,k;
  
  /* normalization */
  for (k=npoles,lambda=1.;k--;) lambda *= (1.-z[k])*(1.-1./z[k]);
  for (n=size;n--;) c[n] *= lambda;
  
  /*----- Loop on poles -----*/
  for (k=0;k<npoles;k++) {
    
    /* forward recursion */
    c[0] = initcausalMW(c,size,z[k]);
    for (n=1;n<size;n++) 
      c[n] += z[k]*c[n-1];
    
    /* backwards recursion */
    c[size-1] = initanticausalMW(c,size,z[k]);
    for (n=size-1;n--;) 
      c[n] = z[k]*(c[n+1]-c[n]);
    
  }
}



//! Main function filling coefficients
void finvsplineMW(float *in,int order,float *out, int width, int height)
{
  double *c,*d,z[5];
  int npoles,nx,ny,x,y;
  
  ny = height; nx = width;
  
  /* initialize poles of associated z-filter */
  switch (order) 
  {
    case 2: z[0]=-0.17157288;  /* sqrt(8)-3 */
      break;
      
    case 3: z[0]=-0.26794919;  /* sqrt(3)-2 */ 
      break;
      
    case 4: z[0]=-0.361341; z[1]=-0.0137254;
      break;
      
    case 5: z[0]=-0.430575; z[1]=-0.0430963;
      break;
      
    case 6: z[0]=-0.488295; z[1]=-0.0816793; z[2]=-0.00141415;
      break;
      
    case 7: z[0]=-0.53528; z[1]=-0.122555; z[2]=-0.00914869;
      break;
      
    case 8: z[0]=-0.574687; z[1]=-0.163035; z[2]=-0.0236323; z[3]=-0.000153821;
      break;
      
    case 9: z[0]=-0.607997; z[1]=-0.201751; z[2]=-0.0432226; z[3]=-0.00212131;
      break;
      
    case 10: z[0]=-0.636551; z[1]=-0.238183; z[2]=-0.065727; z[3]=-0.00752819;
      z[4]=-0.0000169828;
      break;
      
    case 11: z[0]=-0.661266; z[1]=-0.27218; z[2]=-0.0897596; z[3]=-0.0166696; 
      z[4]=-0.000510558;
      break;
      
    default:
      printf("finvspline: order should be in 2..11.\n");
      exit(-1);
  }
  
  npoles = order/2;
  
  /* initialize double array containing image */
  c = (double *)malloc(nx*ny*sizeof(double));
  d = (double *)malloc(nx*ny*sizeof(double));
  for (x=nx*ny;x--;) 
    c[x] = (double)in[x];
  
  /* apply filter on lines */
  for (y=0;y<ny;y++) 
    invspline1DMW(c+y*nx,nx,z,npoles);
  
  /* transpose */
  for (x=0;x<nx;x++)
    for (y=0;y<ny;y++) 
      d[x*ny+y] = c[y*nx+x];
  
  /* apply filter on columns */
  for (x=0;x<nx;x++) 
    invspline1DMW(d+x*ny,ny,z,npoles);
  
  /* transpose directy into image */
  for (x=0;x<nx;x++)
    for (y=0;y<ny;y++) 
      out[y*nx+x] = (float)(d[x*ny+y]);
  
  /* free array */
  free(d);
  free(c);
}



float evaluate_splineMW(float *input, float *ref, float xp, float yp, float *ak, int order, float bg, int width, int height)
{
  
  float  cx[12],cy[12];
  float p=-0.5;
  
  int xi = (int)floor((double)xp); 
  int yi = (int)floor((double)yp);
  
  float res;
  int n1, n2;
  
  if (order == 0) 
  { 
    if (xi<0 || xi>=width || yi<0 || yi>=height)        res = bg; 
    else res = input[yi*width+xi];
    
  } else 
  { 
    
    
    if (xp<0. || xp>=(float)width || yp<0. || yp>=(float)height) res=bg; 
    else {
      
      //xp -= 0.5; yp -= 0.5;
      
      float ux = xp-(float)xi;
      float uy = yp-(float)yi;
      
      switch (order) 
      {
        case 1:  
          n2 = 1;
          cx[0]=ux; cx[1]=1.-ux;
          cy[0]=uy; cy[1]=1.-uy;
          break;
          
        case -3:  
          n2 = 2;
          keysMW(cx,ux,p);
          keysMW(cy,uy,p);
          break;
          
        case 3:  
          n2 = 2;
          spline3MW(cx,ux);
          spline3MW(cy,uy);
          break;
          
        default:  
          n2 = (1+order)/2;
          splinenMW(cx,ux,ak,order);
          splinenMW(cy,uy,ak,order);
          break;
      }
      
      res = 0.; n1 = 1-n2;
      if (xi+n1>=0 && xi+n2<width && yi+n1>=0 && yi+n2<height) {
        
        int adr = yi*width+xi; 
        for (int dy=n1;dy<=n2;dy++) 
          for (int dx=n1;dx<=n2;dx++) 
            res += cy[n2-dy]*cx[n2-dx]*ref[adr+width*dy+dx];
      } else 
        
        for (int dy=n1;dy<=n2;dy++)
          for (int dx=n1;dx<=n2;dx++) 
            res += cy[n2-dy]*cx[n2-dx]*vMW(ref,xi+dx,yi+dy,bg,width,height);
      
    }
    
    
  }	
  
  
  return res;
  
  
}


////////////////////////////////////////////////////////////////////////////////
//! End Spline Interpolation
////////////////////////////////////////////////////////////////////////////////


