/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * Code inspired in Michael Chen Matlab code 
 * http://www.mathworks.com/matlabcentral/fileexchange/26184-em-algorithm-for-gaussian-mixture-model

 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <math.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#include "../../../auxlibs/randmt/randmt.h"

#include "gaussianmix.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846264338327
#endif

//Parameters of the method:
double tolLLH=1.0e-6; //Stop criterium: tolerance of relative difference in 
                      //log-likehood between current and previous iterations 
                      //of the algorithm
int itmax=500; //maximum number of iterations

//Global variables:
//Input:
const double **data; //input data
int ndata; //number of input data
int dim; //dimension of input vectors
int ngauss; //number of Gaussians
//Output:
int *labels; //labels of each data vector 
             //(identifier of the Gaussian it belongs to)
double **mean; //mean vectors of the Gaussians
double ***covar; //covariance matrices of the Gaussians
//Auxiliary:
double *probn; //probability of the n-th data point
double *probk; //probability of the k-th Gaussian
double *weightk; //weight of the k-th Gaussian
double **probnk; //probability of the n-th data point 
                 //belonging to the k-th Gaussian
                 //(conditional probability, probability of k-th Gaussian, 
                 //given the n-th data point)
double **logprobnk; //logarithm of probnk
double *logprobn; //logarithm of probn
unsigned char *used_labels; //used_labels[k]=1 if for some data point 
                            //labels[n]=k, else used_labels[k]=0
//optimal implementation if parallelism is used (OMP flag turned on)
double ***L; //stores Cholesky decomposition of covariance matrices
double **u, **v; //vectors of size dim



//Allocate global variables
void Allocate()
{
    probn=new double[ndata];
    probk=new double[ngauss];
    weightk=new double[ngauss];
    probnk=new double*[ndata];
    for (int n=0; n < ndata; n++) probnk[n]=new double[ngauss];
    logprobnk=new double*[ndata];
    for (int n=0; n < ndata; n++) logprobnk[n]=new double[ngauss];
    logprobn=new double[ndata];
    labels=new int[ndata];
    mean=new double*[ngauss];
    for (int k=0; k < ngauss; k++) mean[k]=new double[dim];
    covar=new double**[ngauss];
    for (int k=0; k < ngauss; k++) {
        covar[k]=new double*[dim];
        for (int i=0; i < dim; i++) covar[k][i]=new double[dim];
    }
    used_labels=new unsigned char[ngauss];
    L=new double**[ngauss];
    u=new double*[ndata];
    v=new double*[ndata];
    for (int n=0; n < ngauss; n++) {
        L[n]=new double*[dim];
        for (int i=0; i < dim; i++) L[n][i]=new double[dim];
    }
    for (int n=0; n < ndata; n++) {
        u[n]=new double[dim];
        v[n]=new double[dim];
    }

}

void ResetAllocate(int ngaussNew)
{
    //create auxiliary data structures
    double *probk_aux, **probnk_aux, **mean_aux, ***covar_aux;
    probk_aux=new double[ngaussNew];
    probnk_aux=new double*[ndata];
    for (int n=0; n < ndata; n++) probnk_aux[n]=new double[ngaussNew];
    mean_aux=new double*[ngaussNew];
    for (int k=0; k < ngaussNew; k++) mean_aux[k]=new double[dim];
    covar_aux=new double**[ngaussNew];
    for (int k=0; k < ngaussNew; k++) {
        covar_aux[k]=new double*[dim];
        for (int i=0; i < dim; i++) covar_aux[k][i]=new double[dim];
    }
    
    //copy from original data structures
    int knew=0;
    for (int k=0; k < ngauss; k++) {
        if (used_labels[k]) {
            probk_aux[knew]=probk[k];
            memcpy(mean_aux[knew], mean[k], dim*sizeof(double));
            for (int i=0; i < dim; i++) memcpy(covar_aux[knew][i], 
                                               covar[k][i], dim*sizeof(double));
            knew++;
        }
    }
    for (int n=0; n < ndata; n++) {
        knew=0;
        for (int k=0; k < ngauss; k++) {
            if (used_labels[k]) {
                probnk_aux[n][knew]=probnk[n][k];
                knew++;
            }
        }
    }
        
    //delete old data structures
    delete[] probk;
    delete[] weightk;
    for (int n=0; n < ndata; n++) delete[] probnk[n];
    delete[] probnk;
    for (int k=0; k < ngauss; k++) delete[] mean[k];
    delete[] mean;
    for (int k=0; k < ngauss; k++) {
        for (int i=0; i < dim; i++)  delete[] covar[k][i];
        delete[] covar[k];
    }
    delete[] covar;
    for (int n=0; n < ndata; n++) delete[] logprobnk[n];
    delete[] used_labels;
         
    //reassign pointers
    ngauss=ngaussNew;
    weightk=new double[ngauss];
    for (int n=0; n < ndata; n++) logprobnk[n]=new double[ngauss];
    used_labels=new unsigned char[ngauss];
    probk=probk_aux;
    probnk=probnk_aux;
    mean=mean_aux;
    covar=covar_aux;

    //delete auxiliary data structures
    delete[] probk_aux;
    for (int n=0; n < ndata; n++) delete[] probnk_aux[n];
    delete[] probnk_aux;
    for (int k=0; k < ngauss; k++) delete[] mean_aux[k];
    delete[] mean_aux;
    for (int k=0; k < ngauss; k++) {
        for (int i=0; i < dim; i++)  delete[] covar_aux[k][i];
        delete[] covar_aux[k];
    }
    delete[] covar_aux;

}

//Deallocate auxiliary variables
void DeAllocate()
{
    delete[] probn;
    delete[] weightk;
    for (int n=0; n < ndata; n++) delete[] probnk[n];
    delete[] probnk;
    for (int n=0; n < ndata; n++) delete[] logprobnk[n];
    delete[] logprobnk;
    delete[] logprobn;
    delete[] used_labels;
    /*for (int i=0; i < dim; i++) delete[] L[i];
    delete[] L;
    delete[] u;
    delete[] v;*/
    for (int n=0; n < ngauss; n++) {
        for (int i=0; i < dim; i++) delete[] L[n][i];
        delete[] L[n];
    }
    for (int n=0; n < ndata; n++) {
        //for (int i=0; i < dim; i++) delete[] L[n][i];
        //delete[] L[n];
        delete[] u[n];
        delete[] v[n];
    }
    delete[] L;
    delete[] u;
    delete[] v;
}

//squared distance between two size-dimensional points
double distance2_points(double *a, double *b, int size)
{
    double dist2=0;
    
    for (int n=0; n < size; n++) dist2+=((a[n]-b[n])*(a[n]-b[n]));
    return dist2;
}

//randomly choose ngauss points from the input data
//assign each point of the data set to the closest chosen point
void Initialization(void)
{
    //initial means of Gaussians: random samples of the initial data
    printf("Random initial means\n");
    //mt_init_auto();
    init_randmt_auto();
    int iddata;
    double a;
    unsigned char repeated;
    for (int k=0; k < ngauss; k++) {
        do {
            //a=mt_drand53(); //random number between 0 and 1
            a=rand_unif(); //random number between 0 and 1
            iddata=(int) (a*(double) (ndata-1)); //random integer number 
                                                 //between 0 and ndata-1
            printf("iddata=%i  (range:0-%i)\n", iddata, ndata-1);
            memcpy(mean[k], data[iddata], dim*sizeof(double));
            //check for repeated values
            repeated=0;
            for (int kk=0; (kk < k) && !repeated; kk++) 
                repeated=(distance2_points(mean[k], mean[kk], dim) == 0);
        } while (repeated);
    }

    //label the data set: assign the closest Gaussian to each point
    //for each point compute the distances to all the Gaussian centers (means)
    //chose the Gaussian with smaller distance
    printf("Label data set\n");
    double dist, distmin;
    int idmin;
    for (int n=0; n < ndata; n++) {
        idmin=0;
        distmin=distance2_points((double *) data[n], mean[0], dim);
        for (int k=1; k < ngauss; k++) {
            dist=distance2_points((double *) data[n], mean[k], dim);
            if (dist < distmin) {
                distmin=dist;
                idmin=k;
            }
        }
        labels[n]=idmin;
        //Initialize probnk
        memset(probnk[n], 0, ngauss*sizeof(double));
        probnk[n][idmin]=1;
    }
    
}

//Cholesky decomposition of square (n x n) matrix A
//Output: L, such that L.L^T = A
unsigned char CholeskyMatrix(double **A, double **L, int n)
{
    double sum;
    
    //initialization: L=A
    for (int i=0; i < n; i++) memcpy(L[i], A[i], n*sizeof(double));
    
    //compute L
    for (int i=0; i < n; i++) {
        for (int j=0; j < n; j++) {
            sum=L[i][j];
            for (int k=i-1; k >= 0; k--) sum-=(L[i][k]*L[j][k]);
            if (i == j) {
                if (sum <= 0) {
                    printf("Error in Cholesky decomposition\n");
                    return 0;
                }
                L[i][i]=sqrt(sum);
            } else L[j][i]=sum/L[i][i];
        }
    }
    for (int i=0; i < n; i++) 
        for (int j=0; j < i; j++) L[j][i]=0.0;
 
  return 1;
}

//Solve system of equations using Cholesky decomposition: L.v=u
//Output: v
void CholeskySolve(double **L, double *u, double *v, int n)
{
    double sum;
    for (int i=0; i < n; i++) {
        sum=u[i];
        for (int j=0; j < i; j++) sum-=(L[i][j]*v[j]);
        v[i]=sum/L[i][i];
    }
}

//output= -1/2 * covarterm - dim/2 * log(2pi) - 1/2 * log(det(covar))
//where: covarterm=(x[n]-mu[k])' * covar^{-1} * (x[n]-mu[k])
double loggausspdf(int n, int k)
{
    double output;
    //compute vector u=(x[n]-mu[k])
    //for (int i=0; i < dim; i++) u[i]=data[n][i]-mean[k][i];
    for (int i=0; i < dim; i++) u[n][i]=data[n][i]-mean[k][i];
    //use Cholesky decomposition to compute covarterm
    //(x-mu)' * covar^{-1} * (x-mu) = |L^{-1} * (x-mu)|^2,
    //where covar=L*L' (L: Cholesky matrix, lower triangular matrix)
    CholeskySolve(L[k], u[n], v[n], dim);
    double covarterm=0;
    for (int i=0; i < dim; i++) covarterm+=(v[n][i]*v[n][i]);
    //compute log(det(covar)) using Cholesky matrix:
    //det(covar)=det(L*L^T)=det(L)*det(L^T)=det(L)^2
    //det(L)=L[0][0]*L[1][1]*...*L[dim-1][dim-1]
    double logdetcovar=0;
    for (int i=0; i < dim; i++) logdetcovar+=log(L[k][i][i]);
    logdetcovar*=2.0;
    
    output=-0.5*(covarterm+dim*log(2*M_PI)+logdetcovar);
        
    return output;
}





//compute log(Prob(x[n])=log(sum_k (gausspdf(n|k)*prob(k)))
//where gausspdf(n|k)*prob(k) = exp(logprobnk[n][k])
//define z_nk = logprobnk[n][k]
//therefore log(Prob(x[n])=log(sum_k(exp(z_nk))=
//                        =z_max+log(sum_k(exp(z_nk-z_max)))
//where z_max=max_k {z_nk}
//more information: google 'log sum exp'
double logsumexp(int n)
{
    double output, max, sum;
    
    max=logprobnk[n][0];
    for (int k=1; k < ngauss; k++) 
        if (logprobnk[n][k] > max) max=logprobnk[n][k];
    sum=0;
    for (int k=0; k < ngauss; k++) sum+=exp(logprobnk[n][k]-max);
    output=max+log(sum);
    
    return output;
}

//Maximization step of EM algorithm
void maximization(void)
{
    //compute weight and probability of each Gaussian
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int k=0; k < ngauss; k++) {
        weightk[k]=0;
        for (int n=0; n < ndata; n++) weightk[k]+=probnk[n][k];
        probk[k]=weightk[k]/(double) ndata;
        if (weightk[k] == 0) {
            fprintf(stdout, "ERROR: Gaussian weight is 0\n");
        }
    }
    
    //compute mean of each Gaussian
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int k=0; k < ngauss; k++) {
        memset(mean[k], 0, dim*sizeof(double));
        for (int n=0; n < ndata; n++) 
            for (int i=0; i < dim; i++) mean[k][i]+=(probnk[n][k]*data[n][i]);
        for (int i=0; i < dim; i++) mean[k][i]/=weightk[k];
    }
    
    //compute covariance matrix of each Gaussian
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int k=0; k < ngauss; k++) {
        for (int i=0; i < dim; i++) memset(covar[k][i], 0, dim*sizeof(double));
        for (int n=0; n < ndata; n++) 
            for (int i=0; i < dim; i++)
                for (int j=0; j < dim; j++) 
                    covar[k][i][j]+=(probnk[n][k]*(data[n][i]-mean[k][i])* \
                                    (data[n][j]-mean[k][j]));
        for (int i=0; i < dim; i++)
            for (int j=0; j < dim; j++) 
                covar[k][i][j]/=weightk[k];
        
        //add a prior for numerical stability
        for (int i=0; i < dim; i++) covar[k][i][i]+=1.0e-6;
    }
    
}


//Expectation step of EM algorithm
double expectation(void)
{
    double LLH;

    //Cholesky decomposition of covariance matrices
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int k=0; k < ngauss; k++) {
      CholeskyMatrix(covar[k], L[k], dim); 
    }
  
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int n=0; n < ndata; n++) {
        for (int k=0; k < ngauss; k++) {
            logprobnk[n][k]=loggausspdf(n, k);
            logprobnk[n][k]+=log(probk[k]);
        }
        logprobn[n]=logsumexp(n);
        for (int k=0; k < ngauss; k++) logprobnk[n][k]-=logprobn[n];
    }

    
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int n=0; n < ndata; n++) 
        for (int k=0; k < ngauss; k++) 
            probnk[n][k]=exp(logprobnk[n][k]);

    
    LLH=0;
    for (int n=0; n < ndata; n++) LLH+=logprobn[n];
    LLH/=(double) ndata;
    
    return LLH;
}

//Compute Gaussian Mixture (GM) for a set of vectors
//Input:
//data[i][j]=element j of i-th data vector
//data[i]=data vector i (dimension=dim)
//ndata=number of data vectors
//dim=dimension of each data vector (number of components)
//ngauss=number of Gaussians in the mixture
//Output:
//ngauss: final number of Gaussians
//labels: 
// labels[i]=label of vector i (identifier of the Gaussian it belongs to)
//means:
// mean[k]=mean vector (dim components) of k-th Gaussian
//	   mean[k][i]: i-th components of mean vector of k-th Gaussian 
// covar[k]: matrix of covariances of k-th Gaussian
//	   covar[k][i][j]: i-th components of variance vector of k-th Gaussian 
// prob[k]: probability of the k-th Gaussian
void GaussianMixture(const double **dataIn, const int ndataIn, 
                     const int dimIn, const int ngaussIn,
		             int &ngaussOut, int **labelsOut, 
		             double ***meanOut, double ****covarOut,
                     double **probOut)
{
    //Initialize global variables
    data=dataIn;
    ndata=ndataIn;
    dim=dimIn;
    ngauss=ngaussIn;

    //allocate memory
    Allocate();

    //INITIALIZATION: initialize means and labels of each Gaussian
    printf("    Initialization\n");
    Initialization();
    
    //ITERATE:
    unsigned char converged=0;
    int it=0;
    double LLH, LLHprev=0;
    double maxprobnk;
    int idmaxprobnk;
    int nused;
    do {
        printf("    iter %i\n", it);
        //MAXIMIZATION: estimate mean and covariances
        printf("        Maximization\n");
        maximization();
        //EXPECTATION: estimate likelihoods
        printf("        Expectation\n");
        LLH=expectation();
  
        printf("        LLH=%2.2f\n", (float) LLH);

        printf("        Recompute labels\n");
        //recompute labels: assign Gaussian with highest probnk
        memset(used_labels, 0, ngauss);
        for (int n=0; n < ndata; n++) {
            maxprobnk=probnk[n][0];
            idmaxprobnk=0;
            for (int k=1; k < ngauss; k++) {
                if (probnk[n][k] > maxprobnk) {
                    maxprobnk=probnk[n][k];
                    idmaxprobnk=k;
                }
            }
            labels[n]=idmaxprobnk;
            used_labels[idmaxprobnk]=1;
        }
        //check number of assigned Gaussians
        nused=0;
        for (int k=0; k < ngauss; k++) if (used_labels[k]) nused++;
        
        if (nused < ngauss) {
            fprintf(stdout, "WARNING: Reducing the number of \
                    Gaussians to %i\n", nused);
            ResetAllocate(nused);
        }
        
        if (it > 0) {
            converged=(LLH-LLHprev < tolLLH*fabs(LLH));
        }
        LLHprev=LLH;
        it++;
    } while ((it < itmax) && (!converged));
    
    if (converged) fprintf(stdout, "Converged in %i iterations\n", it);
    else fprintf(stdout, "Not converged in %i iterations\n", it);
    
    fprintf(stdout, "Final number of Gaussians: %i\n", ngauss);
    
    //OUTPUT
    ngaussOut=ngauss;
    *labelsOut=labels;
    *meanOut=mean;
    *covarOut=covar;
    *probOut=probk;

    //clean up
    DeAllocate();
}

//Compute probnk=prob(gaussian k | given vector n)
//probnk=prob(k|n)=gausspdf(n, k)*prob(k)/prob(x(n))
//where prob(x(n))=sum_k (gausspdf(n, k)*prob(k))
double **probnk_GaussianMixture(const double **dataIn, 
                                const int ndataIn, const int dimIn, 
                                const int ngaussIn,
                                double **meanIn, double ***covarIn, 
                                double *probIn)
{
    double **probnkOut;
    
    //Initialize global variables
    data=dataIn;
    ndata=ndataIn;
    dim=dimIn;
    ngauss=ngaussIn;
    mean=meanIn;
    covar=covarIn;
    probk=probIn;
    
    //allocate memory
    probnkOut=new double*[ndata];
    for (int n=0; n < ndata; n++) probnkOut[n]=new double[ngauss];
    
    probn=new double[ndata];
    probnk=probnkOut;
    logprobnk=new double*[ndata];
    for (int n=0; n < ndata; n++) logprobnk[n]=new double[ngauss];
    logprobn=new double[ndata];

    L=new double**[ngauss];
    u=new double*[ndata]; 
    v=new double*[ndata];
    for (int n=0; n < ngauss; n++) {
        L[n]=new double*[dim];
        for (int i=0; i < dim; i++) L[n][i]=new double[dim];
    }
    for (int n=0; n < ndata; n++) {
        u[n]=new double[dim];
        v[n]=new double[dim];
    }
    
    //Cholesky decomposition of covariance matrices
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int k=0; k < ngauss; k++) {
      CholeskyMatrix(covar[k], L[k], dim);
    }
    
    //compute probnk=prob(k|n)=gausspdf(n, k)*prob(k)/prob(x(n))
    //where prob(x(n))=sum_k (gausspdf(n, k)*prob(k))
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int n=0; n < ndata; n++) {
        for (int k=0; k < ngauss; k++) {
            logprobnk[n][k]=loggausspdf(n, k);
            logprobnk[n][k]+=log(probk[k]);
        }
        logprobn[n]=logsumexp(n);
        for (int k=0; k < ngauss; k++) logprobnk[n][k]-=logprobn[n];
    }  
    
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int n=0; n < ndata; n++) 
        for (int k=0; k < ngauss; k++) 
            probnk[n][k]=exp(logprobnk[n][k]);
    
    //deallocate memory
    delete[] probn;
    for (int n=0; n < ndata; n++) delete[] logprobnk[n];
    delete[] logprobnk;
    delete[] logprobn;
    for (int n=0; n < ngauss; n++) {
        for (int i=0; i < dim; i++) delete[] L[n][i];
        delete[] L[n];
    }  
    for (int n=0; n < ndata; n++) {
        delete[] u[n];
        delete[] v[n];
    }
    delete[] L;
    delete[] u;
    delete[] v;

    return probnkOut;
}


//For each input vector, compute labels: assign Gaussian with highest probnk
//where probnk=prob(gaussian k | given vector n)
int *labelsGaussianMixture(const double **dataIn, const int ndataIn, 
                           const int dimIn, const int ngaussIn,
		     	           double **meanIn, double ***covarIn, double *probIn)
{
  int *labelsOut;
    
  //Initialize global variables
  ndata=ndataIn;
  ngauss=ngaussIn;

  labelsOut=new int[ndata];

  //compute probnk
  probnk=probnk_GaussianMixture(dataIn, ndataIn, dimIn, ngaussIn, 
                                meanIn, covarIn, probIn);

  //compute labels: assign Gaussian with highest probnk
  double maxprobnk;
  int idmaxprobnk; 

  for (int n=0; n < ndata; n++) {
    maxprobnk=probnk[n][0];
    idmaxprobnk=0;
    for (int k=1; k < ngauss; k++) {
      if (probnk[n][k] > maxprobnk) {
        maxprobnk=probnk[n][k];
        idmaxprobnk=k;
      }
    }
    labelsOut[n]=idmaxprobnk;
  }

  //deallocate memory
  for (int n=0; n < ndata; n++) delete[] probnk[n];
  delete[] probnk;

  return labelsOut;
}



