/*
 * Copyright (c) 2011 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "gradients_lib.h"

//Compute gradients of an image 
//Methods for gradient computation:
//Sobel: classical Sobel operator
//Scharr: similar to Sobel, but with better rotational symmetry 
//       (see wikipedia: https://en.wikipedia.org/wiki/Sobel_operator)
//Bilinear: gradient of bilinear interpolate of 4 pixels, gradient at center 
//          of the 4 pixels
template<class TYPE>
void compute_gradient(float **Gx_out, float **Gy_out, TYPE *im, int w, int h,
                      const char *method)
{
  int i, j;
  float *Gx, *Gy;
  
  if (*Gx_out == NULL) {
    Gx=new float[w*h]; 
  } else Gx=*Gx_out;
  if (*Gy_out == NULL) {
    Gy=new float[w*h]; 
  } else Gy=*Gy_out;

  if (strcmp(method, "bilinear")) { //Sobel or Scharr
    float im1, im2, im3, im4, im6, im7, im8, im9;
    //gradient zero at the image borders
    memset(Gx, 0, w*h*sizeof(float));
    memset(Gy, 0, w*h*sizeof(float));
    for (j=1; j < h-1; j++)
      for (i=1; i < w-1; i++) {
        im1=(float) im[i-1+(j-1)*w];
        im2=(float) im[i+(j-1)*w];
        im3=(float) im[i+1+(j-1)*w];
        im4=(float) im[i-1+j*w];
        im6=(float) im[i+1+j*w];
        im7=(float) im[i-1+(j+1)*w];
        im8=(float) im[i+(j+1)*w];
        im9=(float) im[i+1+(j+1)*w];
        if (!strcmp(method, "sobel")) {
          Gx[i+j*w]=((im3+2*im6+im9)-(im1+2*im4+im7))/4.0;
          Gy[i+j*w]=((im7+2*im8+im9)-(im1+2*im2+im3))/4.0;
        } else {
        //Scharr operator: better rotational symmetry (see wikipedia)
          Gx[i+j*w]=((3*im3+10*im6+3*im9)-(3*im1+10*im4+3*im7))/16.0;
          Gy[i+j*w]=((3*im7+10*im8+3*im9)-(3*im1+10*im2+3*im3))/16.0;
        }
      }
  } else { //bilinear
    float im1, im2, im3, im4;
    for (j=0; j < h-1; j++)
      for (i=0; i < w-1; i++) {
        im1=(float) im[i+j*w];
        im2=(float) im[i+1+j*w];
        im3=(float) im[i+(j+1)*w];
        im4=(float) im[i+1+(j+1)*w];
        Gx[i+j*w]=((im2+im4)-(im1+im3))/2.0f;
        Gy[i+j*w]=((im3+im4)-(im1+im2))/2.0f;
      }
    //duplicate gradient in the last row and column
    for (i=0; i < w; i++) {
      Gx[i+(h-1)*w]=Gx[i+(h-2)*w];
      Gy[i+(h-1)*w]=Gy[i+(h-2)*w];
    }
    for (j=0; j < h; j++) {
      Gx[w-1+j*w]=Gx[w-2+j*w];
      Gy[w-1+j*w]=Gy[w-2+j*w];
    }
  }
    
  if (*Gx_out == NULL) *Gx_out=Gx;
  if (*Gy_out == NULL) *Gy_out=Gy;    
  
}



//Compute gradients of an image using bilinear method
void compute_gradient(float **Gx_out, float **Gy_out, unsigned char *im, int w, int h)
{
   //default: bilinear
   compute_gradient(Gx_out, Gy_out, im, w, h, "bilinear");
}



//returns a (2r+1)x(2r+1) square grid 
//sampling a 2D Gaussian function
//The discrete Gaussian is obtained by taking (2r+1)x(2r+1) samples
//in the range (-3sigma, 3sigma)x(-3sigma, 3sigma)
//The samples are taken at positions 
//(x_k, y_l)= (k * (3sigma)/(r+1), l * (3sigma)/(r+1))
//with k and l in the set-r, -r+1, ...-2, -1, 0, 1, 2, ..., r-1, r 
//The distance between samples is 3sigma/(r+1)
//Since the distance between image values (pixels) is 1 pixel
//then 3sigma/(r+1) = 1 pixel, therefore sigma=(r+1)/3 pixels
//The set of samples is normalized such that sum(samples)=1
//if flag isdisk is active set to zero the values in the grid
//whose distance to the center is above r
double *gaussiankernel2D(int r, unsigned char isdisk)
{
  double sigma=(float) (r+1)/3;
  double *gauss=new double[(2*r+1)*(2*r+1)];
  //center of the squared grid
  double invden=1/(2.0f*sigma*sigma);
  double sum=0;
  int n=0;
  for (int y=-r; y <= r; y++) 
    for (int x=-r; x <= r; x++) {
      gauss[n]=exp(-(x*x+y*y)*invden);
      if (isdisk && (x*x+y*y > r*r)) gauss[n]=0;
      sum+=gauss[n];
      n++;
    }
  //normalize
  for (int n=0; n < (2*r+1)*(2*r+1); n++) gauss[n]/=sum;
  
  /*for (int j=0; j <= 2*r; j++) {
    for (int i=0; i <= 2*r; i++) {
      printf("%2.2f  ", gauss[i+j*(2*r+1)]);
    }
    printf("\n");
  }*/
  
  return gauss;
}


//Compute structure tensor MS (2 x 2 second moment matrix)
//associated to pixel (i0, j0)
//- Define for each pixel (i, j) a 2x2 matrix M(i, j)=G (tensor product) G=(Gx^2 GxGy; GyGx Gy^2)
//- MS=sum_{(i,j) in neighborhood of (i0, j0)} M(i, j)*weight(i, j)
//Output: MS
unsigned char get_structure_tensor(int i0, int j0, int r, 
                                   struct gradient_info *gdata, int w, int h,
                                   double *weights, double M[2][2])
{
  double ggx, ggy, weight=1;
  M[0][0]=0; M[0][1]=0;
  M[1][0]=0; M[1][1]=0;
  if (!weights) weight=1/((2*r+1)*(2*r+1)); //use uniform square weights
  int m=0;
  for (int j=j0-r; j <= j0+r; j++)
    for (int i=i0-r; i <= i0+r; i++) {
      if ((i < 0) || (i >= w) || (j < 0) || (j >= h)) return 0;
      ggx=(double) gdata[i+j*w].gx;
      ggy=(double) gdata[i+j*w].gy;
      if (weights) {
        M[0][0]+=ggx*ggx*weights[m];
        M[0][1]+=ggx*ggy*weights[m];
        M[1][1]+=ggy*ggy*weights[m];
      } else {
        M[0][0]+=ggx*ggx*weight;
        M[0][1]+=ggx*ggy*weight;
        M[1][1]+=ggy*ggy*weight;        
      }
      m++;
    }
  M[1][0]=M[0][1];
  
  return 1;
}


//Compute smooth gradient values
//Gradient vector computed from the neighborhood (radius: rad) of a pixel 
//given an initial estimate of the gradients
//Method:
//- Each pixel of the neighborhood has a gradient G, computed using Sobel or 
//  Scharr operators or bilinear interpolation
//  (function compute_gradient)
//- Associate to each pixel (i, j) a 2x2 matrix M(i, j)= G (tensor product) G =
//                                                     = (Gx^2 GxGy; GyGx Gy^2)
//- MS=sum_{for all pixels in neighborhood} M(i, j)
//- Diagonalize MS
//- The eigenvector associated to the largest eigenvalue of MS is an estimate of 
//  the gradient vector of the neighborhood
//- The largest eigenvalue of MS is equal to the squared magnitude of this 
//  gradient vector
// methodGrad='sobel', 'scharr' or 'bilinear'
struct gradient_info *compute_gradient_smooth(struct gradient_info *gdata0, 
                                              unsigned char *im, int w, int h, 
                                              int r)
{  
  
  struct gradient_info *gdata=new struct gradient_info[w*h];

  float r2=(float) r * (float) r;
  
  //Gaussian smoothing window ((2r+1) x (2r+1)), (disk)
  double *weights=gaussiankernel2D(r, 1);
  
  //compute elements of autocorrelation matrix M=(A C, C B) and Cornerness
  double M[2][2];
  for (int j0=r; j0 < h-r; j0++) {
	for (int i0=r; i0 < w-r; i0++) {
      gdata[i0+j0*w].x=gdata0[i0+j0*w].x;
      gdata[i0+j0*w].y=gdata0[i0+j0*w].y;
      get_structure_tensor(i0, j0, r, gdata0, w, h, weights, M);
	  //diagonalize
	  double trM, detM, disc, lambda, ux, uy, modu, modG;
	  trM=M[0][0]+M[1][1];
	  detM=M[0][0]*M[1][1]-M[1][0]*M[0][1];
	  if (trM == 0) {
	    gdata[i0+j0*w].gx=0;
        gdata[i0+j0*w].gy=0;
	  } else {
	    disc=trM*trM-4*detM;
	    if (disc < 0) {
		  printf("Warning: unexpected value in gradients matrix\n");
	      gdata[i0+j0*w].gx=0;
		  gdata[i0+j0*w].gy=0;
		} else {
		  lambda=(trM+sqrt(disc))/2; //largest eigenvalue
		  if (M[0][1] != 0) {
			if (M[0][0] > M[1][1]) { //if M[0][1] is close to 0, then u_normalized is close to (1, 0)
              ux=lambda-M[1][1];
              uy=M[0][1]; 
            } else { //if M[0][1] is close to 0, then u_normalized is close to (0, 1)
              ux=M[0][1];
              uy=lambda-M[0][0]; 
            }
		  } else {
			if (M[0][0] > M[1][1]) {
			  ux=1;
			  uy=0;
			} else {
			  ux=0;
			  uy=1;
			}
		  }
		  //normalize u
		  modu=sqrt(ux*ux+uy*uy);
		  ux/=modu;
		  uy/=modu;
          //assign orientation: gradient in the direction of increasing values
          double sumUp=0, sumDown=0, eprod;
          for (int j=j0-r; j <= j0+r; j++)
            for (int i=i0-r; i <= i0+r; i++) 
              if ((i-i0)*(i-i0)+(j-j0)*(j-j0) <= r2) { //inside circle with center (i0, j0) and radius rad
                //scalar product with gradient vector
                //if > 0, means the pixel is above the edge at (i0, j0)
                //if < 0, means the pixel is below the edge at (i0, j0)
                eprod=((double) (i-i0))*ux+((double) (j-j0))*uy;
                //remark that if bilinear gradient is used we should use 
                //image values at position im[i+0.5+(j+0.5)*w]
                //we don't do it for simplicity and because we don't think 
                //the result won't change much
                if (eprod > 0) sumUp+=im[i+j*w];
                if (eprod < 0) sumDown+=im[i+j*w];
              }
          if (sumUp < sumDown) { //take oposite orientation
            ux=-ux;
            uy=-uy;
          }
		  //get gradient vector
		  modG=sqrt(lambda);
		  gdata[i0+j0*w].gx=ux*modG;
		  gdata[i0+j0*w].gy=uy*modG;
		}
	  }
	}
  }

  //extend values by symmetry
  for (int i=r; i < w-r; i++) 
	for (int j=0; j < r; j++) {
	  gdata[i+(r-1-j)*w].gx=gdata[i+(j+r)*w].gx;
	  gdata[i+(r-1-j)*w].gy=gdata[i+(j+r)*w].gy;
	  gdata[i+(h-r+j)*w].gx=gdata[i+(h-r-1-j)*w].gx;
	  gdata[i+(h-r+j)*w].gy=gdata[i+(h-r-1-j)*w].gy;
	}
  
  for (int j=r; j < h-r; j++) 
	for (int i=0; i < r; i++) {
	  gdata[(r-1-i)+j*w].gx=gdata[(i+r)+j*w].gx;
	  gdata[(r-1-i)+j*w].gy=gdata[(i+r)+j*w].gy;
	  gdata[(w-r+i)+j*w].gx=gdata[(w-r-1-i)+j*w].gx;
	  gdata[(w-r+i)+j*w].gy=gdata[(w-r-1-i)+j*w].gy;
	}
  
  for (int i=0; i < r; i++)
	for (int j=0; j < r; j++) {
	  gdata[(r-1-i)+j*w].gx=gdata[(i+r)+j*w].gx;
	  gdata[(r-1-i)+j*w].gy=gdata[(i+r)+j*w].gy;
	  gdata[(r-1-i)+(h-r+j)*w].gx=gdata[(i+r)+(h-r+j)*w].gx;
	  gdata[(r-1-i)+(h-r+j)*w].gy=gdata[(i+r)+(h-r+j)*w].gy;
	  gdata[(w-r+i)+j*w].gx=gdata[(w-r-1-i)+j*w].gx;
	  gdata[(w-r+i)+j*w].gy=gdata[(w-r-1-i)+j*w].gy;
	  gdata[(w-r+i)+(h-r+j)*w].gx=gdata[(w-r-1-i)+(h-r+j)*w].gx;
	  gdata[(w-r+i)+(h-r+j)*w].gy=gdata[(w-r-1-i)+(h-r+j)*w].gy;
	}
  
  for (int n=0; n < w*h; n++) {
    gdata[n].gmod2=gdata[n].gx*gdata[n].gx+gdata[n].gy*gdata[n].gy;
    gdata[n].x=gdata0[n].x;
    gdata[n].y=gdata0[n].y;
  }
  
  delete[] weights;
  return gdata;
}


//Compute gradients of an image and store in data structure
//Gradient computation methods: "scharr", "sobel", "bilinear"
struct gradient_info *get_gradient_info(unsigned char *im, int w, int h, 
                                        const char *methodGrad)
{
  struct gradient_info *gdata=new struct gradient_info[w*h];

  float *Gx=NULL, *Gy=NULL;
  compute_gradient(&Gx, &Gy, im, w, h, methodGrad);

  for (int n=0; n < w*h; n++) {
	gdata[n].x=n%w;
	gdata[n].y=n/w;
	gdata[n].gx=Gx[n];
	gdata[n].gy=Gy[n];
	gdata[n].gmod2=Gx[n]*Gx[n]+Gy[n]*Gy[n];
  }

  delete[] Gx;
  delete[] Gy;
  return gdata;
}



//Obtain Canny extrema of an image
//extrema in the direction of the gradient
struct gradient_info *get_canny_extrema(struct gradient_info *gdata, int w, int h, float th, int &nextrema)
{
  struct gradient_info *gext=NULL;

  unsigned char *extrema=new unsigned char[w*h];
  memset(extrema, 0, w*h);
  int n;
  nextrema=0;
  float gref, gprev, gnext;
  float th2=th*th;
  double vx, vy, v, x0, y0, xnext, ynext, xprev, yprev;
  int iinext, jjnext, iiprev, jjprev;
  for (int y=1; y < h-1; y++) {
	for (int x=1; x < w-1; x++) {
	  n=x+y*w;
	  gref=gdata[n].gmod2;
	  if (gref >= th2) {
	    v=(double) sqrt((double) gref);
	    vx=gdata[n].gx/v;
	    vy=gdata[n].gy/v;
		x0=(double) gdata[n].x;
		y0=(double) gdata[n].y;
		xnext=x0+vx;
		ynext=y0+vy;
		xprev=x0-vx;
		yprev=y0-vy;
        iinext=nearest(xnext);
        jjnext=nearest(ynext);
        iiprev=nearest(xprev);
        jjprev=nearest(yprev);
		if ((iinext >= 0) && (iinext < w) && (jjnext >= 0) && (jjnext < h) &&
			(iiprev >= 0) && (iiprev < w) && (jjprev >= 0) && (jjprev < h)) {
		  gnext=gdata[iinext+jjnext*w].gmod2;
		  gprev=gdata[iiprev+jjprev*w].gmod2;
		  if ((gref > gprev) && (gref > gnext)) { 
		    extrema[n]=1;	  
		    nextrema++;
		  }
		}
	  }
	}
  }

  gext=new struct gradient_info[nextrema];
  int next=0;
  for (int n=0; n < w*h; n++) {
	if (extrema[n]) memcpy(&(gext[next++]), &(gdata[n]), sizeof(struct gradient_info)); 
  }

  delete[] extrema;
  return gext;
}

//Obtain Harris extrema of an image
//implemented as described in: 
// http://kiwi.cs.dal.ca/~dparks/CornerDetection/harris.htm
//use 3x3 Gaussian window
struct gradient_info *get_Harris_extrema(struct gradient_info *gdata, int w, int h, 
											float thk, float thT, int &nextrema)
{
  struct gradient_info *gext=NULL;

  //find local extrema 
  //gradient[n]> max{gradient[n-1], gradient[n+1], gradient[n-w], gradient[n+w]}
  unsigned char *extrema=new unsigned char[w*h];
  memset(extrema, 0, w*h);
  int n;
  nextrema=0;

  
  //allocate memory
  double A, B, C;
  double *Cornerness=new double[w*h];
  memset(Cornerness, 0, w*h*sizeof(double));

  //Gaussian smoothing window (3 x 3), (disk)
  double *gauss=gaussiankernel2D(1, 1);

  //compute elements of autocorrelation matrix M=(A C, C B) and Cornerness
  double M[2][2];
  for (int y=1; y < h-1; y++) {
	for (int x=1; x < w-1; x++) {
      get_structure_tensor(x, y, 1, gdata, w, h, gauss, M);
      A=M[0][0];
      B=M[1][1];
      C=M[0][1];
	  Cornerness[x+y*w]=(A*B-C*C)-thk*(A+B);
	}
  }
  delete[] gauss;
  
  //find maximum of cornerness
  double cmax=Cornerness[0];
  for (int n=1; n < w*h; n++) if (Cornerness[n] > cmax) cmax=Cornerness[n];
  
  //set to 0 low values of cornerness
  for (int n=0; n < w*h; n++) if (Cornerness[n] < thT*cmax) Cornerness[n]=0;

  //select local maxima (use 4-connectivity)
  double gref, gprev, gnext, gup, gdown;
  for (int y=1; y < h-1; y++) {
	for (int x=1; x < w-1; x++) {
	  n=x+y*w;
	  gref=Cornerness[n];
	  gnext=Cornerness[n+1];
	  gprev=Cornerness[n-1];
	  gup=Cornerness[n-w];
	  gdown=Cornerness[n+w];
	  if ((gref > 0) && (gref > gprev) && (gref > gnext) && 
	      (gref > gup) && (gref > gdown)) {
		extrema[n]=1;	  
		nextrema++;
	  }
	}
  }

  gext=new struct gradient_info[nextrema];
  int next=0;
  for (int n=0; n < w*h; n++) {
	if (extrema[n]) memcpy(&(gext[next++]), &(gdata[n]), 
	                       sizeof(struct gradient_info)); 
  }

    
  //clean up
  delete[] Cornerness;

  return gext;
}








