/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "../../../auxlibs/Eigen/Dense"
using namespace Eigen;

#include "../gaussianmix/gaussianmix.h"
#include "patches_analysis_lib.h"



//Principal Component Analysis      
//
// Perform PCA given a covariance matrix. 
// Uses Eigen template library to compute eigenvalues and
// eigenvectors of the covariance matrix 
//Input: C covariance matrix
//Output: vd variances associated to the principal components,
//        PCs principal components
void compute_pca(double **C, double *vd, double **PCs, int dim)
{
  MatrixXd CC(dim, dim);
  for (int i=0; i < dim; i++)
    for (int j=0; j < dim; j++) 
	  CC(i, j)=C[i][j];

  //compute eigenvalues and eigenvectors

  //using EigenSolver
  EigenSolver<MatrixXd> diag(CC);
  VectorXcd lambda = diag.eigenvalues();
  MatrixXcd P = diag.eigenvectors();

  //sort eigenvalues and eigenvectors 
  //and fill output data structures
  VectorXd real_lambda;
  real_lambda=lambda.array().real();

  MatrixXd real_P(dim, dim);
  real_P=P.array().real();

  int *position, pos;
  double *ev, max;
  position=new int[dim];
  ev=new double[dim];

  for (int k=0; k < dim; k++) ev[k]=real_lambda(k);
  for (int k=0; k < dim; k++) {
    max=ev[0];
    pos=0;
    for (int j=1; j < dim; j++) {
      if (ev[j] > max) {	  
        max=ev[j];
        pos=j;
      }
    }
    position[k]=pos;
    ev[pos]=-1;
  }
  
  for (int k=0; k < dim; k++) {
    vd[k]=real_lambda(position[k]);
	for (int m=0; m < dim; m++) {
      PCs[k][m]=real_P(m, position[k]);
	}
  }

  delete[] position;
  delete[] ev;
  return;
}

//Compute principal components of a set of patches (vectors of 'dim' dimensions) 
//The principal components are the eigenvectors of the covariance matrix of the 
//set of patches
void compute_PCA_vectors(double **vectors, int nvectors, int dim, 
						 double **meanOut, double **valPsOut, double ***vecPsOut)
{
  double **C, *mean=NULL;
  double *valPs=NULL, **vecPs=NULL;

  valPs=new double[dim];
  vecPs=new double*[dim];
  for (int k=0; k < dim; k++) vecPs[k]=new double[dim];

  //covariance matrix: 
  //computed as E((x-xm)^2) instead as E(x^2)-E(x)^2 to prevent numerical overflows
    
  mean=new double[dim];
  memset(mean, 0, dim*sizeof(double));
  C=new double*[dim];
  for (int k=0; k < dim; k++) {
	C[k]=new double[dim];
    memset(C[k], 0, dim*sizeof(double));
  }

  //compute average vector
  for (int n=0; n < nvectors; n++) {
    for (int k=0; k < dim; k++) {
      mean[k]+=vectors[n][k];
    }
  }
  for (int k=0; k < dim; k++) mean[k]/=(double) nvectors;    
    
  //compute covariance matrix
  for (int n=0; n < nvectors; n++) {
    for (int k=0; k < dim; k++) {
	  for (int m=0; m <= k; m++) {
		C[k][m]+=(vectors[n][k]-mean[k])*(vectors[n][m]-mean[m]);
	  }
    }
  }
	
  for (int k=0; k < dim; k++) {
	for (int m=0; m <= k; m++) {
	  C[k][m]/=(double) (nvectors-1);
	}
  }

  for (int k=0; k < dim; k++) {
	for (int m=k+1; m < dim; m++) {
	  C[k][m]=C[m][k];
	}
  }

  //compute egenvectors and eigenvalue of the matrix
  compute_pca(C, valPs, vecPs, dim);

  //delete covariance matrix
  for (int k=0; k < dim; k++) delete[] C[k];
  delete[] C;
    
  //check for numerical instabilities
  for (int k=0; k < dim; k++) 
    if (fabs(valPs[k]) < 1.0e-9) valPs[k]=1.0e-9; 

  //return result
  *valPsOut=valPs;
  *vecPsOut=vecPs;
  *meanOut=mean;
}

//L2-norm of a dim-dimensional vector
double vector_norm(double *v, int dim)
{
  double norm=0;

  for (int k=0; k < dim; k++) norm+=(v[k]*v[k]);

  return (sqrt(norm));
}

//Get unitary dim-dimensional vector
void normalize_vector(double *v, int dim)
{
  double norm;

  norm=vector_norm(v, dim);
  if (norm > 0) {
	for (int k=0; k < dim; k++) v[k]/=norm;
  }

}


//Perform Principal Components Analysis (PCA) of the set of patches
//The following information is obtained:
//- average of the set of patches
//- basis of the set of patches (principal components), obtained as eigenvectors
//of the covariance matrix of the set of patches
//- variance of the projections of the original patches on each
//vector of the basis (i.e. eigenvalues associated to each eigenvector)
double *PCApatches(double **patches, int npatches, int dim, int &sizeout)
{
  double *dataPCA=NULL;

  double *mean, *valPs, **vecPs;

  //compute principal components
  compute_PCA_vectors(patches, npatches, dim, &mean, &valPs, &vecPs);

  //normalize eigenvectors
  for (int k=0; k < dim; k++) normalize_vector(vecPs[k], dim);

  //store PCA information
  dataPCA=new double[dim+dim+dim*dim];
  int count=0;
  for (int k=0; k < dim; k++) dataPCA[count++]=mean[k];
  for (int k=0; k < dim; k++) dataPCA[count++]=valPs[k];
  for (int k=0; k < dim; k++)
    for (int m=0; m < dim; m++) dataPCA[count++]=vecPs[k][m];

  sizeout=dim+dim+dim*dim;


  delete[] mean;
  delete[] valPs;
  for (int k=0; k < dim; k++) delete[] vecPs[k];
  delete[] vecPs;
  return dataPCA;
}


//Create image of PCA basis
//npc patches of size wpatch x hpatch
//are displayed in a square image
//nrow patches are displayed per row
void PCAbaseview(unsigned char *Rout, unsigned char *Gout, unsigned char *Bout,
			     int wpatch, int hpatch, double *dataPCA, int dim,
			     int npc, int nrow)
{
  unsigned char gray=(dim == wpatch*hpatch);
  int wmargin, hmargin;
  wmargin=hmargin=2;
  int wout=nrow*wpatch+(nrow-1)*wmargin;

  double *vector, aa;
  double maxvector=0;
  double *maxvectors=new double[npc];
  double *minvec=new double[npc];
  double *maxvec=new double[npc];
  double maxabs, minabs;
  for (int k=0; k < npc; k++) { //for each vector of the basis
    //read vector data and fins minimum and maximum (in absolute value)
    vector=dataPCA+dim+dim+k*dim;
	maxvectors[k]=0;
	minvec[k]=vector[0];
	maxvec[k]=vector[0];
	if (k == 0) { maxabs=vector[0]; minabs=vector[0]; }
	for (int n=0; n < dim; n++) {
	  if (abs(vector[n]) > maxvector) maxvector=abs(vector[n]);
	  if (abs(vector[n]) > maxvectors[k]) maxvectors[k]=abs(vector[n]);
	  if (vector[n] > maxvec[k]) maxvec[k]=vector[n];
	  if (vector[n] < minvec[k]) minvec[k]=vector[n];
	  if (vector[n] > maxabs) maxabs=vector[n];
	  if (vector[n] < minabs) minabs=vector[n];
	}
  }

  int m;
  for (int k=0; k < npc; k++) { //for each vector of the basis
    //read vector data and display them in a square image
    //for each vector, map (min, max)  to (0, 255)  
    //if constant value, then display it as gray (127)
    vector=dataPCA+dim+dim+k*dim;
    int n=0;
	for (int jj=0; jj < hpatch; jj++) {
      for (int ii=0; ii < wpatch; ii++) {
		m=(ii+(k%nrow)*(wpatch+wmargin))+jj*wout+ \
		  (k/nrow)*(wout*(hpatch+hmargin));
	    if (gray) { //for grayscale patches
          if (maxvec[k] != minvec[k]) 
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f; 
		  n++;
		  Rout[m]=(int) (aa+0.5f);
		  Gout[m]=(int) (aa+0.5f);
		  Bout[m]=(int) (aa+0.5f);
	    } else { //for color patches
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f; 
		  n++;
		  Rout[m]=(int) (aa+0.5f);
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f;
		  n++;
		  Gout[m]=(int) (aa+0.5f);
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f;
		  n++;
		  Bout[m]=(int) (aa+0.5f);
	    }
	  }
	}
  }

  delete[] maxvectors;
  delete[] maxvec;
  delete[] minvec;
}

//Create image of average patch of set of patches
void PCAmeanview(unsigned char *Rout, unsigned char *Gout, unsigned char *Bout,
			     int wpatch, int hpatch, double *dataPCA, int dim)
{
    unsigned char gray=(dim == wpatch*hpatch);
    
    double *vector, aa;
    //average patch data stored in data structure obtained when computing
    //PCA of the set of patches
    vector=dataPCA; 
  
  
    //set minimum value (if negative) to 0
    //clip maximum value to 255
    double min=vector[0];
    for (int n=1; n < dim; n++) {
      if (vector[n] < min) min=vector[n];
    }
    
    int n=0;
    for (int m=0; m < wpatch*hpatch; m++) {
        if (gray) {
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Rout[m]=(int) (aa+0.5f);
            Gout[m]=(int) (aa+0.5f);
            Bout[m]=(int) (aa+0.5f);
        } else {
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Rout[m]=(int) (aa+0.5f);
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Gout[m]=(int) (aa+0.5f);
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Bout[m]=(int) (aa+0.5f);
        }
    }
  

        
}



//Compute Gaussian Mixture decomposition for a set of patches (vectors)
//The patches (vectors) are of dimension 'dim'
//The Gaussian Mixture is composed of 'ngauss' Gaussians
//For each Gaussian it is computed: 
//basis vectors, average vector, covariance matrix, 
//probability of the Gaussian in the mixture 
struct GMMpatches *getGMM(double **patches, int npatches, int dim, int ngauss)
{
  struct GMMpatches *gmm=NULL;


  int *labels;
  double **means, ***covar, *prob;

  //compute Gaussian mixture, using EM algorithm 
  //implemented in 'gaussianmix' library
  int ngaussOut;
  GaussianMixture((const double **) patches, npatches, dim, ngauss, ngaussOut, 
                  &labels, &means, &covar, &prob);

  //store GM information in data structure  
  ngauss=ngaussOut;
  gmm=new struct GMMpatches;
  gmm->ngaussian=ngauss;
  gmm->npatches=npatches;
  gmm->dim=dim;
  gmm->labels=labels;
  gmm->GM=new struct gaussian[gmm->ngaussian];
  for (int n=0; n < gmm->ngaussian; n++) {
    gmm->GM[n].dim=gmm->dim;
    gmm->GM[n].prob=prob[n];
	gmm->GM[n].mean=new double[gmm->dim];
	gmm->GM[n].corr=new double*[gmm->dim];
	for (int m=0; m < gmm->dim; m++) gmm->GM[n].corr[m]=new double[gmm->dim];
	gmm->GM[n].pcaVal=NULL;
	gmm->GM[n].pcaVec=NULL;
  }
  //means of each Gaussian
  for (int n=0; n < gmm->ngaussian; n++) memcpy(gmm->GM[n].mean, means[n], dim*sizeof(double));
  //variances of each Gaussian 
  for (int n=0; n < gmm->ngaussian; n++) 
	for (int m=0; m < gmm->dim; m++) memcpy(gmm->GM[n].corr[m], covar[n][m], dim*sizeof(double));
  
    
  //clean up
  for (int n=0; n < ngauss; n++) delete[] means[n];
  delete[] means;
  for (int n=0; n < ngauss; n++) {
    for (int i=0; i < dim; i++)  delete[] covar[n][i];
    delete[] covar[n];
  }
  delete[] covar;
  delete[] prob;

  return gmm;
}

//Compute basis vectors for each component in a Gaussian Mixture (GM)
//Use Principal Components Analysis (PCA) on the covariance matrix associated
//to each Gaussian
void computeGMMPCA(struct GMMpatches *gmm)
{
  for (int n=0; n < gmm->ngaussian; n++) {
	gmm->GM[n].pcaVal=new double[gmm->dim];
	gmm->GM[n].pcaVec=new double*[gmm->dim];
	for (int m=0; m < gmm->dim; m++) gmm->GM[n].pcaVec[m]=new double[gmm->dim];
	compute_pca(gmm->GM[n].corr, gmm->GM[n].pcaVal, gmm->GM[n].pcaVec, 
	            gmm->dim);
    //normalize eigenvectors
    for (int k=0; k < gmm->dim; k++) normalize_vector(gmm->GM[n].pcaVec[k], 
                                                      gmm->dim);
  }

}

//Delete information for a Gaussian component in the mixture
void delete_gaussian(struct gaussian *g)
{
  if (g->mean) delete[] g->mean;
  for (int n=0; n < g->dim; n++) delete[] g->corr[n];
  delete[] g->corr;
  if (g->pcaVal) delete[] g->pcaVal;
  if (g->pcaVec) {
    for (int n=0; n < g->dim; n++) delete[] g->pcaVec[n];
    delete[] g->pcaVec;
  }
}


//Delete Gaussian Mixture information
void delete_GMMpatches(struct GMMpatches *gmm)
{
  delete[] gmm->labels;
  for (int n=0; n < gmm->ngaussian; n++) delete_gaussian(&(gmm->GM[n]));
  delete[] gmm->GM;
  delete gmm;
}

//Save Gaussian Mixture data to file
void saveGMMdata(struct GMMpatches *gmm, const char *name, unsigned char ZWtype)
{
    size_t tmp;
    FILE *out=fopen(name, "wb");
    
    int datatype=3; //Gaussian Mixture data
  if (ZWtype) datatype=6; //Zoran-Weiss Gaussian Mixture
    tmp=fwrite(&datatype, sizeof(int), 1, out);
    tmp=fwrite(&(gmm->ngaussian), sizeof(int), 1, out);
    tmp=fwrite(&(gmm->npatches), sizeof(int), 1, out);
    tmp=fwrite(&(gmm->dim), sizeof(int), 1, out);
    tmp=fwrite(gmm->labels, sizeof(int), gmm->npatches, out);
    for (int k=0; k < gmm->ngaussian; k++) {
        tmp=fwrite(&(gmm->GM[k].prob), sizeof(double), 1, out);
        tmp=fwrite(gmm->GM[k].mean, sizeof(double), gmm->dim, out);
        for (int i=0; i < gmm->dim; i++) 
            tmp=fwrite(gmm->GM[k].corr[i], sizeof(double), gmm->dim, out);    
    }
    
    tmp=fclose(out);
    tmp++;//to prevent compiler warning    
}

//Read Gaussian Mixture data from file
struct GMMpatches *readGMMdata(const char *name)
{
    size_t tmp;
    
    FILE *in=fopen(name, "rb");
    
    int datatype; 
    tmp=fread(&datatype, sizeof(int), 1, in);
    if ((datatype != 3) && (datatype != 6)) return NULL; //Gaussian mixture data
    
    struct GMMpatches *gmm=new struct GMMpatches;
    tmp=fread(&(gmm->ngaussian), sizeof(int), 1, in);
    tmp=fread(&(gmm->npatches), sizeof(int), 1, in);
    tmp=fread(&(gmm->dim), sizeof(int), 1, in);
    gmm->labels=new int[gmm->npatches];
    tmp=fread(gmm->labels, sizeof(int), gmm->npatches, in);
    gmm->GM=new struct gaussian[gmm->ngaussian];
    for (int k=0; k < gmm->ngaussian; k++) {
        gmm->GM[k].dim=gmm->dim;
        tmp=fread(&(gmm->GM[k].prob), sizeof(double), 1, in);
        gmm->GM[k].mean=new double[gmm->dim];
        tmp=fread(gmm->GM[k].mean, sizeof(double), gmm->dim, in);
        gmm->GM[k].corr=new double*[gmm->dim];
        for (int i=0; i < gmm->dim; i++) {
            gmm->GM[k].corr[i]=new double[gmm->dim];
            tmp=fread(gmm->GM[k].corr[i], sizeof(double), gmm->dim, in);     
        }
        gmm->GM[k].pcaVal=NULL;
        gmm->GM[k].pcaVec=NULL;
    }
    
    tmp=fclose(in);
    tmp++;//to prevent compiler warning    
    
    return gmm;
}


//For a given Gaussian in a mixture 
//Create image of vector basis
//npc patches of size wpatch x hpatch
//are displayed in a square image
//nrow patches are displayed per row
void PCAbaseviewGMM(unsigned char *Rout, unsigned char *Gout, unsigned char *Bout,
			        int wpatch, int hpatch, struct GMMpatches *gmm, int idgaussian,
			        int npc, int nrow)
{
  int dim=gmm->dim;

  unsigned char gray=(dim == wpatch*hpatch);
  int wmargin, hmargin;
  wmargin=hmargin=2;
  int wout=nrow*wpatch+(nrow-1)*wmargin;

  double *vector, aa;
  double maxvector=0;
  double *maxvectors=new double[npc];
  double *minvec=new double[npc];
  double *maxvec=new double[npc];
  for (int k=0; k < npc; k++) { //for each vector of the basis
    //read vector data and fins minimum and maximum (in absolute value)
	vector=gmm->GM[idgaussian].pcaVec[k]; 
	maxvectors[k]=0;
	minvec[k]=vector[0];
	maxvec[k]=vector[0];
	for (int n=0; n < dim; n++) {
	  if (abs(vector[n]) > maxvector) maxvector=abs(vector[n]);
	  if (abs(vector[n]) > maxvectors[k]) maxvectors[k]=abs(vector[n]);
	  if (vector[n] > maxvec[k]) maxvec[k]=vector[n];
	  if (vector[n] < minvec[k]) minvec[k]=vector[n];
	}
  }

  int m;
  for (int k=0; k < npc; k++) { //for each vector of the basis
    //read vector data and display them in a square image
    //for each vector, map (min, max)  to (0, 255)  
    //if constant value, then display it as gray (127)
	vector=gmm->GM[idgaussian].pcaVec[k]; 
    int n=0;
	for (int jj=0; jj < hpatch; jj++) {
      for (int ii=0; ii < wpatch; ii++) {
		m=(ii+(k%nrow)*(wpatch+wmargin))+jj*wout+(k/nrow)* \
		               (wout*(hpatch+hmargin));
	    if (gray) {
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f;
		  n++;
		  Rout[m]=(int) (aa+0.5f);
		  Gout[m]=(int) (aa+0.5f);
		  Bout[m]=(int) (aa+0.5f);
	    } else {
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f;
		  n++;
		  Rout[m]=(int) (aa+0.5f);
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f;
		  n++;
		  Gout[m]=(int) (aa+0.5f);
          if (maxvec[k] != minvec[k])
            aa=(vector[n]-minvec[k])*255.0/(maxvec[k]-minvec[k]); 
          else
            aa=127.5f;
		  n++;
		  Bout[m]=(int) (aa+0.5f);
	    }
	  }
	}
  }

  delete[] maxvectors;
  delete[] maxvec;
  delete[] minvec;

}

//For a given Gaussian in a mixture 
//create image of average patch associated to the Gaussian
void PCAmeanviewGMM(unsigned char *Rout, unsigned char *Gout, 
                    unsigned char *Bout,
                    int wpatch, int hpatch, 
                    struct GMMpatches *gmm, int idgaussian)
{
    int dim=gmm->dim;
    unsigned char gray=(dim == wpatch*hpatch);
    
    double *vector, aa;
    vector=gmm->GM[idgaussian].mean; 
  
    //set minimum value (if negative) to 0
    double min=vector[0];
    for (int n=1; n < dim; n++) {
        if (vector[n] < min) min=vector[n];
    }
    
    int n=0;
    for (int m=0; m < wpatch*hpatch; m++) {
        if (gray) {
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Rout[m]=(int) (aa+0.5f);
            Gout[m]=(int) (aa+0.5f);
            Bout[m]=(int) (aa+0.5f);
        } else {
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Rout[m]=(int) (aa+0.5f);
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Gout[m]=(int) (aa+0.5f);
            aa=(min < 0)?(vector[n]-min):(vector[n]);
            if (aa > 255) aa=255;
            n++;
            Bout[m]=(int) (aa+0.5f);
        }
    }
       
}


//Compute probnk=prob(gaussian k | given patch n)
//Note: GMM eventually obtained from a different set of patches
double **getprobnkGMM(double **patches, int npatches, int dim, 
                      struct GMMpatches *gmm)
{
    double **probnk;
    
    //allocate memory
    int ngauss=gmm->ngaussian;
    double *prob=new double[ngauss];
    double **mean=new double*[ngauss];
    for (int k=0; k < ngauss; k++) mean[k]=new double[dim];
    double ***covar=new double**[ngauss];
    for (int k=0; k < ngauss; k++) {
        covar[k]=new double*[dim];
        for (int i=0; i < dim; i++) covar[k][i]=new double[dim];
    }  
    
    //get information from GMMpatches data structure
    for (int k=0; k < ngauss; k++) {
        prob[k]=gmm->GM[k].prob;
        memcpy(mean[k], gmm->GM[k].mean, dim*sizeof(double));
        for (int i=0; i < dim; i++) memcpy(covar[k][i], gmm->GM[k].corr[i], 
                                           dim*sizeof(double));
    }
    
    //compute probnk=prob(gaussian k | given patch n)
    probnk=probnk_GaussianMixture((const double **) patches, 
                                  npatches, dim, 
                                  gmm->ngaussian, mean, covar, prob);
    
    //deallocate memory
    for (int n=0; n < ngauss; n++) delete[] mean[n];
    delete[] mean;
    for (int n=0; n < ngauss; n++) {
        for (int i=0; i < dim; i++)  delete[] covar[n][i];
        delete[] covar[n];
    }
    delete[] covar;
    delete[] prob;
    
    return probnk;
}


//convert patches values (after projection to 3D space) to
//values in the range [0, 255] than can be stored as a RGB color image 
void projectpatches_image(double **patchesProj, int npatches, 
						  unsigned char *R, unsigned char *G, unsigned char *B)
{
  //find max=maximum absolute value of patches values
  double proj1, proj2, proj3, maxproj=0;
  for (int n=0; n < npatches; n++) {
    proj1=patchesProj[n][0];
    proj2=patchesProj[n][1];
    proj3=patchesProj[n][2];
	if (abs(proj1) > maxproj) maxproj=abs(proj1);
	if (abs(proj2) > maxproj) maxproj=abs(proj2);
	if (abs(proj3) > maxproj) maxproj=abs(proj3);
  }

  //range [-max, max] is converted to range [0, 255]
  for (int n=0; n < npatches; n++) {
    proj1=patchesProj[n][0];
    proj2=patchesProj[n][1];
    proj3=patchesProj[n][2];
	R[n]=(int) (proj1*127.5f/maxproj+127.5);
	G[n]=(int) (proj2*127.5f/maxproj+127.5);
	B[n]=(int) (proj3*127.5f/maxproj+127.5);
  }

}

//scalar product of two dim dimensional vectors 
double scalar_product(double *v1, double *v2, int dim)
{
  double p=0;

  for (int k=0; k < dim; k++) p+=v1[k]*v2[k];

  return p;
}

//substraction of two dim dimensional vectors 
void substract_vectors(double *out, double *in2, int dim)
{
  for (int k=0; k < dim; k++) out[k]-=in2[k];
}

//Project patches (vectors) to 3 components (pc1, pc2 and pc3) of the PCA basis
double **projectpatchesPCA(double **patches, int npatches, int dim, 
                           double *dataPCA,
                           int nproj, int *comps)
{
    double **patchesProj=new double*[npatches];
    for (int n=0; n < npatches; n++) patchesProj[n]=new double[nproj];
    
    double *vectoraux=new double[dim];
    double *patchM=new double[dim];
    double proj;
    int kk;
    for (int n=0; n < npatches; n++) {
        memcpy(patchM, patches[n], dim*sizeof(double));
        //substract average vector associated to PCA basis
        //(center of coordinates)
        substract_vectors(patchM, dataPCA, dim);
        //project vectors 
        for (int k=0; k < nproj; k++) {
            kk=(comps)?(comps[k]):(k);
            memcpy(vectoraux, dataPCA+dim+dim+kk*dim, dim*sizeof(double));
            proj=scalar_product(patchM, vectoraux, dim);
            patchesProj[n][k]=proj;
        }
    }
    
    delete[] vectoraux;
    delete[] patchM;
    
    
    return patchesProj;
}





