/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#ifndef PATCHES_ANALYSIS_LIB_H
#define PATCHES_ANALYSIS_LIB_H

//data structure storing data of a Gaussian component from a Gaussian Mixture
struct gaussian {
  int dim; //dimension of the space of vectors
  double prob; //probability associated to the Gaussian in the mixture
  double *mean; //mean of the Gaussian
  double **corr; //covariance matrix of the Gaussian
  double *pcaVal; //variances associated to the basis vectors of the Gaussian
  double **pcaVec; //basis vectors of the Gaussian
};

//data structure storing data of a Gaussian Mixture
struct GMMpatches {
  int ngaussian; //number of Gaussians in the mixture
  int npatches; //number of analysed patches (vectors)
  int dim; //dimension of the analysed patches (vectors)
  int *labels; //label[n] = Gaussian to which patch n is associated
  struct gaussian *GM; //information of each Gaussian in the mixture
};

//performs Principal Components Analysis (PCA) of the set of patches
double *PCApatches(double **patches, int npatches, int dim, int &sizeout);

//creates image of PCA basis
void PCAbaseview(unsigned char *Rout, unsigned char *Gout, unsigned char *Bout,
			     int wpatch, int hpatch, double *dataPCA, int dim,
			     int npc, int nrow);
			     
//creates image of average patch of set of patches
void PCAmeanview(unsigned char *Rout, unsigned char *Gout, unsigned char *Bout,
			     int wpatch, int hpatch, double *dataPCA, int dim);

//computes Gaussian Mixture decomposition for a set of patches (vectors)
struct GMMpatches *getGMM(double **patches, int npatches, int dim, int ngauss);

//computes basis vectors for each component in a Gaussian Mixture (GM)
void computeGMMPCA(struct GMMpatches *gmm);

//deletes Gaussian Mixture information
void delete_GMMpatches(struct GMMpatches *gmm);

//saves Gaussian Mixture data to file
void saveGMMdata(struct GMMpatches *gmm, const char *name, unsigned char ZWtype);

//reads Gaussian Mixture data from file
struct GMMpatches *readGMMdata(const char *name);

//for a given Gaussian in a mixture creates image of vector basis
void PCAbaseviewGMM(unsigned char *Rout, unsigned char *Gout, 
                    unsigned char *Bout,
			        int wpatch, int hpatch, struct GMMpatches *gmm, 
			        int idgaussian,
			        int npc, int nrow);
			        
//for a given Gaussian in a mixture creates image of average patch associated 
//to the Gaussian
void PCAmeanviewGMM(unsigned char *Rout, unsigned char *Gout, 
                    unsigned char *Bout,
                    int wpatch, int hpatch, struct GMMpatches *gmm, 
                    int idgaussian);
			        
//for each path n in a set, computes probnk=prob(Gaussian k | given patch n)
double **getprobnkGMM(double **patches, int npatches, int dim, 
                      struct GMMpatches *gmm);

//converts patches values (after projection to 3D space) to
//values in the range [0, 255] than can be stored as a RGB color image 
void projectpatches_image(double **patchesProj, int npatches, 
						  unsigned char *R, unsigned char *G, unsigned char *B);
						  
//projects patches (vectors) to 3 components (pc1, pc2 and pc3) of the PCA basis
double **projectpatchesPCA(double **patches, int npatches, int dim, 
                           double *dataPCA,
                           int nproj, int *comps);
                           
//scalar product of two vectors of dimension 'dim'                            
double scalar_product(double *v1, double *v2, int dim);
                           


#endif


