/*
 * Copyright (c) 2011 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>


#include "patches_extraction_lib.h"
#include "../gradients/gradients_lib.h"
#include "../../../auxlibs/randmt/randmt.h"
#include "../../../auxlibs/splines/splinesMW.h"


//auxiliary data structure for image interpolation
struct datainterpolation {
    float *I;
    int w, h;
    const char *method; //"nearest", "bilinear" or "spline" (cubic B-spline)
    float *coeffs; //needed for B-splines
    int order; //needed for B-splines
};


//interpolated value of an image at pixel (xx, yy)
float data_interpolate(struct datainterpolation *data, float xx, float yy)
{
  float v=0;
  int w, h;
  float *I;
  
  w=data->w;
  h=data->h;
  I=data->I;
  
  if (!strcmp(data->method, "nearest")) {
    int ii, jj;
    ii=nearest(xx);
    jj=nearest(yy);
    v=I[ii+jj*w];
  }
  
  if (!strcmp(data->method, "bilinear")) {
    int ii, jj;
    float im0, im1, im2, im3, x, y;
    //bilinear interpolate in patch (ii, jj)-(ii+1, jj)-(ii, jj+1)-(ii+1, jj+1)
    ii=(int) xx;
    jj=(int) yy;
    x=xx-(float) ii;
    y=yy-(float) jj;
    im0=I[ii+jj*w];
    im1=I[(ii+1)+jj*w];
    im2=I[ii+(jj+1)*w];
    im3=I[(ii+1)+(jj+1)*w];
    v=(im3-im1+im0-im2)*x*y+(im1-im0)*x+(im2-im0)*y+im0;
  }
  
  if (!strcmp(data->method, "spline")) {
    float  ak[13];
    v=evaluate_splineMW(I, data->coeffs, xx, yy, &ak[0], data->order, 0, w, h); 
  }
  
  return v;
}

//get patches centered at (cx, cy), of size (wpatch, hpatch) 
//and sampled in the direction (vx, vy) (and its orthogonal) 
//use interpolation of the data (nearest neighbour, bilinear or cubic B-spline)
unsigned char get_sampled_patch(struct patch_info *info, 
                                struct datainterpolation *data, 
                                int wpatch, int hpatch, 
								unsigned char use_orientation)
{
  int w=data->w;
  int h=data->h;
    
  float vx, vy, wx, wy, xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3;
  double *patch=NULL;
  
  vx=info->vx;
  vy=info->vy;
  if (!use_orientation) {
    vx=1;
    vy=0;
  }
  //orthogonal vector to (vx, vy)
  wx=-vy;
  wy=vx;
  
  //limits of the patch
  //top left corner
  xx0=info->cx - ((float) wpatch/2.0f)*vx - ((float) hpatch/2.0f)*wx;
  yy0=info->cy - ((float) wpatch/2.0f)*vy - ((float) hpatch/2.0f)*wy;
  //top left point: (0.5, 0.5) offset from top left corner
  xx0+=0.5f*vx+0.5f*wx;
  yy0+=0.5f*vy+0.5f*wy;

  //top right point
  xx1=xx0+((float) (wpatch-1))*vx;
  yy1=yy0+((float) (wpatch-1))*vy;
  
  
  //bottom left point
  xx2=xx0+((float) (hpatch-1))*wx;
  yy2=yy0+((float) (hpatch-1))*wy;
  
  //bottom right point 
  xx3=xx1+((float) (hpatch-1))*wx;
  yy3=yy1+((float) (hpatch-1))*wy;
  
  if ((xx0 < 0) || (xx0 > w-1) || (yy0 < 0) || (yy0 > h-1) || 
      (xx1 < 0) || (xx1 > w-1) || (yy1 < 0) || (yy1 > h-1) ||
      (xx2 < 0) || (xx2 > w-1) || (yy2 < 0) || (yy2 > h-1) ||
      (xx3 < 0) || (xx3 > w-1) || (yy3 < 0) || (yy3 > h-1) ) return 0;
  	
  float xx, yy; 
    
  patch=new double[wpatch*hpatch];
  int np=0;
  int ii;
  for (int j=0; j < hpatch; j++) {
    for (int i=0; i < wpatch; i++) {
      ii=i;
      xx=xx0+(float) ii * vx + (float) j * wx;
      yy=yy0+(float) ii * vy + (float) j * wy;
      patch[np++]=(double) data_interpolate(data, xx, yy);
	  }
  }

  if (!info->patch1) info->patch1=patch;
  else {
    if (!info->patch2) info->patch2=patch;
    else info->patch3=patch;
  }
  return 1;
}



//Allocate data structures to store data of patches at random pixels
//offset=0.5 if gradient computed with bilinear interpolation
//else offset=0
struct patch_info *select_patches_centers_random(int w, int h, 
                                                 int wpatch, int hpatch, 
                                                 float prcpatches,
												 int &npatches, float offset)
{
  struct patch_info *info=NULL;

  //compute number of patches centers
  //range of x coordinates of the centers
  int imin, imax;
  //consider image domain in rectangle defined by (-0.5, -0.5) (top left corner) to (w-1+0.5, h-1+0.5) (bottom right corner)
  //conditions: 
  //imin+offset-wpatch/2 >= -0.5 (left border patch <= left border image) 
  //imax+offset+wpatch/2 <= w-1+0.5 (right border patch <= right border image)
  imin=(int) ceil(((double) wpatch)/2 - 0.5 - offset);
  imax=(int) floor((double) w - 0.5 - offset - ((double) wpatch)/2);
  //range of y coordinates of the centers
  int jmin, jmax;
  //consider image domain in rectangle defined by (-0.5, -0.5) (top left corner) to (w-1+0.5, h-1+0.5) (bottom right corner)
  //conditions: 
  //jmin+offset-hpatch/2 >= -0.5 (left border patch <= left border image) 
  //jmax+offset+hpatch/2 <= h-1+0.5 (right border patch <= right border image)
  jmin=(int) ceil(((double) hpatch)/2 - 0.5 - offset);
  jmax=(int) floor((double) h - 0.5 - offset -((double) hpatch)/2);
  
  int npatchesmax=(imax-imin+1)*(jmax-jmin+1);
  npatches=(int) (prcpatches * (float) npatchesmax);
  printf("npatchesmax=%i  npatches=%i\n", npatchesmax, npatches);

  info=new struct patch_info[npatches];

  init_randmt_auto();
  
  unsigned char *repeatedpatch=new unsigned char[npatchesmax];
  memset(repeatedpatch, 0, npatchesmax);

  double a;
  int idpatch, ipatch, jpatch;
  int np=0;
  if (prcpatches != 1.0) {
      while (np < npatches) {
          a=rand_unif(); //random number between 0 and 1
          idpatch=(int) (a*(double) (npatchesmax-1)); //random integer number between 0 and ndata-1
          if (!repeatedpatch[idpatch]) {
              ipatch=idpatch%(imax-imin+1);
              jpatch=idpatch/(imax-imin+1);
              repeatedpatch[idpatch]=1;
              info[np].cx=(float) (ipatch+imin)+offset;
              info[np].cy=(float) (jpatch+jmin)+offset;
              info[np].vx=1;
              info[np].vy=0;
              info[np].a=1;
              info[np].b=0;
              info[np].patch1=NULL;
              info[np].patch2=NULL;
              info[np].patch3=NULL;
              np++;
          }
      }
  } else { //select all the patches in correlative order
    np=0;
	for (int jpatch=jmin; jpatch <= jmax; jpatch++) {
	  for (int ipatch=imin; ipatch <= imax; ipatch++) {
	    info[np].cx=(float) ipatch+offset;
	    info[np].cy=(float) jpatch+offset;
	    info[np].vx=1;
	    info[np].vy=0;
	    info[np].a=1;
	    info[np].b=0;
	    info[np].patch1=NULL;
	    info[np].patch2=NULL;
	    info[np].patch3=NULL;
	    np++;
	  }
	}
  }


  delete[] repeatedpatch;
  return info;
}


//Allocate data structures to store data of patches at selected extrema
//offset=0.5 if gradient computed with bilinear interpolation
//else offset=0
struct patch_info *select_patches_centers_GradExtrema(struct gradient_info *gradE, 
                                                      int ngradE, int &npatches, 
                                                      float offset)
{
  struct patch_info *info=NULL;

  npatches=ngradE;
  info=new struct patch_info[npatches];
  for (int n=0; n < npatches; n++) {
	info[n].cx=gradE[n].x+offset;
	info[n].cy=gradE[n].y+offset;
	double vx, vy;
	vx=gradE[n].gx/sqrt(gradE[n].gmod2); //unitary vector in the direction 
	                                     //of the gradient
	vy=gradE[n].gy/sqrt(gradE[n].gmod2);
	//use vector in the direction of the edge (orthogonal to gradient vector)
	info[n].vx=vy;
	info[n].vy=-vx;
	info[n].a=1;
	info[n].b=0;
	info[n].patch1=NULL;
	info[n].patch2=NULL;
	info[n].patch3=NULL;
  }

  return info;
}



//Attach gradient information to patches data structures
void add_orientation_info(struct patch_info *info, int ninfo, 
                          struct gradient_info *gradI, 
						  int w, float th)
{
  int i, j, m;
  float th2=th*th;
  for (int n=0; n < ninfo; n++) {
	i=(int) (info[n].cx-0.5f);
	j=(int) (info[n].cy-0.5f);
	m=i+j*w;
	if (gradI[m].gmod2 >= th2) {
	  double vx, vy;
	  vx=gradI[m].gx/sqrt(gradI[m].gmod2); //unitary vector in the direction 
	                                       //of the gradient
	  vy=gradI[m].gy/sqrt(gradI[m].gmod2);
	  //use vector in the direction of the edge (orthogonal to gradient vector)
	  info[n].vx=vy;
	  info[n].vy=-vx;
	} else {
	  info[n].vx=1.0f;
	  info[n].vy=0.0f;
	}
  }

}



//Get patches centered at a set of selected pixels
//Input: input image (R, G, B), patches centers (info, ninfo)
//The function automatically determines if the image is a gray level image
//If R=G=B, then nchannels=1, 
//else nchannels=3 and the intensity I is computed as average of R, G and B
//Parameters: patch size (wpatch, hpatch), 
//            normalization options (use_orientation, use_normMean, use_normVar)
//            interpolation method (spline, nearest, bilinear),
//            gray or color patch ('gray' flag),
//Output: set of patches (patches), number of patches (npatches), 
//        dimension of the patches (dim=wpatch*hpatch*nchannels)
double **get_selected_patches(unsigned char *R, unsigned char *G, 
                              unsigned char *B, int w, int h,
						      struct patch_info *info, int ninfo, 
						      int wpatch, int hpatch,
						      unsigned char use_orientation, 
						      unsigned char use_normMean,
							  unsigned char use_normVar, int &dim, 
							  int &npatches, const char *methodInterp,
                              int npatchesmax)
{
  float *I=NULL;

  //check gray image
  unsigned char is_gray=1;
  for (int n=0; (n < w*h) && (is_gray); n++)
    is_gray=((R[n] == G[n]) && (R[n] == B[n]) && (G[n] == B[n]));

  //compute gray image
  if (is_gray) {
	I=new float[w*h];
	float average;
	for (int n=0; n < w*h; n++) {
	  average=((float) R[n]+ (float) G[n]+ (float) B[n])/3.0f;
	  I[n]=average;
	}
  }
    
  //Initialize interpolation parameters
  struct datainterpolation *dataIn=new struct datainterpolation;
  memset(dataIn, 0, sizeof(struct datainterpolation));
  dataIn->w=w;  
  dataIn->h=h;
  struct datainterpolation *dataInG=NULL, *dataInB=NULL;
  float *Rf=NULL, *Gf=NULL, *Bf=NULL;
  if (is_gray) { //fill input data (only gray case)
      dataIn->I=I;
      dataIn->method=methodInterp;
      dataIn->coeffs=NULL;
      dataIn->order=3;
      if (!strcmp(methodInterp, "spline")) {
          dataIn->coeffs=new float[w*h];
          finvsplineMW(I,dataIn->order,dataIn->coeffs,w,h);
      }
  } else {
      Rf=new float[w*h];
      for (int n=0; n < w*h; n++) Rf[n]=(float) R[n];
      dataIn->I=Rf;
      dataIn->method=methodInterp;
      dataIn->coeffs=NULL;
      dataIn->order=3;
      if (!strcmp(methodInterp, "spline")) {
          dataIn->coeffs=new float[w*h];
          finvsplineMW(Rf,dataIn->order,dataIn->coeffs,w,h);
      }
      Gf=new float[w*h];
      for (int n=0; n < w*h; n++) Gf[n]=(float) G[n];
      dataInG=new struct datainterpolation;
      memset(dataInG, 0, sizeof(struct datainterpolation));
      dataInG->w=w;
      dataInG->h=h;
      dataInG->I=Gf;
      dataInG->method=methodInterp;
      dataInG->coeffs=NULL;
      dataInG->order=3;
      if (!strcmp(methodInterp, "spline")) {
          dataInG->coeffs=new float[w*h];
          finvsplineMW(Gf,dataInG->order,dataInG->coeffs,w,h);
      }
      Bf=new float[w*h];
      for (int n=0; n < w*h; n++) Bf[n]=(float) B[n];
      dataInB=new struct datainterpolation;
      memset(dataInB, 0, sizeof(struct datainterpolation));
      dataInB->w=w;
      dataInB->h=h;
      dataInB->I=Bf;
      dataInB->method=methodInterp;
      dataInB->coeffs=NULL;
      dataInB->order=3;
      if (!strcmp(methodInterp, "spline")) {
          dataInB->coeffs=new float[w*h];
          finvsplineMW(Bf,dataInB->order,dataInB->coeffs,w,h);
      }
  }

  //number of channels
  int kgray=(is_gray)?(1):(3);
  dim=kgray*wpatch*hpatch;

  //Sample patches
  printf("Data type: %s\n", (is_gray)?"gray":"RGB");
  printf("Sample using %s interpolation\n", dataIn->method);
  npatches=0;
  unsigned char validpatch;
  for (int n=0; n < ninfo; n++) {
	if (is_gray) {
      validpatch=get_sampled_patch(&(info[n]), dataIn, wpatch, hpatch, 
                                   use_orientation); 
	} else {
      validpatch=get_sampled_patch(&(info[n]), dataIn, wpatch, hpatch, 
                                   use_orientation); 
      if (validpatch) validpatch=get_sampled_patch(&(info[n]), dataInG, 
                                                   wpatch, hpatch, 
                                                   use_orientation);
      if (validpatch) validpatch=get_sampled_patch(&(info[n]), dataInB, 
                                                   wpatch, hpatch, 
                                                   use_orientation);
	}
	if (validpatch) npatches++;
  }
  
  
  //If the number of extracted patches is larger than npatchesmax,
  //then randomly select npatchesmax patches and delete the rest
  //If npatchesmax == -1, then keep all the extracted patches
  double **patches=NULL;
  if (npatches > 0) {
    unsigned char *randomsel=NULL;
    if ((npatchesmax != -1) && (npatches > npatchesmax)) { //randomly 
                                                           //select patches
      npatches=npatchesmax;
      randomsel=new unsigned char[ninfo];
      memset(randomsel, 0, ninfo);
      //mt_init_auto();
      init_randmt_auto();
      double a;
      int idpatch, np=0;
      while (np < npatches) {
        //a=mt_drand53(); //random number between 0 and 1
        a=rand_unif(); //random number between 0 and 1
        idpatch=(int) (a*(double) (ninfo-1)); //random integer number 
                                              //between 0 and ndata-1
        if (!randomsel[idpatch] && (info[idpatch].patch1)) {
          randomsel[idpatch]=1;
          np++;
        }
      }
      //delete info of non selected patches
      for (int n=0; n < ninfo; n++) {
        if (!randomsel[n] && info[n].patch1) {
          delete[] info[n].patch1;
          info[n].patch1=NULL;
          if (info[n].patch2) {
            delete[] info[n].patch2;
            info[n].patch2=NULL;            
          }
          if (info[n].patch3) {
            delete[] info[n].patch3;
            info[n].patch3=NULL;            
          }
        }
      }
    }
    patches=new double*[npatches];
    for (int n=0; n < npatches; n++) patches[n]=new double[dim];
	int np=0;
    for (int n=0; n < ninfo; n++) {
      if (info[n].patch1) {
        if (!info[n].patch2 && !info[n].patch3) { //only information of I channel
          memcpy(patches[np], info[n].patch1, wpatch*hpatch*sizeof(double));
        } else { //information of R, G and B channels
          for (int k=0; k < wpatch*hpatch; k++) {
            patches[np][3*k]=info[n].patch1[k];
            patches[np][3*k+1]=info[n].patch2[k];
            patches[np][3*k+2]=info[n].patch3[k];
          }
        }
        np++;
      }
	}
    
    //If normalization options are active, 
    //then normalize mean to 0 and/or standard deviation to 1
	if (use_normMean || use_normVar) { 
	    double meanRef, stdRef, meanpatch, stdpatch, a, b, ii, iin, v; 
	    meanRef=0;
	    stdRef=1;
	    int np=0;
        for (int n=0; n < ninfo; n++) {
	        if (info[n].patch1) {
            meanpatch=0; 
            stdpatch=0;
            //compute mean, do not compute std here, 
            //possible numerical overflows
            if (!info[n].patch2 && !info[n].patch3) { //only information 
                                                      //of I channel
              for (int k=0; k < dim; k++) {
                meanpatch+=(double) patches[np][k];
              }
            } else { //always normalize intensity channel=(R+G+B)/3
              for (int k=0; k < wpatch*hpatch; k++) {
                ii=(patches[np][3*k]+patches[np][3*k+1]+patches[np][3*k+2])/3;
                meanpatch+=ii;
              }
            }
            meanpatch/=(double) (wpatch*hpatch);
            //compute std: this method of computation prevents 
            //numerical overflows
            if (!info[n].patch2 && !info[n].patch3) { //only information 
                                                      //of I channel
              for (int k=0; k < dim; k++) {
                v=(double) patches[np][k] - meanpatch;
                stdpatch+=v*v;
              }
            } else { //always normalize intensity channel=(R+G+B)/3
              for (int k=0; k < wpatch*hpatch; k++) {
                ii=(patches[np][3*k]+patches[np][3*k+1]+patches[np][3*k+2])/3;
                v=ii-meanpatch;
                stdpatch+=v*v;
              }
            }
            stdpatch/=(double) (wpatch*hpatch);
            stdpatch=sqrt(stdpatch);
            if (use_normVar && (fabs(stdpatch) > 1)) a=stdRef/stdpatch;
            else a=1; 
            if (use_normMean) b=meanRef-a*meanpatch;
            else b=meanpatch-a*meanpatch;
            
            info[n].a=a;
            info[n].b=b;
            if (!info[n].patch2 && !info[n].patch3) { //only information 
              for (int k=0; k < dim; k++) patches[np][k]=a*patches[np][k]+b;
            } else { //keep original R/G/B ratios
              for (int k=0; k < wpatch*hpatch; k++) {
                ii=(patches[np][3*k]+patches[np][3*k+1]+patches[np][3*k+2])/3;
                iin=a*ii+b;
                if (ii != 0) {
                  patches[np][3*k]=patches[np][3*k]*iin/ii;
                  patches[np][3*k+1]=patches[np][3*k+1]*iin/ii;
                  patches[np][3*k+2]=patches[np][3*k+2]*iin/ii;
                }
              }
            }
            
            np++;
          }
	    }
	} //end of mean and/or variance normalization
	
	//clean up
    if (randomsel) delete[] randomsel;
  } //end of 'if (npatches > 0) '

  //clean up
  if (I) delete[] I;
  if (Rf) delete[] Rf;
  if (Gf) delete[] Gf;
  if (Bf) delete[] Bf;
  if (dataIn->coeffs) delete[] dataIn->coeffs;
  delete dataIn;
  if (dataInG) {
      if (dataInG->coeffs) delete[] dataInG->coeffs;
      delete dataInG;
  }
  if (dataInB) {
      if (dataInB->coeffs) delete[] dataInB->coeffs;
      delete dataInB;
  }
    
  return patches;
}


//Draw line in a grey level image, from point (x1, y1) to point (x2, y2), ç
//and with value 'grey'
int drawgreyline(int x1, int y1, int x2, int y2, unsigned char grey, 
                 unsigned char *out, int width, int height)
{
  int x, y;
  float lambda, d12;

  // Calculamos la distancia entre ambos puntos
  d12=(float) sqrt((float) ((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)));

  if (d12 == 0.0) return 0;

  for (lambda=0.0f; lambda <= 1.0f; lambda+=(1/d12)) {
    x=x1 + (int) (lambda * (float) (x2-x1) + 0.5f);
    y=y1 + (int) (lambda * (float) (y2-y1) + 0.5f);
    if ((x < width) && (x >= 0) && (y < height) && (y >= 0)) out[x+y*width]=grey;  
  }

  return 1;
}



//data structure to store information on L2-distances between patches
struct data_distances {
    int idpatch;
    int idinfo;
    double d;
};

//auxiliary function to sort distances data in increasing order
int order_increasing_dist(const void *pVoid1, const void *pVoid2)
{
    struct data_distances *node1, *node2;
    
    node1=(struct data_distances *) pVoid1;
    node2=(struct data_distances *) pVoid2;
    
    if (node1->d < node2->d) return -1;
    if (node1->d > node2->d) return 1;
    return 0;
}

//squared L2 distance
double distance_patches(double *v1, double *v2, int dim)
{
    double d=0;
    for (int k=0; k < dim; k++) d+=((v1[k]-v2[k])*(v1[k]-v2[k]));
    return d;
}

//Get closest patches (L2 distance) to patch centered (aproximately) at (x0, y0) 
double **select_closest(double **patchesIn, int npatchesIn, 
                        struct patch_info *infoIn, int ninfoIn,
                        int dim, int x0, int y0, int nclosest, float dmax,
                        int &npatchesOut, struct patch_info **infoOut, int &id0)
{
    struct patch_info *infoSel=NULL;
    double **patchesSel=NULL;
    struct data_distances *datadist=new struct data_distances[npatchesIn];
    double d, dmin=-1;
    id0=0;
    int npatches=0;
    float dmax2=dmax*dmax;
    //find reference patch:
    //from all the input patches, select the one whose center is closest
    //to (x0, y0) 
    for (int n=0; n < ninfoIn; n++) {
        if (infoIn[n].patch1) { //valid patch
            d=(infoIn[n].cx-(float) x0)*(infoIn[n].cx-(float) x0)+ \
              (infoIn[n].cy-(float) y0)*(infoIn[n].cy-(float) y0);
            if ((dmax == -1) || (d <= dmax2)) {
              if ((dmin == -1) || (d < dmin)) {
                dmin=d;
                id0=n; //id0: index of selected patch in input set of patches
              }
              datadist[npatches].idpatch=npatches;
              datadist[npatches].idinfo=n;
              npatches++;
            }
        }
    }
    //compute L2 distances from all the patches to the selected patch
    for (int n=0; n < npatches; n++) {
        d=distance_patches(patchesIn[datadist[n].idinfo], patchesIn[id0], dim);
        datadist[n].d=d;
    }
    //sort distances
    qsort(datadist, npatches, sizeof(struct data_distances), 
          order_increasing_dist);

    //get closest patches
    npatchesOut=(nclosest < npatches)?(nclosest):(npatches);
    patchesSel=new double*[npatchesOut];
    infoSel=new struct patch_info[npatchesOut];
    for (int n=0; n < npatchesOut; n++) {
        patchesSel[n]=new double[dim];
        memcpy(patchesSel[n], patchesIn[datadist[n].idinfo],dim*sizeof(double));
        infoSel[n].cx=infoIn[datadist[n].idinfo].cx;
        infoSel[n].cy=infoIn[datadist[n].idinfo].cy;
        infoSel[n].vx=infoIn[datadist[n].idinfo].vx;
        infoSel[n].vy=infoIn[datadist[n].idinfo].vy;
        infoSel[n].a=infoIn[datadist[n].idinfo].a;
        infoSel[n].b=infoIn[datadist[n].idinfo].b;
        infoSel[n].patch1=NULL;
        infoSel[n].patch2=NULL;
        infoSel[n].patch3=NULL;
        if (!infoIn[datadist[n].idinfo].patch2 && 
            !infoIn[datadist[n].idinfo].patch3) { //1 channel
          infoSel[n].patch1=new double[1]; //just make it a non-NULL pointer
        } else { //3 channels
          infoSel[n].patch1=new double[1];//just make it a non-NULL pointer
          infoSel[n].patch2=new double[1];//just make it a non-NULL pointer
          infoSel[n].patch3=new double[1];//just make it a non-NULL pointer
        }
    }
    
    delete[] datadist;
    *infoOut=infoSel;
    return patchesSel;
}


//label: integer indicating some information about the stored data
void writeDdatalabel(double *data, int size, int label, const char *name)
{
    size_t tmp;
    FILE *out=fopen(name, "wb");
    tmp=fwrite(&label, sizeof(int), 1, out);
    double sizeD=(double) size;
    tmp=fwrite(&sizeD, sizeof(double), 1, out);
    tmp=fwrite(data, sizeof(double), size, out);
    tmp++;//to prevent compiler warning
    fclose(out);
}


//Write patches data to file
void write2Ddata(double **data, int n, int dim, const char *name)
{
  size_t tmp;
  FILE *out=fopen(name, "wb");
  tmp=fwrite(&n, sizeof(int), 1, out);
  tmp=fwrite(&dim, sizeof(int), 1, out);
  for (int i=0; i < n; i++)
    tmp=fwrite(data[i], sizeof(double), dim, out);
  tmp++;//to prevent compiler warning
  fclose(out);
}

//Read patches data from file
double **read2Ddata(const char *name, int &n, int &dim)
{
  size_t tmp;
  FILE *in=fopen(name, "rb");
  tmp=fread(&n, sizeof(int), 1, in);
  tmp=fread(&dim, sizeof(int), 1, in);
  double **data=new double*[n];
  for (int i=0; i < n; i++) {
    data[i]=new double[dim];
    tmp=fread(data[i], sizeof(double), dim, in);
  }
  fclose(in);
  tmp++;//to prevent compiler warning
  return data;
}

//Read patches label from file
//label: integer indicating some information about the stored data
double *readDdatalabel(const char *name, int &size)
{
    size_t tmp;
    FILE *in=fopen(name, "rb");
    int label;
    tmp=fread(&label, sizeof(int), 1, in);
    double sizeD;
    tmp=fread(&sizeD, sizeof(double), 1, in);
    size=(int) sizeD;
    double *data=new double[size];
    tmp=fread(data, sizeof(double), size, in);
    fclose(in);
    tmp++;//to prevent compiler warning
    return data;
}

//Read patches label from file
//label: integer indicating some information about the stored data
int readlabel(const char *name)
{
    size_t tmp;
    FILE *in=fopen(name, "rb");
    int label;
    tmp=fread(&label, sizeof(int), 1, in);
    fclose(in);
    tmp++;//to prevent compiler warning
    return label;    
}

//save patches additional information (location, gradient, etc) to binary file
void save_patch_info(struct patch_info *info, int ninfo, int w, int h, 
                     int wpatch, int hpatch, 
                     unsigned char use_orientation, 
                     const char *name)
{
    int nvalid=0;
    for (int n=0; n < ninfo; n++) if (info[n].patch1) nvalid++;
    
    FILE *out=fopen(name, "wb");
    fwrite(&nvalid, sizeof(int), 1, out);
    fwrite(&w, sizeof(int), 1, out);
    fwrite(&h, sizeof(int), 1, out);
    fwrite(&wpatch, sizeof(int), 1, out);
    fwrite(&hpatch, sizeof(int), 1, out);
    fwrite(&use_orientation, sizeof(unsigned char), 1, out);
    unsigned char use_patch2, use_patch3;
    for (int n=0; n < ninfo; n++) {
        if (info[n].patch1) {
            fwrite(&(info[n].cx), sizeof(float), 1, out);   
            fwrite(&(info[n].cy), sizeof(float), 1, out);   
            fwrite(&(info[n].vx), sizeof(float), 1, out);   
            fwrite(&(info[n].vy), sizeof(float), 1, out);   
            fwrite(&(info[n].a), sizeof(float), 1, out);   
            fwrite(&(info[n].b), sizeof(float), 1, out);   
            use_patch2=(info[n].patch2 != NULL);
            use_patch3=(info[n].patch3 != NULL);
            fwrite(&use_patch2, sizeof(unsigned char), 1, out);
            fwrite(&use_patch3, sizeof(unsigned char), 1, out);
        }
    }
    fclose(out);
}

//read patches additional information (location, gradient, etc) from binary file
struct patch_info *read_patch_info(const char *name, int &ninfo, int &w, int &h, 
                                   int &wpatch, int &hpatch, 
                                   unsigned char &use_orientation)
{
    size_t tmp;
    FILE *in=fopen(name, "rb");
    tmp=fread(&ninfo, sizeof(int), 1, in);
    tmp=fread(&w, sizeof(int), 1, in);
    tmp=fread(&h, sizeof(int), 1, in);
    tmp=fread(&wpatch, sizeof(int), 1, in);
    tmp=fread(&hpatch, sizeof(int), 1, in);
    tmp=fread(&use_orientation, sizeof(unsigned char), 1, in);
    struct patch_info *info=new struct patch_info[ninfo];
    unsigned char use_patch2, use_patch3;
    for (int n=0; n < ninfo; n++) {
        tmp=fread(&(info[n].cx), sizeof(float), 1, in);   
        tmp=fread(&(info[n].cy), sizeof(float), 1, in);   
        tmp=fread(&(info[n].vx), sizeof(float), 1, in);   
        tmp=fread(&(info[n].vy), sizeof(float), 1, in);   
        tmp=fread(&(info[n].a), sizeof(float), 1, in);   
        tmp=fread(&(info[n].b), sizeof(float), 1, in);   
        info[n].patch1=new double[1]; //just make it a non-NULL pointer
        info[n].patch2=NULL;
        info[n].patch3=NULL;
        tmp=fread(&use_patch2, sizeof(unsigned char), 1, in);
        tmp=fread(&use_patch3, sizeof(unsigned char), 1, in);
        if (use_patch2) info[n].patch2=new double[1]; //just make it a 
                                                      //non-NULL pointer
        if (use_patch3) info[n].patch3=new double[1]; //just make it a 
                                                      //non-NULL pointer        
    }    
    fclose(in);
    tmp++;//to prevent compiler warning
    return info;
}


//deallocate memory
void delete_patch_info(struct patch_info *info, int ninfo)
{
  for (int n=0; n < ninfo; n++) {
	if (info[n].patch1) delete[] info[n].patch1;
	if (info[n].patch2) delete[] info[n].patch2;
	if (info[n].patch3) delete[] info[n].patch3;
  }
  delete[] info;
}

//synthesizes 8 patches using a linear combination of basis vectors
//weighted by the standard deviation associated to each vector and a 
//an additional weighting factor
//(1 or -1 for the first 3 vectors and a random value from a 
//normal distribution (mean=0, std=1) for the rest of vectors)
//The first 3 vectors use a factor 1 or -1 to ensure that they contribute
//the most in the synthesis of patches 
double **simpatches(double *mean, double *val, double **vec, int dim)
{
    int npatchesSim=8;
    
    double **patchesSim=new double*[npatchesSim];
    for (int n=0; n < npatchesSim; n++) patchesSim[n]=new double[dim];
    
    double z;
    //all possible factors used for the three first vectors of the basis
    //and the 8 synthesized patches:
    //patch 1: weights (1, 1, 1)
    //patch 2: weights (1, 1, -1)
    //...
    //patch 8: weights (-1, -1, -1)
    double zz[3][8];
    zz[0][0]=zz[0][1]=zz[0][2]=zz[0][3]=1;
    zz[0][4]=zz[0][5]=zz[0][6]=zz[0][7]=-1;
    zz[1][0]=zz[1][1]=zz[1][4]=zz[1][5]=1;
    zz[1][2]=zz[1][3]=zz[1][6]=zz[1][7]=-1;
    zz[2][0]=zz[2][2]=zz[2][4]=zz[2][6]=1;
    zz[2][1]=zz[2][3]=zz[2][5]=zz[2][7]=-1;
    init_randmt_auto();
    for (int n=0; n < npatchesSim; n++) {
      memcpy(patchesSim[n], mean, dim*sizeof(double)); 
      for (int k=0; k < dim; k++) {
        z=rand_normal();
        if (k < 3) z=zz[k][n];        
        for (int m=0; m < dim; m++) {
          //check variance > 0, to prevent numerical errors 
          //weight each vector by its standard deviation
          if (val[k] >= 0) patchesSim[n][m]+= z *sqrt(val[k])*vec[k][m];
        }
      }
    }
      
    return patchesSim;
}




