/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "../auxlibs/rgbcube/rgbcube.h"
#include "library/patches/patches_extraction_lib.h"
#include "library/patches/patches_analysis_lib.h"

//Main code to create 3D representations of the set of patches
//
//Input: set of patches, Gaussian Mixture basis vectors, PCA basis 
//       vectors, indices of the 3 PCA vectors on which the patches 
//       are projected
//
//Output: different views of the 3D representation of the patches, with
//        a color that depends on the Gaussian each patch is associated to



//get different views of the 3D space representation of patches
//we are assuming that patches have already been projected to 3D space
void get3Dviews(double **patches, int npatches, const char *nameviews, 
                int w, int h,
                double **probnk, int ngaussians)
{
    char nameview[128];

    unsigned char *R=new unsigned char[npatches];
    unsigned char *G=new unsigned char[npatches];
    unsigned char *B=new unsigned char[npatches];
    
    
    //convert patches values (after projection to 3D space) to
    //values in the range [0, 255] than can be stored as a RGB color image 
    //The RGB cube associated to this image provides a 3D representation of
    //the patches (see IPOL publication
    //   Image Color Cube Dimensional Filtering an Visualization,
    //   J.L. Lisani, A. Buades, J-M. Morel, 2011 
    //for more information about the visualization of the RGB color cube 
    projectpatches_image(patches, npatches, R, G, B);
    
    //save image
    unsigned char *imgRGB=(unsigned char *) malloc(3*npatches);
    RGB2output(R, G, B, imgRGB, npatches);
    int nc=3; /*color image*/
    sprintf(nameview, "%s_RGB.png", nameviews);
    io_png_write_u8(nameview, imgRGB, npatches, 1, nc);
    
    //generate views
    //This function generates the parameters of the projection planes for a 
    //rotating sequence of views of the RGB cube
    //(code from auxiliary RGBcube library)
    //The computed parameters are stored in an auxiliary text file
    RGBviews_sequence_params("aux.txt");
    
    //assign colors to each patch depending on the Gassian to which it belongs
    unsigned char *IR=new unsigned char[npatches];
    unsigned char *IG=new unsigned char[npatches];
    unsigned char *IB=new unsigned char[npatches];
    if (probnk) { //if Gaussian information is available
        int ncolors=10; //maximum 10 different colors
        unsigned char *colorR=new unsigned char[ncolors];
        unsigned char *colorG=new unsigned char[ncolors];
        unsigned char *colorB=new unsigned char[ncolors];
        colorR[0]=255; colorG[0]=0; colorB[0]=0;
        colorR[1]=0; colorG[1]=0; colorB[1]=255;
        colorR[2]=0; colorG[2]=255; colorB[2]=0;
        colorR[3]=255; colorG[3]=255; colorB[3]=0;
        colorR[4]=255; colorG[4]=0; colorB[4]=255;
        colorR[5]=0; colorG[5]=255; colorB[5]=255;
        colorR[6]=128; colorG[6]=0; colorB[6]=0;
        colorR[7]=0; colorG[7]=0; colorB[7]=128;
        colorR[8]=0; colorG[8]=128; colorB[8]=0;
        colorR[9]=128; colorG[9]=128; colorB[9]=0;
        
        int kmax, icolor;
        double pmax;

        for (int n=0; n < npatches; n++) { //for each patch        
          //find Gaussian for which prob(gaussian | given patch) is highest
          kmax=0;
          pmax=probnk[n][kmax];
          for (int k=1; k < ngaussians; k++) {
            if (probnk[n][k] > pmax) {
              pmax=probnk[n][k];
              kmax=k; //index of Gaussian
            }
          }
          icolor=kmax%ncolors; //color assigned to Gaussian

          IR[n]=(int) (colorR[icolor]+0.5f);
          IG[n]=(int) (colorG[icolor]+0.5f);
          IB[n]=(int) (colorB[icolor]+0.5f);
          //printf("   RGB=(%i, %i, %i)\n", IR[n], IG[n], IB[n]);
        }

        delete[] colorR;
        delete[] colorG;
        delete[] colorB;
    } else { //default: red points
        for (int n=0; n < npatches; n++) {
            IR[n]=255;
            IG[n]=0;
            IB[n]=0;
        }        
    }
    
    
    unsigned char *Rview=new unsigned char[w*h];
    unsigned char *Gview=new unsigned char[w*h];
    unsigned char *Bview=new unsigned char[w*h];    
    unsigned char *imgview=(unsigned char *) malloc(3*w*h);
    
    
    //read parameters of projection planes for the different views of the 
    //RGB cube (computed above and stored in auxiliary text file)
    FILE *params;
    int nviews;
    float P[3], u[3], v[3];
    params=fopen("aux.txt", "r");
    if (fscanf(params, "%i", &nviews)) {} //to prevent compiler warning
    for (int n=0; n < nviews; n++) {
      if (fscanf(params, "%f %f %f %f %f %f %f %f %f", 
               &P[0], &P[1], &P[2], &u[0], &u[1], &u[2], 
               &v[0], &v[1], &v[2])) {} //to prevent compiler warning
        //get 2D projections (views) of RGB cube
        projectRGBcubeB(R, G, B, npatches, IR, IG, IB, 
                        Rview, Gview, Bview, w, h, P, u, v);
        //write the PNG output image 
        RGB2output(Rview, Gview, Bview, imgview, w*h);
        int nc=3; // color image
        sprintf(nameview, "%s_%i.png", nameviews, 100+n);
        io_png_write_u8(nameview, imgview, w, h, nc);
    }
    
    fclose(params);
    
    //clean up
    free(imgRGB);
    delete[] R;
    delete[] G;
    delete[] B;
    free(imgview);
    delete[] Rview;
    delete[] Gview;
    delete[] Bview;
    delete[] IR;
    delete[] IG;
    delete[] IB;
}


int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput1 = {(char *) "input1", NULL, (char *) "input patches"};
    parameters.push_back(&pInput1);
    ParStruct pInput2 = {(char *) "input2", NULL, (char *) "input PCA basis"};
    parameters.push_back(&pInput2);
    ParStruct pInput3 = {(char *) "input3", NULL, (char *) "input GM basis"};
    parameters.push_back(&pInput3);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    
    std::vector <OptStruct *> options;
    OptStruct oA  = {(char *) "a:", 0, (char *) "1", NULL, (char *) "First projection vector"}; 
    options.push_back(&oA);
    OptStruct oB  = {(char *) "b:", 0, (char *) "2", NULL, (char *) "Second projection vector"}; 
    options.push_back(&oB);
    OptStruct oC  = {(char *) "c:", 0, (char *) "3", NULL, (char *) "Third projection vector"}; 
    options.push_back(&oC);
    OptStruct oS  = {(char *) "s:", 0, (char *) "300", NULL, (char *) "size (pixels) of output images"}; 
    options.push_back(&oS);

    if (!parsecmdline((char *) "patches_3dprojs", (char *) "Projections on 3 basis vectors", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namepatches = pInput1.value;
    const char *namebasisPCA = pInput2.value;
    const char *namebasisGM = pInput3.value;
    const char *nameviews = pOutput.value;
    
    int pc1=atoi(oA.value);
    int pc2=atoi(oB.value);
    int pc3=atoi(oC.value);    
    int w=atoi(oS.value);
    int h=w;
    
    
    int npatches, dim;
    double **patches=NULL, **patchesProj=NULL;
    
    //read patches data
    patches=read2Ddata(namepatches, npatches, dim);
    printf("npatches=%i\n", npatches);
    
    if (!patches) {
        fprintf(stdout, "Unable to open patches information\n");
        return EXIT_FAILURE;        
    }
    
    int ncomponents=3;
    int *tabcomponents=new int[3];
    tabcomponents[0]=pc1;
    tabcomponents[1]=pc2;
    tabcomponents[2]=pc3;
    
    //read PCA basis vectors
    double *basisPCA=NULL;
    int sizedata;
    basisPCA=readDdatalabel(namebasisPCA, sizedata);

    
    //read Gaussian Mixture info
    int *labels=NULL;
    labels=new int[npatches];
    memset(labels, 0, npatches*sizeof(int));
    struct GMMpatches *gm=NULL;
    gm=readGMMdata(namebasisGM);
    printf("npatches=%i (%i)\n",npatches,gm->npatches);
    //compute basis of each Gaussian component
    computeGMMPCA(gm);
          

    
    //Project patches to 3 components (pc1, pc2 and pc3) of PCA basis
    //display 3D views and assign a different color to each Gaussian
    
    //compute probnk=prob(gaussian k | given patch n)
    double **probnk=getprobnkGMM(patches, npatches, gm->dim, gm);
    
    //get patches projections
    patchesProj=projectpatchesPCA(patches, npatches, dim, basisPCA, 
                                 ncomponents, tabcomponents);
    if (!patchesProj) {
        fprintf(stdout, "Unable to reconstruct patches\n");
    }
    
    //create 3D views
    char nameviewGM[120];
    sprintf(nameviewGM, "%sPCA", nameviews);
    get3Dviews(patchesProj, npatches, (const char *) nameviewGM, 
               w, h, probnk, gm->ngaussian);
    
    
    
    //clean up
    delete[] basisPCA;

    if (gm) delete_GMMpatches(gm);
    if (probnk) {
      for (int n=0; n < npatches; n++) delete[] probnk[n];
      delete[] probnk;
    }
        
    if (patches) {
        for (int n=0; n < npatches; n++) delete[] patches[n];
        delete[] patches;
    }
    if (patchesProj) {
        for (int n=0; n < npatches; n++) delete[] patchesProj[n];
        delete[] patchesProj;
    }
    delete[] tabcomponents;
    if (labels) delete[] labels;
    
    return EXIT_SUCCESS;

}
    

