/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "library/patches/patches_extraction_lib.h"


//create image displaying 'npatches' patches in 'nrow' rows and 'ncol' columns 
//with a distance of 'margin' pixels between them
//first displayed patch: index n0 (0 is for reference patch)
unsigned char *get_patches_image(double **patches, 
                                 struct patch_info *infopatches, 
                                 int npatches, int n0, int spatch, 
                                 int nchannels, int nrow, int ncol, int margin,
                                 int &wt, int &ht)
{    
    wt=nrow*spatch+(nrow-1)*margin;
    ht=ncol*spatch+(ncol-1)*margin;

    unsigned char *I=new unsigned char[wt*ht];
    unsigned char *img=(unsigned char *) malloc(3*wt*ht);
    unsigned char *IG=NULL; 
    unsigned char *IB=NULL;
    if (nchannels > 1) {
        IG=new unsigned char[wt*ht];
        IB=new unsigned char[wt*ht];
        memset(IG, 0, wt*ht);
        memset(IB, 0, wt*ht);
    }
    memset(I, 0, wt*ht);
    //get patches values
    float Iout, Ioutden;
    int m;
    for (int n=0; n < npatches; n++) {
      for (int j=0; j < spatch; j++) 
        for (int i=0; i < spatch; i++) {
          if (nchannels == 1) 
              Iout= patches[n0+n][i+j*spatch];
          else 
              Iout= (patches[n0+n][3*(i+j*spatch)]+ \
                     patches[n0+n][3*(i+j*spatch)+1]+ \
                     patches[n0+n][3*(i+j*spatch)+2])/3.0;
          //denormalize patch variance and mean, if it has been normalized
          //display with mean=40 and standard deviation=128
          Ioutden=Iout;
          if ((infopatches[0].a != 1) || (infopatches[0].b != 0)) { //normalized patch
              if (infopatches[0].a != 1) { //normalized variance (std=1)
                  //renormalize to std_out=40
                  Ioutden=Iout*40.0f;
              } 
              if (infopatches[0].b != 0) { //normalized mean (mean=0)
                  //renormalize to mean_out=128
                  Ioutden=Ioutden+128.0f;
              } 
          }
          //clip out-of-range values
          if (Ioutden > 255) Ioutden=255;
          if (Ioutden < 0) Ioutden=0;
          
          //display patch value
          m=(i+(n%nrow)*(spatch+margin))+j*wt+(n/nrow)*(wt*(spatch+margin));
          if (nchannels == 1) {
              I[m]=(int) Ioutden;
          } else {
              //get normalized R, G, B values
              if (Iout != 0) {
                  float v;
                  v=patches[n0+n][3*(i+j*spatch)]*Ioutden/Iout;
                  if (v > 255) v=255;
                  I[m]=(int) (v+0.5f);
                  v=patches[n0+n][3*(i+j*spatch)+1]*Ioutden/Iout;
                  if (v > 255) v=255;
                  IG[m]=(int) (v+0.5f);
                  v=patches[n0+n][3*(i+j*spatch)+2]*Ioutden/Iout;
                  if (v > 255) v=255;
                  IB[m]=(int) (v+0.5f);
              }
          }
        }
    }
    if (nchannels == 1) 
        RGB2output(I, I, I, img, wt*ht);
    else
        RGB2output(I, IG, IB, img, wt*ht);

    delete[] I;
    if (IG) delete[] IG;
    if (IB) delete[] IB;
    
    return img;
}


//Main code to select the closest (L2 distance) patches to a reference patch

//Input: patches data (output of patches_extract.cpp), central pixel of
//       reference patch, number of selected patches, size of search area of
//       closest patches
//Output: data of selected patches (patches values and extraction 
//        information for each patch), image of reference patch, image of 10 
//        closest patches to reference patch

int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput = {(char *) "input", NULL, (char *) "root name of input files"};
    parameters.push_back(&pInput);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    ParStruct pX = {(char *) "x", NULL, (char *) "x-coordinate of selected point"};
    parameters.push_back(&pX);
    ParStruct pY = {(char *) "y", NULL, (char *) "y-coordinate of selected point"};
    parameters.push_back(&pY);
    ParStruct pN = {(char *) "n", NULL, (char *) "number of closest points"};
    parameters.push_back(&pN);
    

    std::vector <OptStruct *> options;
    OptStruct oS = 	{(char *) "s:", 0, NULL, NULL, (char *) "search radius"};
    options.push_back(&oS);

    if (!parsecmdline((char *) "patches_closest", (char *) "Get closest patches to selected one", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namein = pInput.value;
    const char *nameout = pOutput.value;
    int x0=atoi(pX.value);
    int y0=atoi(pY.value);
    int nclosest=atoi(pN.value);
    int rsearch = -1;
    if (oS.flag) rsearch = atoi(oS.value);

    
    int npatchesIn, dim, ninfoIn, w, h, wpatch, hpatch;
    struct patch_info *infoIn=NULL;
    double **patchesIn=NULL;
    unsigned char orient;
    
    //Read patches data from file
    char namedatain[120];
    strcpy(namedatain, namein);
    strcat(namedatain, ".dat");
    patchesIn=read2Ddata(namedatain, npatchesIn, dim);
    printf("npatchesIn=%i\n", npatchesIn);
    
    //Read additional patches data from file (location in image, gradient info)
    strcpy(namedatain, namein);
    strcat(namedatain, "_loc.dat");
    infoIn=read_patch_info(namedatain, ninfoIn, w, h, wpatch, hpatch, orient);
    
    if (!patchesIn || !infoIn) {
        fprintf(stdout, "Unable to open input patches information\n");
        return EXIT_FAILURE;        
    }
    
    struct patch_info *infoOut=NULL;
    double **patchesOut=NULL;
    int npatchesOut=0;
    int id0;
    //get closest patches (L2 distance) to patch centered (aproximately) 
    //at (x0, y0) 
    patchesOut=select_closest(patchesIn, npatchesIn, infoIn, ninfoIn, dim, 
                              x0, y0, nclosest, rsearch, npatchesOut, &infoOut, 
                              id0);

    //save number of closest patches to text file
    printf("npatchesOut=%i\n", npatchesOut);
    FILE *f=fopen("closestinfo.txt", "w");
    fprintf(f, "nclosest final: %i\n", npatchesOut);
    fclose(f);
  
    //clean up and exit if no patches found
    if (npatchesOut == 0) {
      printf("No patches in the selected region\n");
      if (infoIn) delete_patch_info(infoIn, ninfoIn);
      if (patchesIn) {
        for (int n=0; n < npatchesIn; n++) delete[] patchesIn[n];
        delete[] patchesIn;
      }
      return EXIT_SUCCESS;      
    }
 
    //save patches data
    char namedataout[120];
    strcpy(namedataout, nameout);
    strcat(namedataout, ".dat");
    write2Ddata(patchesOut, npatchesOut, dim, (const char *) namedataout);
    strcpy(namedataout, nameout);
    strcat(namedataout, "_loc.dat");
    save_patch_info(infoOut, npatchesOut, w, h, wpatch, hpatch, orient, 
                    (const char *) namedataout);
  
  
    //Get image of reference patch= patchesOut[0]
    int spatch, nchannels;
    float v=sqrt((double) dim);
    int vi=(int) v;
    //patch size
    if (v == vi) { //dim=wpatch^2 (1 channel image)
        spatch=vi;
        nchannels=1;
    } else { //dim != wpatch^2, assume dim/3 = wpatch^2 (3 channels image)
        spatch=(int) sqrt((double) (dim/3));
        nchannels=3;
    }
    printf("%i channels\n", nchannels);
    int wt=spatch;
    int ht=spatch;  
    unsigned char *I=new unsigned char[wt*ht];
    unsigned char *IG=NULL, *IB=NULL;
    if (nchannels > 1) {
        IG=new unsigned char[wt*ht];
        IB=new unsigned char[wt*ht];
        memset(IG, 0, wt*ht);
        memset(IB, 0, wt*ht);
    }
    unsigned char *img=(unsigned char *) malloc(3*wt*ht);
    char nameimout[120];
    float Iout, Ioutden;
    memset(I, 0, wt*ht);
    //get patch values
    for (int j=0; j < spatch; j++) 
      for (int i=0; i < spatch; i++) {
        if (nchannels == 1) 
            Iout= patchesOut[0][i+j*spatch];
        else 
            Iout= (patchesOut[0][3*(i+j*spatch)]+ \
                   patchesOut[0][3*(i+j*spatch)+1]+ \
                   patchesOut[0][3*(i+j*spatch)+2])/3.0;
        //denormalize patch variance and mean, if it has been normalized
        //display with mean=40 and standard deviation=128
        Ioutden=Iout;
        if ((infoOut[0].a != 1) || (infoOut[0].b != 0)) { //normalized patch
            if (infoOut[0].a != 1) { //normalized variance (std=1)
                //renormalize to std_out=40
                Ioutden=Iout*40.0f;
            } 
            if (infoOut[0].b != 0) { //normalized mean (mean=0)
                //renormalize to mean_out=128
                Ioutden=Ioutden+128.0f;
            } 
        }
        //clip out-of-range values
        if (Ioutden > 255) Ioutden=255;
        if (Ioutden < 0) Ioutden=0;
        if (nchannels == 1) {
            I[i+j*wt]=(int) Ioutden;
        } else {
            //get normalized R, G, B values
            if (Iout != 0) {
                float v;
                v=patchesOut[0][3*(i+j*spatch)]*Ioutden/Iout;
                if (v > 255) v=255;
                I[i+j*wt]=(int) (v+0.5f);
                v=patchesOut[0][3*(i+j*spatch)+1]*Ioutden/Iout;
                if (v > 255) v=255;
                IG[i+j*wt]=(int) (v+0.5f);
                v=patchesOut[0][3*(i+j*spatch)+2]*Ioutden/Iout;
                if (v > 255) v=255;
                IB[i+j*wt]=(int) (v+0.5f);
            }
        }
      }
      
    if (nchannels == 1) 
        RGB2output(I, I, I, img, wt*ht);
    else
        RGB2output(I, IG, IB, img, wt*ht);

    sprintf(nameimout, "%s_ref.png", nameout);
    io_png_write_u8((const char *) nameimout, img, wt, ht, 3); //3 channels   

    delete[] img;
    delete[] I;
    if (IG) delete[] IG;
    if (IB) delete[] IB;
    
  
    //draw 10 closest
    //create patches image (horizontal)
    int ntop=(npatchesOut-1 > 10)?(10):(npatchesOut-1);
    int margin=2; //2 pixels separation between patches images

    //patch size
    v=sqrt((double) dim);
    vi=(int) v;
    if (v == vi) { //dim=wpatch^2 (1 channel image)
        spatch=vi;
        nchannels=1;
    } else { //dim != wpatch^2, assume dim/3 = wpatch^2 (3 channels image)
        spatch=(int) sqrt((double) (dim/3));
        nchannels=3;
    }

    //create output image of top patches arranged horizontally
    //do not display reference patch
    img=get_patches_image(patchesOut, infoOut, ntop, 1, spatch, nchannels, 
                          ntop, 1, margin, wt, ht);    
    sprintf(nameimout, "%s_top10.png", nameout);
    io_png_write_u8((const char *) nameimout, img, wt, ht, 3); //3 channels
    free(img);


    //create output square image of all the patches (including reference patch)
    //compute width and height of output image
    //npatchesOut-1 patches are displayed, we try to arrange them into a square image
    int nOut=npatchesOut; 
    int nrow=(int) sqrt((double) nOut);
    int ncol=(int) ceil((double) nOut/(double) nrow);
    img=get_patches_image(patchesOut, infoOut, nOut, 0, spatch, nchannels, 
                          nrow, ncol, margin, wt, ht);    
    sprintf(nameimout, "%s_all.png", nameout);
    io_png_write_u8((const char *) nameimout, img, wt, ht, 3); //3 channels
    free(img);

  
    //clean up
    if (infoIn) delete_patch_info(infoIn, ninfoIn);
    if (infoOut) delete_patch_info(infoOut, npatchesOut);
    if (patchesIn) {
        for (int n=0; n < npatchesIn; n++) delete[] patchesIn[n];
        delete[] patchesIn;
    }
    if (patchesOut) {
        for (int n=0; n < npatchesOut; n++) delete[] patchesOut[n];
        delete[] patchesOut;
    }
    
    return EXIT_SUCCESS;

}
    

