/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "library/patches/patches_extraction_lib.h"

//Main code to select the patches closest (L2 distance) to  
//each one of the patches closest to a reference patch
//
//Input: patches data (output of patches_extract.cpp), central pixel of
//       reference patch, number of selected patches, size of search area of
//       closest patches, original input image
//Output: images of the centers of the patches closest (L2 distance) to 
//        a reference patch (center of reference patch in blue, rest in red;
//        reference patches: initial selection plus patches closest to 
//        the initial selection), histograms of centers of closest patches

int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput = {(char *) "input", NULL, (char *) "root name of input files"};
    parameters.push_back(&pInput);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    ParStruct pX = {(char *) "x", NULL, (char *) "x-coordinate of selected point"};
    parameters.push_back(&pX);
    ParStruct pY = {(char *) "y", NULL, (char *) "y-coordinate of selected point"};
    parameters.push_back(&pY);
    ParStruct pN = {(char *) "n", NULL, (char *) "number of closest points"};
    parameters.push_back(&pN);
    

    std::vector <OptStruct *> options;
    OptStruct oS = 	{(char *) "s:", 0, NULL, NULL, (char *) "search radius"};
    options.push_back(&oS);
    OptStruct oO = 	{(char *) "o:", 0, NULL, NULL, (char *) "original image"};
    options.push_back(&oO);
    OptStruct oZ = 	{(char *) "z:", 0, NULL, NULL, (char *) "zoom factor for visualization (only values 0, 3, 5, 7, 9)"};
    options.push_back(&oZ);


    if (!parsecmdline((char *) "patches_closest2closest", (char *) "Get closest patches to the closest patches to the selected one", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namein = pInput.value;
    const char *nameout = pOutput.value;
    int x0=atoi(pX.value);
    int y0=atoi(pY.value);
    int nclosest=atoi(pN.value);
    int rsearch = -1;
    if (oS.flag) rsearch = atoi(oS.value);
    const char *nameoriginal=NULL;
    if (oO.flag) nameoriginal = oO.value;
    int zoom=0;
    if (oZ.flag) zoom = atoi(oZ.value);


    int npatchesIn, dim, ninfoIn, w, h, wpatch, hpatch;
    struct patch_info *infoIn=NULL;
    double **patchesIn=NULL;
    unsigned char orient;
    
    //Read patches data from file
    char namedatain[120];
    strcpy(namedatain, namein);
    strcat(namedatain, ".dat");
    patchesIn=read2Ddata(namedatain, npatchesIn, dim);
    printf("npatchesIn=%i\n", npatchesIn);
    
    //Read additional patches data from file (location in image, gradient info)
    strcpy(namedatain, namein);
    strcat(namedatain, "_loc.dat");
    infoIn=read_patch_info(namedatain, ninfoIn, w, h, wpatch, hpatch, orient);
    
    if (!patchesIn || !infoIn) {
        fprintf(stdout, "Unable to open input patches information\n");
        return EXIT_FAILURE;        
    }
    
    printf("Compute patches\n");
    struct patch_info **infoOut=new struct patch_info *[nclosest];
    double ***patchesOut=new double **[nclosest];
    int *npatchesOut=new int[nclosest];
    for (int n=0; n < nclosest; n++) {
      infoOut[n]=NULL;
      patchesOut[n]=NULL;
      npatchesOut[n]=0;
    }

    int id0;
    //get closest patches (L2 distance) to patch centered (aproximately) 
    //at (x0, y0) (reference patch)
    patchesOut[0]=select_closest(patchesIn, npatchesIn, infoIn, ninfoIn, dim, 
                                 x0, y0, nclosest, rsearch, npatchesOut[0], 
                                 &(infoOut[0]), id0);
    
    //for each one of the patches closest to the reference patch                          
    for (int n=1; n < npatchesOut[0]; n++) {
      //get closest patches (L2 distance) to a n-th patch
      patchesOut[n]=select_closest(patchesIn, npatchesIn, infoIn, ninfoIn, dim, 
                                   (int) infoOut[0][n].cx, 
                                   (int) infoOut[0][n].cy, 
                                   nclosest, rsearch, npatchesOut[n], 
                                   &(infoOut[n]), id0);
    }
  
  
    //save number of closest patches to text file
    FILE *f=fopen("closestinfo.txt", "w");
    fprintf(f, "nclosest final: %i\n", npatchesOut[0]);
    fclose(f);

    //clean up and exit if no patches found
    if (npatchesOut[0] == 0) {
      printf("No patches in the selected region\n");
      if (infoIn) delete_patch_info(infoIn, ninfoIn);
      if (patchesIn) {
        for (int n=0; n < npatchesIn; n++) delete[] patchesIn[n];
        delete[] patchesIn;
      }
      if (patchesOut) {
        for (int m=0; m < nclosest; m++) {
          if (patchesOut[m]) {
            for (int n=0; n < npatchesOut[m]; n++) delete[] patchesOut[m][n];
            delete[] patchesOut[m];
          }
        }
        delete[] patchesOut;
      }
      if (npatchesOut) delete[] npatchesOut;
      return EXIT_SUCCESS;      
    }
  
  
  
    //Create image of patches centers 
    //Use an odd zoom factor to resize the patches images
    int zoomfactor=1;
    if (zoom != 0) {
      if ((zoom != 3) && (zoom != 5) && (zoom != 7) && (zoom != 9)) zoomfactor=3;
      else zoomfactor=zoom;
    }
    int wC=zoomfactor*w;
    int hC=zoomfactor*h;
    unsigned char *RC=new unsigned char[wC*hC];
    unsigned char *GC=new unsigned char[wC*hC];
    unsigned char *BC=new unsigned char[wC*hC];
    unsigned char *imgC=(unsigned char *) malloc(3*wC*hC);
    unsigned char *Rin=NULL;
    unsigned char *Gin=NULL;
    unsigned char *Bin=NULL;
    unsigned char *imgin=NULL;
  
  
    for (int m=0; m < npatchesOut[0]; m++) {
  
      if (nameoriginal) { //read input image and resize it
        size_t ww, hh;
        imgin = io_png_read_u8_rgb(nameoriginal, &ww, &hh);
        input2RGB(imgin, &Rin, &Gin, &Bin, w*h);
        if (zoomfactor == 1) {
          memcpy(RC, Rin, wC*hC);
          memcpy(GC, Gin, wC*hC);
          memcpy(BC, Bin, wC*hC);
        } else {
          for (int j=0; j < hC; j++)
            for (int i=0; i < wC; i++) {
              RC[i+j*wC]=Rin[(i/zoomfactor)+(j/zoomfactor)*w];
              GC[i+j*wC]=Gin[(i/zoomfactor)+(j/zoomfactor)*w];
              BC[i+j*wC]=Bin[(i/zoomfactor)+(j/zoomfactor)*w];              
            }
        }
      }
  
      int i, j, cR, cG, cB;
      for (int n=npatchesOut[m]-1; n >= 0; n--) {
        cR=255; cG=0; cB=0; //mark centers of patches in red
        if (n == 0) { cR=0; cG=0; cB=255;} //use blue for the reference patch
        if (infoOut[m][n].patch1) {
          i=(int) infoOut[m][n].cx;
          j=(int) infoOut[m][n].cy;
          if (zoomfactor == 1) {
            RC[i+j*w]=cR;
            GC[i+j*w]=cG;
            BC[i+j*w]=cB;
            if (i > 0) {
              RC[(i-1)+j*w]=cR;
              GC[(i-1)+j*w]=cG;
              BC[(i-1)+j*w]=cB;
            }
            if (j > 0) {
              RC[i+(j-1)*w]=cR;
              GC[i+(j-1)*w]=cG;
              BC[i+(j-1)*w]=cB;
            }
            if (i+1 < w) {
              RC[(i+1)+j*w]=cR;
              GC[(i+1)+j*w]=cG;
              BC[(i+1)+j*w]=cB;
            }
            if (j+1 < h) {
              RC[i+(j+1)*w]=cR;
              GC[i+(j+1)*w]=cG;
              BC[i+(j+1)*w]=cB;
            }
          } else {
            RC[(i*zoomfactor+zoomfactor/2)+(j*zoomfactor+zoomfactor/2)*wC]=cR;
            GC[(i*zoomfactor+zoomfactor/2)+(j*zoomfactor+zoomfactor/2)*wC]=cG;
            BC[(i*zoomfactor+zoomfactor/2)+(j*zoomfactor+zoomfactor/2)*wC]=cB;            
            int ii, jj;
            for (int kk=0; kk < 4; kk++) {
              ii=-1; jj=0;
              if (kk == 1) { ii=1; jj=0; }
              if (kk == 2) { ii=0; jj=-1; }
              if (kk == 3) { ii=0; jj=1; }
              RC[(i*zoomfactor+zoomfactor/2+ii)+(j*zoomfactor+zoomfactor/2+jj)*wC]=cR;
              GC[(i*zoomfactor+zoomfactor/2+ii)+(j*zoomfactor+zoomfactor/2+jj)*wC]=cG;
              BC[(i*zoomfactor+zoomfactor/2+ii)+(j*zoomfactor+zoomfactor/2+jj)*wC]=cB;           
            }
          }
        }
      }
      RGB2output(RC, GC, BC, imgC, wC*hC);
      char nameimout2[120];
      sprintf(nameimout2, "%s_%i_C.png", nameout, m);
      io_png_write_u8((const char *) nameimout2, imgC, wC, hC, 3); //3 channels
  
    }
  
  
    //Create image of 2D histogram of centers of closest patches
    //Compute 2D histogram
    int *hist=new int[w*h]; //initialize to 0
    memset(hist, 0, w*h*sizeof(int));
    for (int m=0; m < npatchesOut[0]; m++) {
      int i, j;
      for (int n=0; n < npatchesOut[m]; n++) {
        if (infoOut[m][n].patch1) {
          i=(int) infoOut[m][n].cx;
          j=(int) infoOut[m][n].cy;
          hist[i+j*w]++; //increase count at bin (i, j)
        }
      }
    }
    //get maximum of histogram
    int hmax=0;
    for (int n=0; n < w*h; n++) if (hist[n] > hmax) hmax=hist[n];
    //create output image, linearly remap values from [0, hmax] to [0, 255]
    memset(RC, 0, wC*hC);
    float v;
    for (int j=0; j < h; j++) {
      for (int i=0; i < w; i++) {
        v=(float) hist[i+j*w]*255.0f/ (float) hmax;
        //v=(hist[i+j*w] != 0)?(255):(0);
        for (int jj=-zoomfactor/2; jj <= zoomfactor/2; jj++) {
          for (int ii=-zoomfactor/2; ii <= zoomfactor/2; ii++) {
            if ((i*zoomfactor+ii >= 0) && (i*zoomfactor+ii < wC) && 
                (j*zoomfactor+jj >= 0) && (j*zoomfactor+jj < hC))
              RC[(i*zoomfactor+ii)+(j*zoomfactor+jj)*wC]=(unsigned char) v;
          }
        }
      }
    }
    RGB2output(RC, RC, RC, imgC, wC*hC);
    char nameimout2[120];
    sprintf(nameimout2, "%s_mean.png", nameout);
    io_png_write_u8((const char *) nameimout2, imgC, wC, hC, 3); //3 channels
  
    delete[] hist;

    free(imgC);
    delete[] RC;
    delete[] GC;
    delete[] BC;
  
  
  
    //Create image of 2D centered histogram
    int wD, hD;
    wD=w;
    hD=h;
    int *histD=new int[wD*hD];
    memset(histD, 0, wD*hD*sizeof(int));
    float ux, uy, vx, vy, Ox, Oy, vv;
    //for each one of the patches closest to the reference patch   
    //(including the reference patch, i.e. index = 0)                          
    for (int m=0; m < npatchesOut[0]; m++) {
      //center and orientation of patch 
      Ox=infoOut[m][0].cx;
      Oy=infoOut[m][0].cy;
      vx=infoOut[m][0].vx;
      vy=infoOut[m][0].vy;
      vv=sqrt(vx*vx+vy*vy);
      int i, j;
      //read list of patches closest to the patch
      for (int n=0; n < npatchesOut[m]; n++) {
        //coordinates of center of closest patch in reference system 
        //with center at (Ox, Oy) and coordinates vectors (vx, vy)
        if (infoOut[m][n].patch1) {
          ux=infoOut[m][n].cx-Ox;
          uy=infoOut[m][n].cy-Oy;
          i=(int) ((ux*vx+uy*vy)/vv + 0.5f);
          j=(int) ((ux*(-vy)+uy*vx)/vv + 0.5f);
          if ((i+wD/2 >= 0) && (i+wD/2 < wD) && (j+hD/2 >= 0) && (j+hD/2 < hD)) 
              histD[i+wD/2+(j+hD/2)*wD]++; //increase histogram count
          
        }
      }
    }
    //display histogram values after linear mapping of original range
    //of values to [0, 255]
    wC=zoomfactor*wD;
    hC=zoomfactor*hD;
    unsigned char *IC=new unsigned char[wC*hC];
    imgC=(unsigned char *) malloc(3*wC*hC);
    hmax=0;
    for (int n=0; n < wD*hD; n++) if (histD[n] > hmax) hmax=histD[n];
    memset(IC, 0, wC*hC);
    for (int j=0; j < hD; j++) {
      for (int i=0; i < wD; i++) {
        v=(float) histD[i+j*wD]*255.0f/ (float) hmax;
        for (int jj=-zoomfactor/2; jj <= zoomfactor/2; jj++) {
          for (int ii=-zoomfactor/2; ii <= zoomfactor/2; ii++) {
            if ((i*zoomfactor+ii >= 0) && (i*zoomfactor+ii < wC) && 
                (j*zoomfactor+jj >= 0) && (j*zoomfactor+jj < hC))
              IC[(i*zoomfactor+ii)+(j*zoomfactor+jj)*wC]=(unsigned char) v;
          }
        }
      }
    }
    RGB2output(IC, IC, IC, imgC, wC*hC);
    char nameimout3[120];
    sprintf(nameimout3, "%s_mean_centered.png", nameout);
    io_png_write_u8((const char *) nameimout3, imgC, wC, hC, 3); //3 channels
  
    //display in logarithmic scale
    float hmaxL=0;
    for (int n=0; n < wD*hD; n++) 
        if ((histD[n] > 0) && (log((float) histD[n]) > hmaxL)) 
            hmaxL=log((float) histD[n]); //maximum value in log scale
    memset(IC, 0, wC*hC);
    for (int j=0; j < hD; j++) {
      for (int i=0; i < wD; i++) {
        if (histD[i+j*wD] > 0) 
            v=55.0f+log((float) histD[i+j*wD])*200.0f/ (float) hmaxL;
        else v=0;
        for (int jj=-zoomfactor/2; jj <= zoomfactor/2; jj++) {
          for (int ii=-zoomfactor/2; ii <= zoomfactor/2; ii++) {
            if ((i*zoomfactor+ii >= 0) && (i*zoomfactor+ii < wC) && 
                (j*zoomfactor+jj >= 0) && (j*zoomfactor+jj < hC))
              IC[(i*zoomfactor+ii)+(j*zoomfactor+jj)*wC]=(unsigned char) v;
          }
        }
      }
    }
    RGB2output(IC, IC, IC, imgC, wC*hC);
    char nameimout4[120];
    sprintf(nameimout4, "%s_mean_centered_log.png", nameout);
    io_png_write_u8((const char *) nameimout4, imgC, wC, hC, 3); //3 channels
  
  
    delete[] histD;
  
    free(imgC);
    delete[] IC;

  
  
  
    //clean up
    if (infoIn) delete_patch_info(infoIn, ninfoIn);
    if (patchesIn) {
        for (int n=0; n < npatchesIn; n++) delete[] patchesIn[n];
        delete[] patchesIn;
    }
    if (infoOut) {
      for (int m=0; m < nclosest; m++) {
        if (infoOut[m]) delete_patch_info(infoOut[m], npatchesOut[m]);
      }
      delete[] infoOut;
    }
    if (patchesOut) {
      for (int m=0; m < nclosest; m++) {
        if (patchesOut[m]) {
          for (int n=0; n < npatchesOut[m]; n++) delete[] patchesOut[m][n];
          delete[] patchesOut[m];
        }
      }
      delete[] patchesOut;
    }
    if (npatchesOut) delete[] npatchesOut;
  
    if (imgin) free(imgin);
    if (Rin) delete[] Rin;
    if (Gin) delete[] Gin;
    if (Bin) delete[] Bin;

    
    return EXIT_SUCCESS;

}
    

