/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "library/gradients/gradients_lib.h"
#include "library/patches/patches_extraction_lib.h"
#include "library/patches/patches_analysis_lib.h"


//Main code to extract patches from an image
//    
//Input: image and extraction parameters (extraction method, interpolation
//       method, gradient computation method, patch size, normalization options)
//        
//Output: patches data (patches values and extraction information for
//        each patch)


int main(int argc, char **argv)
{

    std::vector<ParStruct *> parameters;
    ParStruct pInput = {(char *) "input", NULL, (char *) "input image"};
    parameters.push_back(&pInput);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);

    std::vector <OptStruct *> options;
    OptStruct oC  = {(char *) "c:", 0, (char *) "canny", NULL, (char *) "Type of patch center (canny, harris, random)"}; 
    options.push_back(&oC);
    OptStruct oW  = {(char *) "w:", 0, (char *) "8", NULL, (char *) "Patch size"}; 
    options.push_back(&oW);
    OptStruct oI  = {(char *) "i:", 0, (char *) "spline", NULL, (char *) "Interpolation method (spline, nearest, bilinear)"}; 
    options.push_back(&oI);
    OptStruct oG  = {(char *) "g:", 0, (char *) "scharr", NULL, (char *) "Gradient computation method (scharr, sobel, bilinear)"}; 
    options.push_back(&oG);
    OptStruct oOnorm = 	{(char *) "o:", 0, NULL, NULL, (char *) "Orientation normalization flag"};
    options.push_back(&oOnorm);
    OptStruct oMnorm = 	{(char *) "m:", 0, NULL, NULL, (char *) "Mean normalization flag"};
    options.push_back(&oMnorm);
    OptStruct oVnorm = 	{(char *) "v:", 0, NULL, NULL, (char *) "Variance normalization flag"};
    options.push_back(&oVnorm);
    OptStruct oMax = 	{(char *) "x:", 0, NULL, NULL, (char *) "Maximum number of extracted patches"};
    options.push_back(&oMax);
    OptStruct oS = 	{(char *) "s:", 0, NULL, NULL, (char *) "Use smooth gradients"};
    options.push_back(&oS);

    if (!parsecmdline((char *) "patches_extract", (char *) "Patches extraction", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namein = pInput.value;
    const char *nameout = pOutput.value;
    
    //Parameters of the method
    //Method for the extraction of the patch centers
    const char *method = oC.value;
    //Patch size
    int wpatch = atoi(oW.value);
    //Interpolation method
    const char *methodInterp = oI.value;
    //Gradient computation method
    const char *methodGrad = oG.value;
    //Flag for orientation normalization
    unsigned char orient = 0;
    if (oOnorm.flag) orient = (atoi(oOnorm.value) == 0)?(0):(1);
    //Flag for mean normalization
    unsigned char normMean = 0;
    if (oMnorm.flag) normMean = (atoi(oMnorm.value) == 0)?(0):(1);
    //Flag for variance normalization
    unsigned char normVar = 0;
    if (oVnorm.flag) normVar = (atoi(oVnorm.value) == 0)?(0):(1);
    //Maximum number of extracted patches
    int npatchesmax = -1;
    if (oMax.flag) npatchesmax = atoi(oMax.value);
    //Use smooth gradients
    int radsmooth=2; //default parameter for gradient smoothing
    if (!oS.flag) radsmooth = 0; //if not set then do not smooth gradients
    

    //Default extraction thresholds
    
    //Minimum magnitude of the gradient at the extracted points
    double thgrad=10.0; //Gradient threshold 
    
    //Parameters for extraction of Harris points
    if (!strcmp(method, "harris")) thgrad=0.01; //Cornerness threshold for 
                                                //Harris points
    double thk=0.05; //Parameter for cornerness computation at Harris points
    
    //Parameter for the extraction of random points
    double prc=100.0; //Percentage of randomly extracted points (default: all)
    


    int hpatch=wpatch;      
    prc/=100.0f;
    int dim=wpatch*hpatch;
    
    //Read input image data
    size_t ww, hh;
    int w, h; //input image dimensions
    unsigned char *img; //image data
    unsigned char *R, *G, *B; //image data organized in R, G and B channels
    unsigned char *I;
    /* read the PNG input image */
    img = io_png_read_u8_rgb(namein, &ww, &hh);
    w=(int) ww;
    h=(int) hh;
    input2RGB(img, &R, &G, &B, w*h);
    //compute intensity channel (average of R, G and B)
    I=RGBtoI(R, G, B, w*h);

    //get gradient info (from intensity channel)
    struct gradient_info *gradI=NULL;
    struct gradient_info *gradE=NULL;
    int nextrema=0;
    //compute gradient if 'canny' or 'harris', or oriented patches
    if (strcmp(method, "random") || orient) { 
        gradI=get_gradient_info(I, w, h, methodGrad);
        if (radsmooth > 0) {
          //smooth gradient values
          struct gradient_info *gradS=compute_gradient_smooth(gradI, I, w, h, 
                                                              radsmooth);
          delete[] gradI;
          gradI=gradS;
        }
    }
  
    //auxiliary variable for the computation of gradients 
    float offset=0;
    if (!strcmp(methodGrad, "bilinear")) offset=0.5f;
  
    //select patches
    struct patch_info *infoPatches=NULL;
    int ninfoPatches;
    if (!strcmp(method, "random")) {
        //allocate data structures to store data of patches at random pixels
        infoPatches=select_patches_centers_random(w, h, wpatch, hpatch, prc, 
                                                  ninfoPatches, offset);
        //attach gradient information
        if (gradI) add_orientation_info(infoPatches, ninfoPatches, gradI, w, 
                                        thgrad);
    }
    if (!strcmp(method, "canny")) {
        //get Canny extrema
        gradE=get_canny_extrema(gradI, w, h, thgrad, nextrema);
        //allocate data structures to store data of patches at selected extrema
        infoPatches=select_patches_centers_GradExtrema(gradE, nextrema, 
                                                       ninfoPatches, offset);      
    }
    if (!strcmp(method, "harris")) {
        //get Harris extrema
        struct gradient_info *gradI0;
        gradI0=get_gradient_info(I, w, h, methodGrad); //no smoothing
        gradE=get_Harris_extrema(gradI0, w, h, thk, thgrad, nextrema);
        //use smoothed gradients for patches sampling
        int m;
        for (int n=0; n < nextrema; n++) {
          m=gradE[n].x+gradE[n].y*w;
          if ((gradE[n].x != gradI[m].x) || (gradE[n].y != gradI[m].y)) {
            printf("Error gradI0\n"); 
          }
          memcpy(&(gradE[n]), &(gradI[m]), sizeof(struct gradient_info)); 
        }
        delete[] gradI0;
        //allocate data structures to store data of patches at selected extrema
        infoPatches=select_patches_centers_GradExtrema(gradE, nextrema, 
                                                       ninfoPatches, offset);
    }    
    
    //get patches centered at the selected pixels
    double **patches=NULL;
    int npatches=0;
    if (infoPatches) patches=get_selected_patches(R, G, B, w, h, 
                                                  infoPatches, ninfoPatches,
                                                  wpatch, hpatch, 
                                                  orient, normMean, normVar, 
                                                  dim, npatches, methodInterp, 
                                                  npatchesmax);
    //Exit if no patches extracted
    if (patches == NULL) {
        printf("No patches found\n");
        free(img);
        delete[] R;
        delete[] G;
        delete[] B;
        delete[] I;
        if (gradI) delete[] gradI;
        if (gradE) delete[] gradE;
        if (infoPatches) delete_patch_info(infoPatches, ninfoPatches);    
        return EXIT_SUCCESS;
    }

 
    //save patches data
    char namedataout[120];
    strcpy(namedataout, nameout);
    strcat(namedataout, ".dat");
    //check if color image or gray image
    write2Ddata(patches, npatches, dim, (const char *) namedataout);
    printf("npatches=%i\n", npatches);
    strcpy(namedataout, nameout);
    strcat(namedataout, "_loc.dat");
    save_patch_info(infoPatches, ninfoPatches, w, h, wpatch, hpatch, 
                    orient, (const char *) namedataout);
  
  
    //open auxiliary file storing the number of extracted patches
    FILE *faux=fopen("npatches.txt", "wt");
    fprintf(faux, "%i\n", npatches);
    //close auxiliary file
    fclose(faux);
  

    //clean up
    free(img);
    delete[] R;
    delete[] G;
    delete[] B;
    delete[] I;
    if (gradI) delete[] gradI;
    if (gradE) delete[] gradE;
    if (infoPatches) delete_patch_info(infoPatches, ninfoPatches);
    if (patches) {
        for (int n=0; n < npatches; n++) delete[] patches[n];
        delete[] patches;
    }
    
    return EXIT_SUCCESS;

}
    

