/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "library/patches/patches_extraction_lib.h"
#include "library/patches/patches_analysis_lib.h"


//Main code to apply Gaussian Mixture (GM) analysis to a set of patches
//
//Input: set of patches, number of Gaussians in the mixture
//
//Output (for each Gaussian): basis vectors and corresponding variances, 
//        image displaying the basis vectors, average patch of the patches
//        contributing to the Gaussian, graphical representation of the variance 
//        of each basis component, probability of the Gaussian in the mixture
        
int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput = {(char *) "input", NULL, (char *) "input patches"};
    parameters.push_back(&pInput);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    ParStruct pN = {(char *) "n", NULL, (char *) "number of Gaussians"};
    parameters.push_back(&pN);
    
    std::vector <OptStruct *> options;

    if (!parsecmdline((char *) "patches_gm", (char *) "Gaussian Mixture from a set of patches", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namepatches = pInput.value;
    const char *nameout = pOutput.value;
    int ngaussians=atoi(pN.value);


    int npatches, dim;
    double **patches=NULL;
    
    //Read patches data from file
    patches=read2Ddata(namepatches, npatches, dim);
    printf("npatches=%i\n", npatches);
    
    //data structure that will hold GM data
    struct GMMpatches *gm=NULL;
  
    int ntries=0, ntriesmax=3;
    unsigned char errorGM;
    do { //Try to compute GMM up to ntriesmax times 
         //(with different inital seeds)
         
      //compute Gaussian Mixture decomposition for a set of patches
      gm=getGMM(patches, npatches, dim, ngaussians);    
      
      //compute basis vectors for each Gaussian
      computeGMMPCA(gm);
    
      //check if result is correct
      //check that basis vectors are orthonormal (i.e. unitary and orthogonal)
      double *vector1=new double[dim];
      double *vector2=new double[dim];
      double vv;
      errorGM=0;
      for (int i=0; i < gm->ngaussian; i++) {
        for (int k1=0; k1 < dim; k1++) {
          memcpy(vector1, gm->GM[i].pcaVec[k1], dim*sizeof(double));
          for (int k2=k1; k2 < dim; k2++) {
            memcpy(vector2, gm->GM[i].pcaVec[k2], dim*sizeof(double));
            vv=scalar_product(vector1, vector2, dim);
            if ((k2 == k1) && (fabs(vv-1) > 0.001)) {
              //printf("Warning non-unitary vector %i, Gaussian %i: %2.2f\n", 
              //k1, i, vv);
              errorGM=1;
            }
            if ((k2 != k1) && (fabs(vv) > 0.001)) {
              //printf("Warning non-orthogonal vectors %i-%i, 
              //Gaussian %i: %2.2f\n", k1, k2, i, vv);
              errorGM=1;
            }
          }
        }
      }
      delete[] vector1;
      delete[] vector2;
      
      if (errorGM) {
        if (ntries+1 < ntriesmax) {
            printf("Error in computation of Gaussian Mixture ... let's try again\n");
            delete_GMMpatches(gm);
        }
        //wait 1 second, to prevent "random" initialization of GM with 
        //the same set of values than in the previous attempt
        unsigned long int seedprev;
        unsigned long int seed = (unsigned long int) time(NULL);
        do {
          seedprev=seed;
          seed = (unsigned long int) time(NULL);
        } while (seed == seedprev);
      }
      ntries++;
      //
    } while (errorGM && (ntries < ntriesmax));
  
    if (errorGM) {
      //display warning message and create auxiliary file indicating the problem
      printf("Error in computation of Gaussian Mixture. Some basis vectors may be unreliable.\n");
      FILE *faux=fopen("gmwarning.txt", "wt");
      fprintf(faux, "Error in computation of Gaussian Mixture. Some basis vectors may be unreliable.\n");
      fclose(faux);
    }
  
    //Save GM data to file
    char namedataout[120];
    strcpy(namedataout, nameout);
    strcat(namedataout, ".dat");
    saveGMMdata(gm, (const char *) namedataout, 0);
    
    //images of GMM basis: one image for each Gaussian
    int wpatch;
    float v=sqrt((double) dim);
    int vi=(int) v;
    if (v == vi) { //dim=wpatch^2 (1 channel image)
        wpatch=vi;
        printf("1 channel image\n");
    } else { //dim != wpatch^2, assume dim/3 = wpatch^2 (3 channels image)
        wpatch=(int) sqrt((double) (dim/3));
        printf("3 channels images\n");
    }
    int hpatch=wpatch;
    int npc=dim;
    int wmargin, hmargin;
    wmargin=hmargin=2;//2-pixels margin between displayed images of patches
    //compute width and height of output image
    //dim patches are displayed, we try to arrange them into a square image
    int nrow=(int) sqrt((double) npc);
    int wout=nrow*wpatch+(nrow-1)*wmargin;
    int ncol=(int) ceil((double) npc/(double) nrow);
    int hout=ncol*hpatch+(ncol-1)*hmargin;
    unsigned char *Rout=new unsigned char[wout*hout];
    unsigned char *Gout=new unsigned char[wout*hout];
    unsigned char *Bout=new unsigned char[wout*hout];
    unsigned char *img=(unsigned char *) malloc(3*wout*hout);
    memset(Rout, 0, wout*hout);
    memset(Gout, 0, wout*hout);
    memset(Bout, 0, wout*hout);
  
    char nameimout[120];
    //create auxiliary file 'gaussprobs.txt' 
    //and write probabilities of each Gaussian
    FILE *faux=fopen("gaussprobs.txt", "wt");
    for (int idgaussian=0; idgaussian < gm->ngaussian; idgaussian++) {
      memset(Rout, 0, wout*hout);
      memset(Gout, 0, wout*hout);
      memset(Bout, 0, wout*hout);
      //for each Gaussian, draw basis vectors as patches
      PCAbaseviewGMM(Rout, Gout, Bout, wpatch, hpatch,
                     gm, idgaussian, npc, nrow);
      RGB2output(Rout, Gout, Bout, img, wout*hout);
      sprintf(nameimout, "%s_%i.png", nameout, idgaussian);
      io_png_write_u8((const char *) nameimout, img, wout, hout, 3);//3 channels
      //write probability of the Gaussian in auxiliary file 'gaussprobs.txt'
      fprintf(faux, "prob %i: %2.2f\n", idgaussian, 
              gm->GM[idgaussian].prob*100.0f);
    }
    fclose(faux);
    
    //clean up
    free(img);
    delete[] Rout;
    delete[] Gout;
    delete[] Bout;
    
    
    //images of average patches: one image for each Gaussian
    wout=wpatch;
    hout=hpatch;
    Rout=new unsigned char[wout*hout];
    Gout=new unsigned char[wout*hout];
    Bout=new unsigned char[wout*hout];
    img=(unsigned char *) malloc(3*wout*hout);
    memset(Rout, 0, wout*hout);
    memset(Gout, 0, wout*hout);
    memset(Bout, 0, wout*hout);
    char nameimout2[120];
    for (int idgaussian=0; idgaussian < gm->ngaussian; idgaussian++) {
        memset(Rout, 0, wout*hout);
        memset(Gout, 0, wout*hout);
        memset(Bout, 0, wout*hout);
        //draw average patch
        PCAmeanviewGMM(Rout, Gout, Bout, wpatch, hpatch, gm, idgaussian);
        //save image to file
        RGB2output(Rout, Gout, Bout, img, wout*hout);
        sprintf(nameimout2, "%s_%i_mean.png", nameout, idgaussian);
        io_png_write_u8((const char *) nameimout2, img, wout, hout, 3); 
    }
    
    //clean up
    free(img);
    delete[] Rout;
    delete[] Gout;
    delete[] Bout;
    
    //For each Gaussian, create graphical representation of the variance 
    //of each basis component (i.e. magnitude of eigenvalues associated to
    //to each eigenvector of the Gaussian basis)
    unsigned char *Ivp=new unsigned char[dim*dim];
    unsigned char *imgEigV=(unsigned char *) malloc(3*dim*dim);
    char nameimout3[120];
    for (int idgaussian=0; idgaussian < gm->ngaussian; idgaussian++) {
        double maxEigV=gm->GM[idgaussian].pcaVal[0];
        memset(Ivp, 0, dim*dim);
        for (int k=0; k < dim; k++) { //x-coordinate of graphic: index of vector
            for (int j=dim-1; 
                 j >= dim-(int) dim*gm->GM[idgaussian].pcaVal[k]/maxEigV; j--)
                //y-coordinate: magnitude of eigenvalue associated to vector
                Ivp[k+j*dim]=255;
        }
        //save image to file
        RGB2output(Ivp, Ivp, Ivp, imgEigV, dim*dim);
        sprintf(nameimout3, "%s_%i_EigV.png", nameout, idgaussian);
        io_png_write_u8((const char *) nameimout3, imgEigV, dim, dim, 3);
    }
  
  
    //clean up
    free(imgEigV);
    delete[] Ivp;

    //Compute standard deviation of each GM: sqrt(sum(eigenvalues))
    //and save values to text file
    FILE *faux2=fopen("gaussstd.txt", "wt");
    for (int idgaussian=0; idgaussian < gm->ngaussian; idgaussian++) {
        double stdPCA=0;
        for (int k=0; k < dim; k++) stdPCA+=gm->GM[idgaussian].pcaVal[k];
        stdPCA=sqrt(stdPCA)/sqrt(dim);
        fprintf(faux2, "std %i: %2.2f\n", idgaussian, stdPCA);
    }
    fclose(faux2);
    
    //clean up
    if (patches) {
        for (int n=0; n < npatches; n++) delete[] patches[n];
        delete[] patches;
    }
    if (gm) delete_GMMpatches(gm);
    
    return EXIT_SUCCESS;
    
}

