/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "library/patches/patches_extraction_lib.h"
#include "library/patches/patches_analysis_lib.h"


//Main code to analyse which patches from a given set are 
//associated to each component of a Gaussian Mixture
//
//Input: set of patches, basis of a Gaussian Mixture
//
//Output: image displaying the centers of the input patches with a
//        color that depends on the Gaussian they are associated to; for each 
//        Gaussian, image of the 10 patches associated to the Gaussian that are
//        closest to a reference patch (the first patch in the input set) 


//auxiliary data structure to store distance from patch to a given Gaussian
struct data_distancesP {
  int idpatch;
  double d;
};

//sort patches according to increasing distance to Gaussian
int order_increasing_distP(const void *pVoid1, const void *pVoid2)
{
  struct data_distancesP *node1, *node2;
  
  node1=(struct data_distancesP *) pVoid1;
  node2=(struct data_distancesP *) pVoid2;
  
  if (node1->d < node2->d) return -1;
  if (node1->d > node2->d) return 1;
  return 0;
}

//squared L2 distance between patches
double distance_patchesP(double *v1, double *v2, int dim)
{
  double d=0;
  for (int k=0; k < dim; k++) d+=((v1[k]-v2[k])*(v1[k]-v2[k]));
  return d;
}


int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput1 = {(char *) "input1", NULL, (char *) "input patches"};
    parameters.push_back(&pInput1);
    ParStruct pInput2 = {(char *) "input2", NULL, (char *) "input patches info"};
    parameters.push_back(&pInput2);
    ParStruct pInput3 = {(char *) "input3", NULL, (char *) "input basis"};
    parameters.push_back(&pInput3);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    
    std::vector <OptStruct *> options;
    OptStruct oZ = 	{(char *) "z:", 0, NULL, NULL, (char *) "zoom factor for visualization (only values 0, 3, 5, 7, 9)"};
    options.push_back(&oZ);
    OptStruct oC  = {(char *) "c", 0, NULL, NULL, (char *) "draw 10 closest patches to each Gaussian"}; 
    options.push_back(&oC);
    OptStruct oO = 	{(char *) "o:", 0, NULL, NULL, (char *) "original image"};
    options.push_back(&oO);

    if (!parsecmdline((char *) "patches_gmprobs", (char *) "GMprobs", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namepatches = pInput1.value;
    const char *nameinfo = pInput2.value;
    const char *namebasis = pInput3.value;
    const char *nameout = pOutput.value;
    
    int zoom=0;
    if (oZ.flag) zoom = atoi(oZ.value);
    unsigned char closest= oC.flag;
    const char *nameorig=NULL;
    if (oO.flag) nameorig = oO.value;




    int npatches, dim, ninfo, w, h, wpatch, hpatch, ngauss;
    unsigned char orient;
    struct patch_info *info=NULL;
    double **patches=NULL;
 
  
    char nameimout[120];

    //read patches data
    patches=read2Ddata(namepatches, npatches, dim);
    printf("npatches=%i\n", npatches);
    //read patches additional info (location, orientation, etc)
    info=read_patch_info(nameinfo, ninfo, w, h, wpatch, hpatch, orient);
    
    if (!patches || !info) {
        fprintf(stdout, "Unable to open patches information\n");
        return EXIT_FAILURE;        
    }

    //read GM data from file
    struct GMMpatches *gm=NULL;
    gm=readGMMdata(namebasis);
    ngauss=gm->ngaussian;
    printf("npatches=%i (%i)  ngaussians=%i\n",npatches,gm->npatches, ngauss);

    unsigned char *R=new unsigned char[w*h];
    unsigned char *G=new unsigned char[w*h];
    unsigned char *B=new unsigned char[w*h];
    unsigned char *img=(unsigned char *) malloc(3*w*h);
    int i, j;

    //compute labels
    int *labels=NULL;
    
    //compute probnk=prob(gaussian k | given patch n)
    double **probnk=NULL;
    probnk=getprobnkGMM(patches, npatches, dim, gm);
 
    //create image with pixel(center of patch)=color(k),with k=argmax p(k|patch) 
    int ncolors=10; //maximum 10 different colors
    unsigned char *colorR=new unsigned char[ncolors];
    unsigned char *colorG=new unsigned char[ncolors];
    unsigned char *colorB=new unsigned char[ncolors];
    colorR[0]=255; colorG[0]=0; colorB[0]=0;
    colorR[1]=0; colorG[1]=0; colorB[1]=255;
    colorR[2]=0; colorG[2]=255; colorB[2]=0;
    colorR[3]=255; colorG[3]=255; colorB[3]=0;
    colorR[4]=255; colorG[4]=0; colorB[4]=255;
    colorR[5]=0; colorG[5]=255; colorB[5]=255;
    colorR[6]=128; colorG[6]=0; colorB[6]=0;
    colorR[7]=0; colorG[7]=0; colorB[7]=128;
    colorR[8]=0; colorG[8]=128; colorB[8]=0;
    colorR[9]=128; colorG[9]=128; colorB[9]=0;

    //use only odd zoom factors
    int zoomfactor=1;
    if (zoom != 0) {
      if ((zoom != 3) && (zoom != 5) && (zoom != 7) && (zoom != 9)) zoomfactor=3;
      else zoomfactor=zoom;
    }
    int wC=zoomfactor*w;
    int hC=zoomfactor*h;
    unsigned char *RC=new unsigned char[wC*hC];
    unsigned char *GC=new unsigned char[wC*hC];
    unsigned char *BC=new unsigned char[wC*hC];
    unsigned char *imgC=(unsigned char *) malloc(3*wC*hC);
  
    if (nameorig) {
      //background=original image
      size_t ww, hh;
      unsigned char *img0; //image data
      unsigned char *R0, *G0, *B0; //image data organized in R, G and B channels
      /* read the PNG input image */
      img0 = io_png_read_u8_rgb(nameorig, &ww, &hh);
      input2RGB(img0, &R0, &G0, &B0, w*h);  
      //rescale input image
      if (zoomfactor == 1) {
        memcpy(RC, R0, wC*hC);
        memcpy(GC, G0, wC*hC);
        memcpy(BC, B0, wC*hC);
      } else {
        for (int j=0; j < hC; j++)
          for (int i=0; i < wC; i++) {
            RC[i+j*wC]=R0[(i/zoomfactor)+(j/zoomfactor)*w];
            GC[i+j*wC]=G0[(i/zoomfactor)+(j/zoomfactor)*w];
            BC[i+j*wC]=B0[(i/zoomfactor)+(j/zoomfactor)*w];              
          }
      }
      free(img0);
      delete[] R0;
      delete[] G0;
      delete[] B0;
    } else {
      //back background
      memset(RC, 0, wC*hC);
      memset(GC, 0, wC*hC);
      memset(BC, 0, wC*hC);
    }
    
    //draw centers of patches
    int kmax, icolor, cR, cG, cB;
    double pmax;
    for (int n=0; n < npatches; n++) { //for each patch
      i=(int) info[n].cx;
      j=(int) info[n].cy;
      kmax=0;
      pmax=probnk[n][kmax];
      //find Gaussian for which prob(gaussian | given patch) is highest
      for (int k=1; k < ngauss; k++) {
        if (probnk[n][k] > pmax) {
          pmax=probnk[n][k];
          kmax=k; //index of Gaussian
        }
      }
      //draw central pixel with a different color for each Gaussian
      icolor=kmax%ncolors;
      cR=colorR[icolor];
      cG=colorG[icolor];
      cB=colorB[icolor];
      if (zoomfactor == 1) {
        RC[i+j*w]=cR;
        GC[i+j*w]=cG;
        BC[i+j*w]=cB;
        if (i > 0) {
          RC[(i-1)+j*w]=cR;
          GC[(i-1)+j*w]=cG;
          BC[(i-1)+j*w]=cB;
        }
        if (j > 0) {
          RC[i+(j-1)*w]=cR;
          GC[i+(j-1)*w]=cG;
          BC[i+(j-1)*w]=cB;
        }
        if (i+1 < w) {
          RC[(i+1)+j*w]=cR;
          GC[(i+1)+j*w]=cG;
          BC[(i+1)+j*w]=cB;
        }
        if (j+1 < h) {
          RC[i+(j+1)*w]=cR;
          GC[i+(j+1)*w]=cG;
          BC[i+(j+1)*w]=cB;
        }
      } else {
        RC[(i*zoomfactor+zoomfactor/2)+(j*zoomfactor+zoomfactor/2)*wC]=cR;
        GC[(i*zoomfactor+zoomfactor/2)+(j*zoomfactor+zoomfactor/2)*wC]=cG;
        BC[(i*zoomfactor+zoomfactor/2)+(j*zoomfactor+zoomfactor/2)*wC]=cB; 
        int ii, jj;
        for (int kk=0; kk < 4; kk++) {
          ii=-1; jj=0;
          if (kk == 1) { ii=1; jj=0; }
          if (kk == 2) { ii=0; jj=-1; }
          if (kk == 3) { ii=0; jj=1; }
          RC[(i*zoomfactor+zoomfactor/2+ii)+\
             (j*zoomfactor+zoomfactor/2+jj)*wC]=cR;
          GC[(i*zoomfactor+zoomfactor/2+ii)+\
             (j*zoomfactor+zoomfactor/2+jj)*wC]=cG;
          BC[(i*zoomfactor+zoomfactor/2+ii)+\
             (j*zoomfactor+zoomfactor/2+jj)*wC]=cB;           
        }
      }
      
    }
    //save image
    RGB2output(RC, GC, BC, imgC, wC*hC);
    sprintf(nameimout, "%s_probGM.png", nameout);
    io_png_write_u8((const char *) nameimout, imgC, wC, hC, 3); //3 channels
    free(imgC);
    delete[] RC;
    delete[] GC;
    delete[] BC;
  
  
  
    //for each Gaussian, draw 10 patches closest to the reference patch 
    if (closest) {
      int *labels=new int[npatches];
      int kmax;
      double pmax;
      for (int n=0; n < npatches; n++) { //for each patch
        kmax=0;
        pmax=probnk[n][kmax];
        //find Gaussian for which prob(gaussian | given patch) is highest
        for (int k=1; k < ngauss; k++) {
          if (probnk[n][k] > pmax) {
            pmax=probnk[n][k];
            kmax=k; //index of Gaussian
          }
        }
        labels[n]=kmax; //label of patch = index of Gaussian with highest prob
      }
      
      for (int k=0; k < ngauss; k++) { //for each Gaussian
        int npatchesK=0;
        //assume that reference patch is patches[0], 
        //so do not include it in computations
        for (int n=1; n < npatches; n++) 
          if (labels[n] == k) npatchesK++; //number of patches assigned to
                                           //the Gaussian
          
        //compute L2-distances between patches assigned to the Gaussian
        //and the reference patch
        struct data_distancesP *datadist=new struct data_distancesP[npatchesK];
        int idK=0;
        double d;
        for (int n=1; n < npatches; n++) {
          if (labels[n] == k) { //assume that reference patch is patches[0]
            d=distance_patchesP(patches[0], patches[n], dim);
            datadist[idK].idpatch=n;
            datadist[idK].d=d;
            idK++;
          }
        }
        //sort distances
        qsort(datadist, npatchesK, sizeof(struct data_distancesP), 
              order_increasing_distP);
        //draw 10 closest patches to reference patch
        //create patches image (horizontal)
        int ntop=(npatchesK > 10)?(10):(npatchesK);
        int margin=2;
        int spatch, nchannels;
        float v=sqrt((double) dim);
        int vi=(int) v;
        if (v == vi) { //dim=wpatch^2 (1 channel image)
          spatch=vi;
          nchannels=1;
        } else { //dim != wpatch^2, assume dim/3 = wpatch^2 (3 channels image)
          spatch=(int) sqrt((double) (dim/3));
          nchannels=3;
        }
        int wt=spatch*ntop+margin*(ntop-1);
        int ht=spatch;  
        unsigned char *I=new unsigned char[wt*ht];
        unsigned char *IG=NULL, *IB=NULL;
        if (nchannels > 1) {
            IG=new unsigned char[wt*ht];
            IB=new unsigned char[wt*ht];
            memset(IG, 0, wt*ht);
            memset(IB, 0, wt*ht);
        }
        unsigned char *img2=(unsigned char *) malloc(3*wt*ht);
        char nameimout[120];
        //normalize mean and std of intensity of displayed patch to 128 and 40
        float Iout, Ioutden;
        memset(I, 0, wt*ht);
        for (int n=0; n < ntop; n++) {
          for (int j=0; j < spatch; j++) 
            for (int i=0; i < spatch; i++) {
              if (nchannels == 1) 
                Iout=patches[datadist[n].idpatch][i+j*spatch];
              else 
                Iout= (patches[datadist[n].idpatch][3*(i+j*spatch)]+\
                       patches[datadist[n].idpatch][3*(i+j*spatch)+1]+\
                       patches[datadist[n].idpatch][3*(i+j*spatch)+2])/3.0;
              //denormalize patch variance and mean, if it has been normalized
              Ioutden=Iout;
              if ((info[datadist[n].idpatch].a != 1) || 
                  (info[datadist[n].idpatch].b != 0)) { //normalized patch
                if (info[datadist[n].idpatch].a != 1) { //normalized variance 
                                                        //(std=1)
                  //renormalize to std_out=40
                  Ioutden=Iout*40.0f;
                }
                if (info[datadist[n].idpatch].b != 0) { //normalized mean 
                                                        //(mean=0)
                  //renormalize to mean_out=128
                  Ioutden=Ioutden+128.0f;
                }
              }
              if (Ioutden > 255) Ioutden=255;
              if (Ioutden < 0) Ioutden=0;
              if (nchannels == 1) { //grayscale patch
                I[i+n*(spatch+margin)+j*wt]=(int) Ioutden;
              } else { //color patch: apply same normalization factor 
                       //to all channels
                if (Iout != 0) {
                  float v;
                  v=patches[datadist[n].idpatch][3*(i+j*spatch)]*Ioutden/Iout;
                  if (v > 255) v=255;
                  I[i+n*(spatch+margin)+j*wt]=(int) (v+0.5f);
                  v=patches[datadist[n].idpatch][3*(i+j*spatch)+1]*Ioutden/Iout;
                  if (v > 255) v=255;
                  IG[i+n*(spatch+margin)+j*wt]=(int) (v+0.5f);
                  v=patches[datadist[n].idpatch][3*(i+j*spatch)+2]*Ioutden/Iout;
                  if (v > 255) v=255;
                  IB[i+n*(spatch+margin)+j*wt]=(int) (v+0.5f);
                }
              }
            }
        }
        if (nchannels == 1) 
          RGB2output(I, I, I, img2, wt*ht);
        else
          RGB2output(I, IG, IB, img2, wt*ht);
          
        //save image
        sprintf(nameimout, "%s_top10_%i.png", nameout, k);
        io_png_write_u8((const char *) nameimout, img2, wt, ht, 3); //3 channels 
        
        delete[] I;
        if (IG) delete[] IG;
        if (IB) delete[] IB;
        free(img2);
        delete[] datadist;
      }
      
      delete[] labels;
      labels=NULL;
    }
  
  

    //clean up
    free(img);
    delete[] R;
    delete[] G;
    delete[] B;
    if (info) delete_patch_info(info, ninfo);
    if (patches) {
        for (int n=0; n < npatches; n++) delete[] patches[n];
        delete[] patches;
    }
    if (gm) delete_GMMpatches(gm);
    if (labels) delete[] labels;
    if (probnk) {
      for (int n=0; n < npatches; n++) delete[] probnk[n];
      delete[] probnk;
    }
    
    return EXIT_SUCCESS;
}

