/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "../auxlibs/rgbcube/rgbcube.h"
#include "library/patches/patches_extraction_lib.h"
#include "library/patches/patches_analysis_lib.h"

//Main code to compute basis of a set of patches using PCA
//        
//Input: set of patches
//
//Output: basis of the set of patches, and corresponding variances, 
//        image displaying the basis vectors, 
//        image displaying the average patch of the input set,
//        graphical representation of the variance of each basis component

int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput = {(char *) "input", NULL, (char *) "input patches"};
    parameters.push_back(&pInput);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    
    std::vector <OptStruct *> options;

    if (!parsecmdline((char *) "patches_pca", (char *) "PCA of a set of patches", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namepatches = pInput.value;
    const char *nameout = pOutput.value;


    int npatches, dim;
    double **patches=NULL;
    double *dataPCA=NULL;    
    
    //Read patches data from file
    patches=read2Ddata(namepatches, npatches, dim);
    
    //Perform Principal Components Analysis (PCA) of the set of patches
    int sizePCA;
    dataPCA=PCApatches(patches, npatches, dim, sizePCA);
        
    //Save PCA data to file
    char namedataout[120];
    strcpy(namedataout, nameout);
    strcat(namedataout, ".dat");
    int label=2; //label=2 : auxiliar variable that indicates that the
                 //          stored data correspond to PCA 
    writeDdatalabel(dataPCA, sizePCA, label, (const char *) namedataout); 
    
    
    //image of PCA basis
    int wpatch;
    float v=sqrt((double) dim);
    int vi=(int) v;
    //get size of the patches
    if (v == vi) { //dim=wpatch^2 (1 channel image)
        wpatch=vi;
    } else { //dim != wpatch^2, assume dim/3 = wpatch^2 (3 channels image)
        wpatch=(int) sqrt((double) (dim/3));
    }
    int hpatch=wpatch;
    int npc=dim;
    int wmargin, hmargin;
    wmargin=hmargin=2; //2-pixels margin between displayed images of patches
    //compute width and height of output image
    //dim patches are displayed, we try to arrange them into a square image
    int nrow=(int) sqrt((double) npc);
    int wout=nrow*wpatch+(nrow-1)*wmargin;
    int ncol=(int) ceil((double) npc/(double) nrow);
    int hout=ncol*hpatch+(ncol-1)*hmargin;
    unsigned char *Rout=new unsigned char[wout*hout];
    unsigned char *Gout=new unsigned char[wout*hout];
    unsigned char *Bout=new unsigned char[wout*hout];
    unsigned char *img=(unsigned char *) malloc(3*wout*hout);
    memset(Rout, 0, wout*hout);
    memset(Gout, 0, wout*hout);
    memset(Bout, 0, wout*hout);
    //draw basis vectors as patches
    PCAbaseview(Rout, Gout, Bout, wpatch, hpatch, dataPCA, dim, npc, nrow);	
    //save image
    RGB2output(Rout, Gout, Bout, img, wout*hout);
    char nameimout[120];
    strcpy(nameimout, nameout);
    strcat(nameimout, ".png");
    io_png_write_u8((const char *) nameimout, img, wout, hout, 3); //3 channels
  
    
    //clean up
    free(img);
    delete[] Rout;
    delete[] Gout;
    delete[] Bout;
    
    
    
    //image of average patch
    wout=wpatch;
    hout=hpatch;
    Rout=new unsigned char[wout*hout];
    Gout=new unsigned char[wout*hout];
    Bout=new unsigned char[wout*hout];
    img=(unsigned char *) malloc(3*wout*hout);
    //draw average patch
    PCAmeanview(Rout, Gout, Bout, wpatch, hpatch, dataPCA, dim);
    //save image to file
    RGB2output(Rout, Gout, Bout, img, wout*hout);
    char nameimout2[120];
    strcpy(nameimout2, nameout);
    strcat(nameimout2, "_mean.png");
    io_png_write_u8((const char *) nameimout2, img, wout, hout, 3); //3 channels
    //clean up
    free(img);
    delete[] Rout;
    delete[] Gout;
    delete[] Bout;
    
    
    //image of eigenvalues
    unsigned char *Ivp=new unsigned char[dim*dim];
    memset(Ivp, 0, dim*dim);
    for (int k=0; k < dim; k++) {
        for (int j=dim-1; j >= dim-(int) dim*dataPCA[dim+k]/dataPCA[dim]; j--)
            Ivp[k+j*dim]=255;
    }
    unsigned char *imgEigV=(unsigned char *) malloc(3*dim*dim);
    RGB2output(Ivp, Ivp, Ivp, imgEigV, dim*dim);
    char nameimout3[120];
    strcpy(nameimout3, nameout);
    strcat(nameimout3, "_EigV.png");
    io_png_write_u8((const char *) nameimout3, imgEigV, dim, dim, 3); //3 channels
    //clean up
    free(imgEigV);
    delete[] Ivp;

    //standard deviation of PCA: sqrt(average(eigenvalues))
    double stdPCA=0;
    for (int k=0; k < dim; k++) stdPCA+=dataPCA[dim+k];
    stdPCA=sqrt(stdPCA)/sqrt(dim);
    FILE *faux=fopen("pcastd.txt", "wt");
    fprintf(faux, "std: %2.2f\n", stdPCA);
    fclose(faux);
    
    if (patches) {
        for (int n=0; n < npatches; n++) delete[] patches[n];
        delete[] patches;
    }
    if (dataPCA) delete[] dataPCA;
    
    return EXIT_SUCCESS;
    
}

