/*
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../auxlibs/parser/parser.h"
#include "../auxlibs/io_png/io_png.h"
#include "../auxlibs/io_RGB/io_RGB.h"
#include "../auxlibs/rgbcube/rgbcube.h"
#include "library/patches/patches_extraction_lib.h"
#include "library/patches/patches_analysis_lib.h"

//Main code to synthesize patches using a random linear combination 
//of basis vectors
//
//Input: basis of the space of patches, and corresponding standard deviations
//
//Output: image of 8 synthesized patches. If the basis are from a Gaussian
//        Mixture, one image of synthesized patches is generated for each Gaussian


int main(int argc, char **argv)
{
    std::vector<ParStruct *> parameters;
    ParStruct pInput = {(char *) "input", NULL, (char *) "input basis"};
    parameters.push_back(&pInput);
    ParStruct pOutput = {(char *) "output", NULL, (char *) "root name of output files"};
    parameters.push_back(&pOutput);
    
    std::vector <OptStruct *> options;
    OptStruct oM  = {(char *) "m:", 0, (char *) "120", NULL, (char *) "Mean of output patches"}; 
    options.push_back(&oM);
    OptStruct oS  = {(char *) "s:", 0, (char *) "40", NULL, (char *) "Standard deviation of output patches"}; 
    options.push_back(&oS);

    if (!parsecmdline((char *) "patches_sim", (char *) "Simulate patches patches", \
                  argc, argv, options, parameters))
        return EXIT_FAILURE;

    const char *namebasis = pInput.value;
    const char *nameout = pOutput.value;


    //Optional: normalize mean and standard deviation for better visualization
    //if stdRef==0 -> do not normalize
    float meanRef=0, stdRef=0;
    if (oM.flag) meanRef = atof(oM.value);
    if (oS.flag) stdRef = atof(oS.value);

      
    int type=0;
    type=readlabel(namebasis); //assume binary file and first value is integer
    printf("type=%i  file=%s\n", type, namebasis);
  
    double ***patchesSim=NULL;
    double *basisPCA=NULL;
    struct GMMpatches *gm=NULL;
    int dim=0, nsim;
    int npatchesSim=8;
    //nsim=number of output images
    //if PCA: 1 image
    //if GM: 1 image per Gaussian component
    if (type == 2) { //type 2: PCA
      int sizedata;
      //read PCA basis data from file
      basisPCA=readDdatalabel(namebasis, sizedata);
      dim=(int) (sqrt(1+ (double) sizedata)-1);
      patchesSim=new double**[1];
      nsim=1;
      double **vec=new double*[dim];
      for (int k=0; k < dim; k++) vec[k]=basisPCA+(2+k)*dim;
      //synthesize patches
      patchesSim[0]=simpatches(basisPCA, basisPCA+dim, vec, dim);
      delete[] vec;
    } 
    if (type == 3) { //type 3: Gaussian Mixture
      //read GM sata from file
      gm=readGMMdata(namebasis);
      //compute basis of for each Gaussian
      computeGMMPCA(gm);
      dim=gm->dim;
      patchesSim=new double**[gm->ngaussian];
      nsim=gm->ngaussian;
      for (int k=0; k < gm->ngaussian; k++) {
        //synthesize patches
        patchesSim[k]=simpatches(gm->GM[k].mean, gm->GM[k].pcaVal, 
                                 gm->GM[k].pcaVec, dim);
      }
    }
    
  
    //create patches image (horizontal)
    int margin=2; //2 pixels marging between synthesized patches
    int nchannels;
    int spatch;
    float v=sqrt((double) dim);
    int vi=(int) v;
    if (v == vi) { //dim=wpatch^2 (1 channel image)
        spatch=vi;
        nchannels=1;
    } else { //dim != wpatch^2, assume dim/3 = wpatch^2 (3 channels image)
        spatch=(int) sqrt((double) (dim/3));
        nchannels=3;
    }
    int w=spatch*npatchesSim+margin*(npatchesSim-1);
    int h=spatch;  
    unsigned char *I=new unsigned char[w*h];
    unsigned char *IG=NULL, *IB=NULL;
    if (nchannels > 1) {
        IG=new unsigned char[w*h];
        memset(IG, 0, w*h);
        IB=new unsigned char[w*h];
        memset(IB, 0, w*h);
    }
    unsigned char *img=(unsigned char *) malloc(3*w*h);
    char nameimout[120];
    float Iout, Ioutden;
    double a, b;
    for (int k=0; k < nsim; k++) { //for each output image
      memset(I, 0, w*h);
      for (int n=0; n < npatchesSim; n++) { //for each simulated patch
      
        //if stdRef flag is active (!= 0), then
        //compute intensity normalization parameters a,b such that
        //the mean and standard deviation of simulated patch are 
        //meanRead and stdRef
        if (stdRef != 0) {
           double mean=0, std=0, size=0;
           for (int j=0; j < spatch; j++) 
             for (int i=0; i < spatch; i++) {
                if (nchannels == 1)
                    Iout=(int) patchesSim[k][n][i+j*spatch];
                else
                    Iout=(int) (patchesSim[k][n][3*(i+j*spatch)]+ \
                                patchesSim[k][n][3*(i+j*spatch)+1]+ \
                                patchesSim[k][n][3*(i+j*spatch)+2])/3.0f;
                mean+=Iout;
                std+=(Iout*Iout);
                size+=1.0f;
             }
           mean/=size;
           std/=size;
           std-=(mean*mean);
           std=sqrt(std);
           if (std != 0) { 
               a=stdRef/std;
               b=meanRef-a*mean;
           } else {
               a=1;
               b=0;
           }
        }
        
        
        for (int j=0; j < spatch; j++) 
          for (int i=0; i < spatch; i++) {
            //get intensity component
            if (nchannels == 1) //grayscale image
              Iout=(int) patchesSim[k][n][i+j*spatch];
            else //color image
              Iout=(int) (patchesSim[k][n][3*(i+j*spatch)]+\
                          patchesSim[k][n][3*(i+j*spatch)+1]+\
                          patchesSim[k][n][3*(i+j*spatch)+2])/3.0f;
            Ioutden=Iout;
            //if stdRef flag is active (!= 0), then
            //normalize mean and std of intensity component
            //using a and b parameters computed above
            if (stdRef != 0) {             
                Ioutden=a*Iout+b;
            }
            //clip out-of-range parameters
            if (Ioutden > 255) Ioutden=255;
            if (Ioutden < 0) Ioutden=0;
            if (nchannels == 1) { //grayscale image
                I[i+n*(spatch+margin)+j*w]=(int) (Ioutden+0.5f);
            } else { //color image: apply normalization factor to each channel
                float v;
                v=patchesSim[k][n][3*(i+j*spatch)]*Ioutden/Iout;
                if (v > 255) v=255;
                I[i+n*(spatch+margin)+j*w]=(int) (v+0.5f);
                v=patchesSim[k][n][3*(i+j*spatch)+1]*Ioutden/Iout;
                if (v > 255) v=255;
                IG[i+n*(spatch+margin)+j*w]=(int) (v+0.5f);
                v=patchesSim[k][n][3*(i+j*spatch)+2]*Ioutden/Iout;
                if (v > 255) v=255;
                IB[i+n*(spatch+margin)+j*w]=(int) (v+0.5f);
            }
          }
      }
        
      if (nchannels == 1)
          RGB2output(I, I, I, img, w*h);
      else
          RGB2output(I, IG, IB, img, w*h);
        
      if (basisPCA) sprintf(nameimout, "%s.png", nameout);
      else sprintf(nameimout, "%s_%i.png", nameout, k);
      
      //save image
      io_png_write_u8((const char *) nameimout, img, w, h, 3); //3 channels
    }
  
  
    //clean up
    if (basisPCA) delete[] basisPCA;
    if (gm) delete_GMMpatches(gm);
    if (patchesSim) {
      for (int k=0; k < nsim; k++) {
        if (patchesSim[k]) {
          for (int n=0; n < npatchesSim; n++) delete[] patchesSim[k][n];
          delete[] patchesSim[k];
        }
      }
      delete[] patchesSim;
    }
    free(img);
    delete[] I;
    if (IG) delete[] IG;
    if (IB) delete[] IB;

    return EXIT_SUCCESS;

}
    

